: run this script through /bin/sh

# bootproxy.sh 192.103.140.1 10 6 192.103.140.1 102

IP="$1"
TARGET="$2"
CONTEXT="$3"
DEST="$4"
OFFSET="$5"
if [ -z "$OFFSET" ]; then
    OFFSET=0
fi
COMMUNITY="$6"

if [    -z "$IP" \
     -o -z "$TARGET" \
     -o "$TARGET" -lt 7 \
     -o -z "$CONTEXT" \
     -o "$CONTEXT" -lt 3 \
     -o -z "$DEST" \
     -o \( "$IP"  = "$DEST" -a "$OFFSET" -lt 100 \) \
     -o \( "$IP" != "$DEST" -a -z "$COMMUNITY" \) ]; then
    echo "usage: bootproxy.sh proxy-IP party-ID ctx-ID dest-IP [offset [community]]" 1>&2
    exit 1
fi
if echo "$IP" | awk '
/^[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*$/ { next; }
	{ exit(1); }
'; then
    true
else
    echo "invalid IP-address: $IP" 1>&2
    echo "usage: bootproxy.sh proxy-IP party-ID ctx-ID dest-IP [offset [community]]" 1>&2
    exit 1
fi
if echo "$DEST" | awk '
/^[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*$/ { next; }
	{ exit(1); }
'; then
    true
else
    echo "invalid IP-address: $IP" 1>&2
    echo "usage: bootproxy.sh proxy-IP party-ID ctx-ID dest-IP [offset [community]]" 1>&2
    exit 1
fi

HARPO=initialPartyId.$IP.$TARGET
DUCKSOUP=initialContextId.$IP.$CONTEXT

F=/tmp/bootparty.$$
trap "rm -f $F" 0 1 2 3 13 15
touch $F
chmod 0600 $F

ps axl | sed -e 's% %%g' | awk '
	{
	    if ($NR == 1) next;
	    printf "Y%d=\"%cecho %c%s%c | /usr/lib/makekey | awk ", \
		flip, 96, 39, substr($1,9,10), 39;
	    printf "%c{ printf substr($1,1,8) }%c%c\"\n", 39, 39, 96;
	    if (flip) flip = 0; else flip = 1;
	    if (!flip) { k++; printf "K%d=\"$Y0$Y1\"\n", k; if (k == 8) exit; }
	}
' > $F
. $F

if [ -z "$COMMUNITY" ]; then
    COMMUNITY="$K3"
fi


# first, create proxy destination

snmpi -p manager -a 3 -m 4 set                      \
           partyTDomain.$HARPO=rfc1157Domain        \
          partyTAddress.$HARPO=0x$DEST.$OFFSET.161  \
    partyMaxMessageSize.$HARPO=8192                 \
         partyCloneFrom.$HARPO=initialPartyId.$IP.3 \
       partyAuthPrivate.$HARPO="$COMMUNITY"         \
            partyStatus.$HARPO=createAndGo || exit 1

# second, create context

snmpi -p manager -a 3 -m 4 set            \
    contextProxyDstParty.$DUCKSOUP=$HARPO \
           contextStatus.$DUCKSOUP=createAndGo || exit 1

DIDX=`snmpi -p manager -a 3 -m 4 get contextIndex.$DUCKSOUP | sed -e 's%.*=%%'`


# third, create access policy

snmpi -p manager -a 3 -m 4 set          \
    aclPrivileges.1.1.$DIDX=128         \
        aclStatus.1.1.$DIDX=createAndGo \
    aclPrivileges.1.2.$DIDX=35          \
        aclStatus.1.2.$DIDX=createAndGo \
    aclPrivileges.2.1.$DIDX=4           \
        aclStatus.1.2.$DIDX=createAndGo \
    aclPrivileges.3.4.$DIDX=43          \
        aclStatus.3.4.$DIDX=createAndGo \
    aclPrivileges.4.3.$DIDX=4           \
        aclStatus.4.3.$DIDX=createAndGo || exit 1

exit 0
