/* isode-snmp.c - 4BSD/ISODE SNMP driver */

/* TODO:

   1. Secret update
 */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/*
 * $Header$
 *
 *
 * $Log$
 */

/*			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <varargs.h>
#ifdef	BSD42
#include <sys/ioctl.h>
#endif
#ifdef	SYS5
#include <fcntl.h>
#endif

#ifndef	KZM
#include <isode/snmp/objects.h>
#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/party.h>
#include <isode/snmp/defsdir.h>
#include <isode/dgram.h>
#include <isode/internet.h>
#include <isode/isoaddrs.h>
#include <isode/sys.file.h>
#include <isode/tailor.h>
#else
#include <objects.h>
#include <SNMP-types.h>
#include <party.h>
#include <defsdir.h>
#include <dgram.h>
#include <internet.h>
#include <isoaddrs.h>
#include <sys.file.h>
#include <tailor.h>
#endif

#include <tk.h>

/*    ENTRY */

#define	MAGIC


static  Tcl_Interp *tclInterp;
static  char   *objects;
static  char   *pParty;
static  int     partysync;
static  int    *watch;



int	NewSNMPcommand (), SNMPhandleCommand ();
void	SNMPdeleteCommand ();

int	SNMPinfoCommand ();

#ifdef	MAGIC
int	SNMPmiscCommand ();
#endif

int	SNMPoid2ode ();



/* SNMP */

struct oHandle {
    int			o_rtrqos;
#define	RTR_EXPONENTIAL	0
#define	RTR_SUPERLINEAR	1
    int			o_retries;
#define	MAXTRIES	10
    u_long		o_timeout;

    int			o_secqos;
#define	SEC_NONE	0x00
#define	SEC_AUTH	0x01
#define	SEC_PRIV	0x02
#define	SEC_ALL		(SEC_AUTH | SEC_PRIV)

    int			o_maxrept;
    int			o_nonrptr;

#ifdef	MAGIC
    int			o_magic;
#endif
};

struct request {
    int			r_id;
    u_long		r_time;
    Tk_TimerToken	r_token;
    int			r_waiting;
};

struct iHandle {
    struct iHandle     *i_next;

    Tcl_Interp	       *i_interp;
    struct pHandle     *i_pHandle;

    char	       *i_callback;
#define	GET_CLOCKS	"pInternal:get_clocks"
    char	       *i_pname;
    u_long		i_id;

    struct request     *i_attempts;
    struct request     *i_current;

    struct snmp_access *i_acl;
    struct type_SNMP_SnmpAuthMsg
		       *i_msg;

    struct oHandle	i_options;
#define	i_rtrqos    i_options.o_rtrqos
#define	i_retries   i_options.o_retries
#define	i_timeout   i_options.o_timeout
#define	i_secqos    i_options.o_secqos
#define	i_maxrept   i_options.o_maxrept
#define	i_nonrptr   i_options.o_nonrptr
#ifdef	MAGIC
#define	i_magic	    i_options.o_magic
#endif

    int			i_zero;
    int			i_update;
    struct snmp_party  *i_sync;
};

struct nHandle {
    struct nHandle     *n_next;
    char	       *n_name;
};

struct pHandle {
    struct pHandle     *p_next;

    char	       *p_name;

    Tcl_Interp	       *p_interp;
    struct nHandle     *p_inform;
    struct nHandle     *p_trap;
    int			p_selected;

    struct snmp_context
		       *p_context;
    struct snmp_access *p_acl;

    struct oHandle	p_options;
#define	p_rtrqos   p_options.o_rtrqos
#define	p_retries   p_options.o_retries
#define	p_timeout   p_options.o_timeout
#define	p_secqos    p_options.o_secqos
#define	p_maxrept   p_options.o_maxrept
#define	p_nonrptr   p_options.o_nonrptr
#ifdef	MAGIC
#define	p_magic	    p_options.o_magic
#endif
    
    struct iHandle     *p_invokes;

    char	       *p_resources;
};
static	struct pHandle *pHead = NULL;
static	struct pHandle *pInternal = NULL;


static struct pair {
    char   *p_old;
    char   *p_new;
}	pairs[] = {
    "INTEGER",     "Integer32",
    "Counter",     "Counter32",
    "Gauge",       "Gauge32",

    NULL
};


static	OID	nullSpecific = NULL;
static	OT	partyAClock = NULL;
static	OT	partyAPrivate = NULL;
static	OT	partyAProtocol = NULL;
static	OT	partyMMessageSize = NULL;
static	PE	unSpecified = NULL;


static	u_long	lastRequest = 1;


#ifdef	MAGIC
#define	MAGIC_NONE		0
#define	MAGIC_NOTMINLEN		1
#define	MAGIC_CONSTRING		2
#define	MAGIC_INDFLEN		3
#define	MAGIC_NONUNITVR		4
#define	MAGIC_NULLCOMM		5
#define	MAGIC_BIGREQID		6
#define	MAGIC_NONZEROES		7
#define	MAGIC_NONZEROEI		8
#define	MAGIC_BDSTPARTY		9
#define	MAGIC_NDSTMATCH	       10
#define	MAGIC_BSRCPARTY	       11
#define	MAGIC_NSRCMATCH	       12
#define	MAGIC_BCONTEXT	       13
#define	MAGIC_BERLESUDP	       14
#define	MAGIC_UDPLESBER	       15

static	int	magicommand = MAGIC_NONE;
#endif


int	send_message (), send_v1_message (), send_v2_message ();

void	recv_message ();

void	lost_message ();

char   *snmp_error (), *snmpV2_exception ();



/* INVOCATIONS */

void	free_iHandle ();
u_long	rto ();



/* OPTIONS */

int	make_oHandle (), conf_pHandle ();



/* ACCESS CONTROL */

struct snmp_access *get_sa ();



/* CONTEXT */

struct snmp_context *find_context (), *fuzzy_context (), *make_context ();



/* PARTY */

static	int		wantsync = 0;	/* sleep short for next sync */
static	u_long		synctime = 0L;	/* time of next clock sync,
					   0 indicates in progress */
static	Tk_TimerToken	synctoken;	/* scheduler token */


void	start_clocks (), find_clocks (), get_clocks (), get_clocks_callback ();
struct snmp_party *make_party ();



/* DGRAM */

struct dHandle {
    int		    d_sd;
    PS		    d_ps;

    int		    d_errno;

    int		    d_tdomain;
    struct sockaddr d_agent;
    struct sockaddr d_manager;
    
    struct snmp_party
		   *d_lastsend;
    struct sockaddr d_lastrecv;

    int		    d_maxsize;

    struct dHandle *d_next;

    int		    d_pktsize;
    char	    d_packet[1];
};
static struct dHandle *dHead = NULL;


struct dHandle *find_dHandle (), *make_dHandle ();
int	make_dHandles ();
#ifdef	TCP
struct dHandle *make_udp_dHandle ();
#endif
int	read_dHandle (), write_dHandle ();



/* MIBS */

static Tcl_HashTable agentTable;
static Tcl_HashTable groupTable;


struct iinfo {
    OT	    ii_ot;
    int	    ii_length;		/*  > 0: then that
				   == 0: IMPLICIT variable-length
				    < 0: EXPLICIT variable-length */
};


int	resolve_iit ();
void	append_iit ();


/* LIBC */

#ifndef	SYS5
long	random ();
#endif
long	time ();

/*    ENTRY */

int	SNMP_Init (interp, myname, oflag, pflag, sflag, wflag, library)
Tcl_Interp     *interp;
char   *library,
       *myname,
       *oflag,
       *pflag;
int	sflag,
       *wflag;
{
    int	    i;
    char   *cp,
	    buffer[BUFSIZ];
    register struct pHandle *p = NULL;
    struct timeval tvs;
    OT	    ot;

    _tailor_logging_only = 1;
    isodetailor (myname, 1);

    tclInterp = interp;
    objects = oflag;
    pParty = pflag;
    partysync = sflag;
    watch = wflag;

    cp = Tcl_GetVar2 (interp, "env", "SNMPTCL_LIBRARY", TCL_GLOBAL_ONLY);
    (void) strcpy (library, cp ? cp : isodefile ("snmptcl", 0));

#ifdef	SYS5
    (void) srand ((unsigned int) time ((long *) 0));
    lastRequest = ((int) rand ()) & 0x7fffffff;
#else
    (void) srandom ((int) time ((long *) 0));
    lastRequest = ((int) random ()) & 0x7fffffff;
#endif
    if (lastRequest == 0)
	lastRequest++;

    ps_len_strategy = PS_LEN_LONG;

    if (loadobjects (objects) == NOTOK) {
	(void) sprintf (buffer, "loadobjects: %s", PY_pepy);
	goto tcl_error;
    }

    for (ot = anchor; ot; ot = ot -> ot_next)
	if (!ot -> ot_syntax
		&& (ot -> ot_index || ot -> ot_augments)
	        && (i = strlen (ot -> ot_text)) > 5
	        && !strcmp (ot -> ot_text + i - 5, "Entry")) {
	    OT	    ot2,
		    otn;

	    otn = ot;
	    for (ot2 = ot -> ot_children; ot2; ot2 = ot2 -> ot_sibling) {
		otn = ot2;
		ot2 -> ot_iit = (caddr_t) ot;
	    }

	    ot = otn;
	}

    if (!(nullSpecific = text2oid ("0.0"))) {
	(void) strcpy (buffer, "text2oid (\"0.0\") failed");
	goto tcl_error;
    }

    if (!(partyAClock = text2obj ("partyAuthClock"))) {
	(void) strcpy (buffer, "text2obj (\"partyAuthClock\") failed");
	goto tcl_error;
    }
    if (!(partyAPrivate = text2obj ("partyAuthPrivate"))) {
	(void) strcpy (buffer, "text2obj (\"partyAuthPrivate\") failed");
	goto tcl_error;
    }
    if (!(partyAProtocol = text2obj ("partyAuthProtocol"))) {
	(void) strcpy (buffer,  "text2obj (\"partyAuthProtocol\") failed");
	goto tcl_error;
    }
    if (!(partyMMessageSize = text2obj ("partyMaxMessageSize"))) {
	(void) strcpy (buffer,  "text2obj (\"partyMaxMessageSize\") failed");
	goto tcl_error;
    }

    if (!(unSpecified = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL))){
no_mem: ;
	(void) strcpy (buffer, "out of memory");
	goto tcl_error;
    }

    if (init_party (pParty, MANAGER, 0, &pParty) == NOTOK) {
	(void) strcpy (buffer, PY_pepy);
	goto tcl_error;
    }

    if (!(p = (struct pHandle *) calloc (1, sizeof *p))
	    || !(p -> p_name = strdup ("pInternal")))
	goto no_mem;
    pInternal = pHead = p;
    p -> p_interp = interp;

    p -> p_rtrqos = RTR_EXPONENTIAL;
    p -> p_retries = 5;
    p -> p_timeout = 3 * 1000L;
    p -> p_secqos = SEC_NONE;
    p -> p_maxrept = 1;
    p -> p_nonrptr = 0;

    if (make_dHandles (interp) != TCL_OK)
	goto tcl_error_aux;

    (void) gettimeofday (&tvs, (struct timezone *) 0);
    synctime = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L;
    start_clocks ((ClientData) 0);

    Tcl_InitHashTable (&agentTable, TCL_STRING_KEYS);
    Tcl_InitHashTable (&groupTable, TCL_STRING_KEYS);

    Tcl_CreateCommand (interp, "newsnmp", NewSNMPcommand,
		       (ClientData) 0, (void (*) ()) 0);
    Tcl_CreateCommand (interp, "snmpinfo", SNMPinfoCommand,
		       (ClientData) 0, (void (*) ()) 0);
#ifdef	MAGIC
    Tcl_CreateCommand (interp, "snmpmisc", SNMPmiscCommand,
		       (ClientData) 0, (void (*) ()) 0);
#endif

/* defined in C for speed (ha!) */
    Tcl_CreateCommand (interp, "snmptcl_proc_oid2ode", SNMPoid2ode,
		       (ClientData) 0, (void (*) ()) 0);

    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
tcl_error_aux: ;
    if (nullSpecific)
	oid_free (nullSpecific), nullSpecific = NULL;
    if (unSpecified)
	pe_free (unSpecified), unSpecified = NULL;
    if (p) {
	if (p -> p_name)
	    free (p -> p_name);
	free ((char *) p);
	pHead = pInternal = NULL;
    }
    return TCL_ERROR;
}

/*  */

/* ARGSUSED */

int	SNMP_Exit (interp)
Tcl_Interp     *interp;
{
    if (partysync && party_changed) {
	char    buffer[BUFSIZ],
		file[BUFSIZ];

	(void) sprintf (file, "%spublic", pParty);
	if (access (file, W_OK) != NOTOK) {
	    (void) sprintf (buffer,
			    "tk_dialog .modal Dialog {%s} question 0 Yes No!",
			    "Synchronize party database before quiting?");
	    if (Tcl_Eval (interp, buffer) == TCL_OK
		    && *interp -> result == '0'
		    && sync_party (pParty, MANAGER, 0) == NOTOK) {
		(void) sprintf (buffer, "stable storage suspect: %s", PY_pepy);
		Tcl_SetResult (interp, buffer, TCL_VOLATILE);
		return TCL_ERROR;
	    }
	}
    }

    return TCL_OK;
}

/*    TCL */

/* ARGSUSED */

static int  NewSNMPcommand (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    char    buffer[BUFSIZ];
    register struct pHandle *p = NULL;

    if (argc < 3) {
	(void) sprintf (buffer, "argc=%d; expecting at least 3", argc);
	goto tcl_error;
    }
    argc--, argv++;

    if (!(p = (struct pHandle *) calloc (1, sizeof *p))) {
no_mem: ;
	(void) strcpy (buffer, "out of memory");
	goto tcl_error;
    }
    p -> p_interp = interp;

    if (!(p -> p_name = strdup (*argv)))
	goto no_mem;
    argc--, argv++;

    p -> p_rtrqos = RTR_EXPONENTIAL;
    p -> p_retries = 5;
    p -> p_timeout = 3 * 1000L;
    p -> p_secqos = SEC_NONE;
    p -> p_maxrept = 1;
    p -> p_nonrptr = 0;
    if (make_oHandle (interp, &p -> p_options, argc - 1, argv + 1) == NOTOK
	    || !(p -> p_context = find_context (interp, *argv)))
	goto tcl_error_aux;
    if (pHead)
	p -> p_next = pHead;
    pHead = p;
    Tcl_CreateCommand (interp, p -> p_name, SNMPhandleCommand, (ClientData) p,
		       SNMPdeleteCommand);

    Tcl_SetResult (interp, p -> p_name, TCL_VOLATILE);
    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
tcl_error_aux: ;
    if (p) {
	if (p -> p_name)
	    free (p -> p_name);
	free ((char *) p);
    }
    return TCL_ERROR;
}

/*  */

static int  SNMPhandleCommand (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    int	    opcode,
	    opmask,
	    vecp;
    u_long  timeout;
    char   *cp,
	    buffer[BUFSIZ],
	   *pname,
	  **vb,
	  **vec = NULL;
    register struct pHandle *p = (struct pHandle *) clientData;
    register struct iHandle *q = NULL;
    struct type_SNMP_PDUs *pdu;
    struct type_SNMP_PDU *parm;
    struct type_SNMP_VarBindList **vp;
    struct snmp_party *agentParty;

    if (argc < 2) {
	(void) sprintf (buffer, "argc=%d; expecting at least 2", argc);
	goto tcl_error;
    }
    pname = *argv;
    argc--, argv++;

    switch (*(cp = *argv)) {
	case 'b':
	    if (strcmp (cp, "bulk") == 0) {
		opcode = type_SNMP_PDUs_get__bulk__request;
		opmask = PARTY_PRIV_BULK;
		break;
	    }
	    goto none;

	case 'c':
	    if (strcmp (cp, "config") == 0) {
		argc--, argv++;

		if (argc > 2) {
extra_args: ;
		    (void) strcpy (buffer, "extra arguments");
		    goto tcl_error;
		}

		return conf_pHandle (interp, *argv, *(argv + 1), p);
	    }
	    goto none;

	case 'd':
	    if (strcmp (cp, "destroy") == 0) {
		if (argc > 1)
		    goto extra_args;

		Tcl_DeleteCommand (interp, p -> p_name);

		Tcl_ResetResult (interp);
		return TCL_OK;
	    }
	    goto none;

	case 'g':
	    if (strcmp (cp, "get") == 0) {
		opcode = type_SNMP_PDUs_get__request;
		opmask = PARTY_PRIV_GET;
		break;
	    }
	    goto none;

	case 'i':
	    if (strcmp (cp, "inform") == 0) {
		opcode = type_SNMP_PDUs_inform__request;
		opmask = PARTY_PRIV_INFO;
		break;
	    }
	    goto none;

	case 'n':
	    if (strcmp (cp, "next") == 0) {
		opcode = type_SNMP_PDUs_get__next__request;
		opmask = PARTY_PRIV_NEXT;
		break;
	    }
	    goto none;

	case 'r':
	    if (strcmp (cp, "register") == 0) {
		int	addP = 1;
		register struct nHandle  *n,
					**np;
		argc--, argv++;

register_again: ;
		if (argc < 2) {
		    (void) strcpy (buffer,
				    "need a notification type and a callback");
		    goto tcl_error;
		}

		if (strcmp (*argv, "-add") == 0) {
		    addP = 1;
		    argc--, argv++;
		    goto register_again;
		}
		else
		    if (strcmp (*argv, "-remove") == 0) {
			addP = 0;
			argc--, argv++;
			goto register_again;
		    }
		    else
			if (**argv == '-') {
			    (void) sprintf (buffer, "unknown option %s",
					    *argv);
			    goto tcl_error;
			}

		if (strcmp (*argv, "inform") == 0)
		    np = &p -> p_inform;
		else
		    if (strcmp (*argv, "trap") == 0)
			np = &p -> p_trap;
		    else {
			(void) sprintf (buffer, "unknown notification type %s",
					*argv);
			goto tcl_error;
		    }
		argc--, argv++;

		if (addP) {
		    for (; *np; np = &((*np) -> n_next))
			if (strcmp ((*np) -> n_name, *argv) == 0)
			    break;
		    if (!*np) {
			if (!(n = (struct nHandle *) calloc (1, sizeof *n)))
			    goto no_mem;
			if (!(n -> n_name = strdup (*argv))) {
			    free ((char *) n);
			    goto no_mem;
			}
			*np = n;
		    }
		}
		else {
		    for (; *np; np = &((*np) -> n_next))
			if (strcmp ((*np) -> n_name, *argv) == 0)
			    break;
		    if (n = *np) {
			*np = n -> n_next;
			free (n -> n_name);
			free ((char *) n);
		    }
		}

		Tcl_ResetResult (interp);
		return TCL_OK;
	    }
#ifdef	MAGIC
	    if (strcmp (cp, "resp") == 0) {
		opcode = type_SNMP_PDUs_response;
		opmask = PARTY_PRIV_GET;
		break;
	    }
#endif
	    goto none;

	case 's':
	    if (strcmp (cp, "set") == 0) {
		opcode = type_SNMP_PDUs_set__request;
		opmask = PARTY_PRIV_SET;
		break;
	    }
	    goto none;

#ifdef	MAGIC
	case 't':
	    if (strcmp (cp, "trap") == 0) {
		opcode = type_SNMP_PDUs_snmpV2__trap;
		opmask = PARTY_PRIV_GET;
		break;
	    }
	    goto none;
#endif

	case 'w':
	    if (strcmp (cp, "wait") == 0) {
		u_long	invokeID;

		argc--, argv++;

		if (argc < 1)
		    invokeID = 0;
		else
		    if (argc > 1)
			goto extra_args;
		    else
			if (sscanf (*argv, "%U", &invokeID) != 1) {
			    (void) strcpy (buffer,
					   "wait needs an invocation-ID");
			    goto tcl_error;
			}

		while (p -> p_invokes) {
		    if (invokeID) {
			for (q = p -> p_invokes; q; q = q -> i_next)
			    if (q -> i_id == invokeID)
				break;
			if (!q)
			    break;
		    }

		    (void) Tk_DoOneEvent (0);
		}
		
		Tcl_ResetResult (interp);
		return TCL_OK;
	    }
	    goto none;

#ifdef	MAGIC
	case 'x':
	    if (strcmp (cp, "xxx") == 0) {
		opcode = type_SNMP_PDUs_xxx;
		opmask = PARTY_PRIV_GET;
		break;
	    }
	    goto none;
#endif

	default: ;
none: ;
	    (void) sprintf (buffer, "expecting an operation, not %s", cp);
	    goto tcl_error;
    }
    argc--, argv++;

    if (!(q = (struct iHandle *) calloc (1, sizeof *q))) {
no_mem: ;
	(void) strcpy (buffer, "out of memory");
	goto tcl_error;
    }
    q -> i_interp = interp;
    q -> i_pHandle = p;
    if (p -> p_invokes)
	q -> i_next = p -> p_invokes;
    p -> p_invokes = q;

    if (++lastRequest & 0x80000000)
	lastRequest = 1;
    q -> i_id = lastRequest;
    if (!(q -> i_callback = strdup (*argv))
	    || !(q -> i_pname = strdup (pname)))
	goto no_mem;
    argc--, argv++;

    if (Tcl_SplitList (interp, *argv, &vecp, &vec) != TCL_OK)
	goto tcl_error_aux;
    argc--, argv++;

    bcopy ((char *) &p -> p_options, (char *) &q -> i_options,
	   sizeof q -> i_options);
    q -> i_timeout = 0;
    if (make_oHandle (interp, &q -> i_options, argc, argv) == NOTOK)
	goto tcl_error_aux;
    timeout = q -> i_timeout;

    if (p -> p_acl)
	q -> i_acl = p -> p_acl;
    else
	if (p -> p_context -> sc_identity) {
	    if (!(q -> i_acl = get_sa (p -> p_context, opmask,
				       q -> i_secqos))) {
		(void) strcpy (buffer,
			       "no access control entry allowing operation on management context");
		goto tcl_error;
	    }
	}
	else {
	    int	    len;
	    u_char *base;
	    register struct snmp_context *sc;

	    base = p -> p_context -> sc_name,
	    len = p -> p_context -> sc_name_len;
	    for (sc = contextHead; sc; sc = sc -> sc_next)
		if (sc -> sc_identity
		        && sc -> sc_name
		        && sc -> sc_name_len >= len
		        && lexnequ ((char *) sc -> sc_name, (char *) base,
				    len) == 0
		        && (q -> i_acl = get_sa (sc, opmask, q -> i_secqos)))
		    break;
	    if (!q -> i_acl) {
		(void) strcpy (buffer,
			       "nothing matching wildcard context for operation");
		goto tcl_error;
	    }
	}

    agentParty = q -> i_acl -> sa_target;
    if (agentParty -> party_timeout == 0) {
	agentParty -> party_timeout = p -> p_timeout;
	agentParty -> party_timemin = 0xffffffff;
	agentParty -> party_timemax = 0;
    }
    if (timeout)
	q -> i_timeout = timeout;
    else {
	int	i;

	q -> i_timeout = agentParty -> party_timeout;
	if (opcode == type_SNMP_PDUs_set__request)
	    for (i = q -> i_retries; i > 0; i--)
		q -> i_timeout += rto (agentParty -> party_timeout, i,
				       q -> i_rtrqos);
    }
    if (opcode == type_SNMP_PDUs_set__request)
	q -> i_retries = 0;
    if (!(q -> i_attempts = (struct request *)
				calloc ((unsigned) (q -> i_retries + 1),
					sizeof *q -> i_attempts)))
	goto no_mem;
    q -> i_current = q -> i_attempts;
#ifdef	MAGIC
    magicommand = q -> i_magic;
#endif

    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	struct type_SNMP_Snmp1157Msg *msg;

	switch (opcode) {
	    case type_SNMP_PDUs_get__bulk__request:
	        opcode = type_SNMP_PDUs_get__next__request;
		break;

	    case type_SNMP_PDUs_inform__request:
		(void) strcpy (buffer, "context held by SNMPv1 agent");
		goto tcl_error;

	    default:
		break;
	}

	if (!(msg = (struct type_SNMP_Snmp1157Msg *) calloc (1, sizeof *msg)))
	    goto no_mem;
	q -> i_msg = (struct type_SNMP_SnmpAuthMsg *) msg;
	msg -> version = int_SNMP_version_version__1;
	if (!(msg -> community =
			    str2qb ((char *) agentParty -> party_AuthPrivate,
				    agentParty -> party_AuthPrivate_len, 1))
	        || !(pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)))
	    goto no_mem;
	msg -> data = pdu;
	pdu -> offset = opcode;
	if (!(parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)))
	    goto no_mem;
#ifdef	MAGIC
	if (magicommand == MAGIC_NONZEROES)
	    parm -> error__status = -1;
	else
	    if (magicommand == MAGIC_NONZEROEI)
		parm -> error__index = 1;
#endif
    }
    else {
	struct type_SNMP_AuthInformation *authinfo;
	struct type_SNMP_SnmpMgmtCom *mgmtmsg;

	if (!(q -> i_msg = (struct type_SNMP_SnmpAuthMsg *)
					calloc (1, sizeof *q -> i_msg))
	        || !(authinfo = (struct type_SNMP_AuthInformation *)
					calloc (1, sizeof *authinfo)))
	    goto no_mem;
	q -> i_msg -> authInfo = authinfo;
	if (!(mgmtmsg = (struct type_SNMP_SnmpMgmtCom *) 
					calloc (1, sizeof *mgmtmsg)))
	    goto no_mem;
	q -> i_msg -> authData = mgmtmsg;
	if (!(pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)))
	    goto no_mem;
	mgmtmsg -> pdu = pdu;
	if (!(parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)))
	    goto no_mem;
	if ((pdu -> offset = opcode) == type_SNMP_PDUs_get__bulk__request) {
	    parm -> error__status = q -> i_nonrptr;
	    parm -> error__index = q -> i_maxrept;
	}
    }
    pdu -> un.get__request = parm;

    vp = &parm -> variable__bindings;
    for (vb = vec; *vb; vb++) {
	register struct type_SNMP_VarBindList *bind;
	register struct type_SNMP_VarBind *v;
	OID	oid;

	if (Tcl_SplitList (interp, *vb, &argc, &argv) != TCL_OK)
	    goto tcl_error_aux;

	if (argc < 1) {
	    (void) sprintf (buffer, "missing OID in variable binding");
	    goto tcl_error_aux2;
	}

	if (!(bind = (struct type_SNMP_VarBindList *) calloc (1,
							      sizeof *bind))) {
no_mem_aux: ;
	    free ((char *) argv);
	    goto no_mem;
	}
	*vp = bind, vp = &bind -> next;

	if (!(v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)))
	    goto no_mem_aux;
	bind -> VarBind = v;

	if (!(oid = str2oid (*argv)) || oid -> oid_nelem < 2) {
	    (void) sprintf (buffer, "bad OID %s in variable binding", *argv);
tcl_error_aux2: ;
	    free ((char *) argv);
	    goto tcl_error;
	}
	if (!(v -> name = oid_cpy (oid)))
	    goto no_mem_aux;

	if (argc == 1) {
	    switch (opcode) {
		case type_SNMP_PDUs_set__request:
		case type_SNMP_PDUs_inform__request:
		    (void) sprintf (buffer, "missing value for %s", *argv);
		    goto tcl_error_aux2;

		default:
		    break;
	    }
	    (v -> value = unSpecified) -> pe_refcnt++;
	}
	else
	    if (argc != 3) {
		(void) sprintf (buffer, "missing syntax and value for %s",
				*argv);
		goto tcl_error_aux2;
	    }
	    else {
		int	result;
		OS	os;
		caddr_t	value;

		if (!(os = text2syn (argv[1]))) {
		    (void) sprintf (buffer, "unknown syntax %s", argv[1]);
		    goto tcl_error_aux2;
		}
		if ((*os -> os_parse) (&value, argv[2], os) == NOTOK) {
		    (void) sprintf (buffer, "invalid value for %s", *argv);
		    goto tcl_error_aux2;
		}
#ifdef	MAGIC
		if ((magicommand == MAGIC_CONSTRING)
		         && (!strcmp (argv[1], "OctetString")
			         || !strcmp (argv[1], "DisplayString"))) {
		    int	    len;
		    register struct qbuf *qb,
					 *pb;
		    
		    if ((len = (qb = (struct qbuf *) value) -> qb_len) > 1
			    && (pb = str2qb (NULLCP, len >> 1, 0))) {
			bcopy (qb -> qb_forw -> qb_data
			           + qb -> qb_len - pb -> qb_len,
			       pb -> qb_data, pb -> qb_len);
			qb -> qb_forw -> qb_len -= pb -> qb_len;
			qb -> qb_len -= pb -> qb_len;
			insque (pb, qb -> qb_forw);
		    }
		}
#endif
		result = (*os -> os_encode) (value, &v -> value);
		(*os -> os_free) (value);
		if (result == NOTOK) {
		    (void) sprintf (buffer, "encoding error for %s", *argv);
		    goto tcl_error_aux2;
		}
	    }

	free ((char *) argv);
    }

    free ((char *) vec), vec = NULL;

    if (send_message (q, 0) == NOTOK) {
	q = NULL;
	goto tcl_error_aux;
    }

    (void) sprintf (buffer, "%U", q -> i_id);
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
#ifdef	MAGIC
    magicommand = MAGIC_NONE;
#endif
    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
tcl_error_aux: ;
    if (vec)
	free ((char *) vec);
    if (q)
	free_iHandle (q, NULLCP, 1);
#ifdef	MAGIC
    magicommand = MAGIC_NONE;
#endif
    return TCL_ERROR;
}

/*  */

static void  SNMPdeleteCommand (clientData)
ClientData	clientData;
{
    register struct pHandle  *p = (struct pHandle *) clientData;
    register struct pHandle **pp;
    register struct iHandle  *q;
    register struct nHandle  *n1,
			     *n2;

    for (pp = &pHead; *pp != p; pp = &((*pp) -> p_next))
	continue;
    *pp = p -> p_next;

    for (n1 = p -> p_inform; n1; n1 = n2) {
	n2 = n1 -> n_next;
	free (n1 -> n_name);
	free ((char *) n1);
    }
    for (n1 = p -> p_trap; n1; n1 = n2) {
	n2 = n1 -> n_next;
	free (n1 -> n_name);
	free ((char *) n1);
    }

    while (q = p -> p_invokes) {
	Tcl_ResetResult (q -> i_interp);
	free_iHandle (q, "destroyed", 1);
    }

    if (p -> p_resources)
	free (p -> p_resources);

    if (p -> p_name)
	free (p -> p_name);

    free ((char *) p);
}

/*  */

static char *agentprop = "\
proc	snmptcl_agentcap	{oid object property} {\n\
    global	snmptcl_library\n\
    global	[set array snmptcl_agentcap_[join [split $oid \".\"] \"_\"]]\n\
\n\
    if {![info exists $array]} {\n\
	if {[file readable $oid]} {\n\
	    source $oid\n\
	} elseif {[file readable $snmptcl_library/agents/$oid]} {\n\
	    source $snmptcl_library/agents/$oid\n\
	} else {\n\
	    error \"no information available\"\n\
	}\n\
    }\n\
\n\
    unset	array\n\
    upvar #0	snmptcl_agentcap_[join [split $oid \".\"] \"_\"] array\n\
\n\
    if {[catch { set value $array($object,$property) }]} {\n\
	if {$property == \"support\"} {\n\
	    set		value	0\n\
	} else {\n\
	    set		value	\"\"\n\
	}\n\
    }\n\
\n\
    return	$value\n\
}\n";


/* ARGSUSED */

static int  SNMPinfoCommand (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    char   *cp,
	    buffer[BUFSIZ];

    if (argc < 2) {
	(void) sprintf (buffer, "argc=%d; expecting at least 2", argc);
	goto tcl_error;
    }
    argc--, argv++;

    switch (*(cp = *argv)) {
	case 'a':
	    if (strcmp (cp, "agentprop") == 0) {
		int	result;
		OID	oid;
		static int once_only = 0;

		argc--, argv++;

		if (argc != 3) {
		    (void) sprintf (buffer, "argc=%d; expecting 3", argc);
		    goto tcl_error;
		}

		if (!once_only) {
		    once_only = 1;
		    if (Tcl_GlobalEval (interp, agentprop) != TCL_OK)
			return TCL_ERROR;
		}

		if (!(oid = text2oid (argv[0]))) {
		    (void) sprintf (buffer, "unable to resolve %s", argv[0]);
		    goto tcl_error;
		}
		cp = strdup (sprintoid (oid));
		oid_free (oid);
		if (!cp) {
		    (void) strcpy (buffer, "out of memory");
		    goto tcl_error;
		}

		if (!(oid = text2oid (argv[1]))) {
		    (void) sprintf (buffer, "unable to resolve %s", argv[1]);
		    goto tcl_error;
		}

		result = Tcl_VarEval (interp, "snmptcl_agentcap ", cp, " ",
				      sprintoid (oid), " ", argv[2], NULLCP);

		oid_free (oid);

		return result;
	    }
	    goto none;

	case 'c':
	    if (strcmp (cp, "ctxlist") == 0) {
		register struct snmp_context *sc;
		Tcl_DString	dstring;

		if (argc > 1)
		    goto extra_args;

		Tcl_DStringInit (&dstring);
		for (sc = contextHead; sc; sc = sc -> sc_next)
		    if (sc -> sc_location == AGENT) {
			Tcl_DStringStartSublist (&dstring);
			cp = u_char2str (sc -> sc_name, sc -> sc_name_len);
			if (*cp == '"') {
			    cp++;
			    cp [strlen (cp) - 1] = NULL;
			}
			(void) Tcl_DStringAppendElement (&dstring, cp);
			if (sc -> sc_identity)
			    (void) Tcl_DStringAppendElement
				    (&dstring, sprintoid (sc -> sc_identity));
			else
			    (void) Tcl_DStringAppendElement (&dstring, "");
			(void) Tcl_DStringAppendElement (&dstring,
							 sc -> sc_status
							         == ROW_ACTIVE
							     ? "1" : "0");
			Tcl_DStringEndSublist (&dstring);
		    }
		Tcl_DStringResult (interp, &dstring);
		Tcl_DStringFree (&dstring);

		return TCL_OK;
	    }
	    goto none;
	    
	case 'm':
	    if (strcmp (cp, "mibprop") == 0) {
		OI	oi;
		OT	ot;

		argc--, argv++;

		if (argc < 2) {
		    (void) sprintf (buffer,
				    "need an object name and a property name");
		    goto tcl_error;
		}

		if (!(oi = text2inst (*argv))) {
		    (void) sprintf (buffer, "no such object %s", *argv);
		    goto tcl_error;
		}
		ot = oi -> oi_type;
		argc--, argv++;

		Tcl_ResetResult (interp);
		switch (*(cp = *argv)) {
		    case 'A':
		        if (strcmp (cp, "ACCESS") == 0) {
			    if (argc > 1) {
extra_args: ;
				(void) strcpy (buffer, "extra arguments");
				goto tcl_error;
			    }

			    if (!ot -> ot_status)
				return TCL_OK;
			    switch (ot -> ot_access) {
				case OT_NONE:
				    Tcl_SetResult (interp, "not-accessible",
						   TCL_STATIC);
				    break;

				case OT_RDONLY:
				    Tcl_SetResult (interp, "read-only",
						   TCL_STATIC);
				    break;

				case OT_WRONLY:
				    Tcl_SetResult (interp, "write-only",
						   TCL_STATIC);
				    break;

				case OT_RDWRITE:
				    Tcl_SetResult (interp, "read-write",
						   TCL_STATIC);
				    break;

				default:
				    break;
			    }
			    return TCL_OK;
			}
		        if (strcmp (cp, "AUGMENTS") == 0) {
			    if (argc > 1)
				goto extra_args;

			    if (ot -> ot_augments)
				Tcl_SetResult (interp, ot -> ot_augments,
					       TCL_STATIC);
			    return TCL_OK;
			}
			goto null;

		    case 'd':
		        if (strcmp (cp, "descriptor") == 0) {
			    if (argc > 1)
				goto extra_args;

			    Tcl_SetResult (interp, ot -> ot_text, TCL_STATIC);
			    return TCL_OK;
			}
			goto null;

		    case 'e':
		        if (strcmp (cp, "enums") == 0) {
			    register struct enum_syntax *es;
			    OS	    os;
			    Tcl_DString	dstring;

			    if (argc > 1)
				goto extra_args;

			    if (!(os = ot -> ot_syntax)
				    || !(es = os -> os_enum))
				return TCL_OK;

			    Tcl_DStringInit (&dstring);
			    for (; es; es = es -> es_next) {
				Tcl_DStringStartSublist (&dstring);
				(void) sprintf (buffer, "%d", es -> es_value);
				(void) Tcl_DStringAppendElement (&dstring,
								 buffer);
				(void) Tcl_DStringAppendElement (&dstring,
								 es ->es_name);
				Tcl_DStringEndSublist (&dstring);
			    }
			    Tcl_DStringResult (interp, &dstring);
			    Tcl_DStringFree (&dstring);

			    return TCL_OK;
			}
			goto null;

		    case 'f':
		        if (strcmp (cp, "full") == 0) {
			    struct OIDentifier oids;

			    if (argc > 1)
				goto extra_args;

			    if (!ot -> ot_index
				    && !ot -> ot_augments
				    && ot -> ot_iit
				    && !((OT) ot -> ot_iit) -> ot_iit
				    && resolve_iit (interp, (OT) ot -> ot_iit)
					    != TCL_OK) {
				ot -> ot_iit = NULL;
				Tcl_ResetResult (interp);
			    }

			    Tcl_AppendElement (interp, ot -> ot_text);
			    Tcl_AppendElement (interp,
					       ot -> ot_status ? "OBJECT-TYPE"
							       : "");
			    Tcl_AppendElement (interp,
					       sprintoid (ot -> ot_name));
			    if (ot -> ot_status) {
				bcopy ((char *) oi -> oi_name, (char *) &oids,
				       sizeof oids);
				oids.oid_elements += ot -> ot_name ->oid_nelem;
			        oids.oid_nelem -= ot -> ot_name -> oid_nelem;
				Tcl_AppendElement (interp, sprintoid (&oids));
				if (ot -> ot_iit)
				    append_iit (interp, ot, &oids);
				else
				    Tcl_AppendElement (interp, "");
			    }
			    else {
				Tcl_AppendElement (interp, "");
				Tcl_AppendElement (interp, "");
			    }
			    return TCL_OK;
			}
			goto null;

		    case 'I':
		        if (strcmp (cp, "INDEX") == 0) {
			    register struct iinfo *iit;

			    if (argc > 1)
				goto extra_args;

			    if (!ot -> ot_index)
				return TCL_OK;

			    if (!ot -> ot_iit
				    && resolve_iit (interp, ot) != TCL_OK)
				return TCL_ERROR;

			    for (iit = (struct iinfo *) ot -> ot_iit;
				     iit -> ii_ot;
				     iit++) {
				if (iit -> ii_length)
				    Tcl_AppendElement (interp, iit -> ii_ot
								   -> ot_text);
				else {
				    (void) sprintf (buffer, "IMPLIED %s",
						    iit -> ii_ot -> ot_text);
				    Tcl_AppendElement (interp, buffer);
				}
			    }

			    return TCL_OK;
			}
			goto null;

		    case 'M':
		        if (strcmp (cp, "MAX-ACCESS") == 0) {
			    if (argc > 1)
				goto extra_args;

			    if (!ot -> ot_status)
				return TCL_OK;
			    switch (ot -> ot_access) {
				case OT_NONE:
				    Tcl_SetResult (interp, "not-accessible",
						   TCL_STATIC);
				    break;

				case OT_RDONLY:
				    Tcl_SetResult (interp, "read-only",
						   TCL_STATIC);
				    break;

				case OT_RDWRITE:
				    Tcl_SetResult (interp, "read-write",
						   TCL_STATIC);
				    break;

				case OT_RDCREAT:
				    Tcl_SetResult (interp, "read-create",
						   TCL_STATIC);
				    break;

				default:
				    break;
			    }
			    return TCL_OK;
			}
			goto null;

		    case 'r':
			if (strcmp (cp, "range") == 0) {
			    OS	    os;
			    if (ot -> ot_range
				    && (os = ot -> ot_syntax)
				    && (!strcmp (os -> os_name, "Integer32")
					    || !strcmp (os -> os_name,
							"INTEGER"))) {
do_range: ;
				(void) sprintf (buffer, "%d",
						ot -> ot_lendpoint);
				Tcl_AppendElement (interp, buffer);

				(void) sprintf (buffer, "%d",
						ot -> ot_rendpoint);
				Tcl_AppendElement (interp, buffer);
			    }
			    return TCL_OK;
			}
			goto null;

		    case 'S':
		        if (strcmp (cp, "STATUS") == 0) {
			    if (argc > 1)
				goto extra_args;

			    switch (ot -> ot_status) {
				case OT_OBSOLETE:
				    Tcl_SetResult (interp, "obsolete",
						   TCL_STATIC);
				    break;

				case OT_CURRENT:
				    Tcl_SetResult (interp, "current",
						   TCL_STATIC);
				    break;

				case OT_OPTIONAL:
				    Tcl_SetResult (interp, "optional",
						   TCL_STATIC);

				case OT_DEPRECATED:
				    Tcl_SetResult (interp, "decpreated",
						   TCL_STATIC);
				    break;

				default:
				    break;
			    }
			    return TCL_OK;
			}
		        if (strcmp (cp, "SYNTAX") == 0) {
			    OS	    os;

			    if (argc > 1)
				goto extra_args;

			    if (os = ot -> ot_syntax) {
				char   *dp = os -> os_textual
				    	         ? os -> os_textual
						 : os -> os_name;
				register struct pair *pp;

				for (pp = pairs; pp -> p_old; pp++)
				    if (strcmp (pp -> p_old, dp) == 0) {
					dp = pp -> p_new;
					break;
				    }

				Tcl_SetResult (interp, dp, TCL_STATIC);
			    }
			    return TCL_OK;
			}
			goto null;

		    case 's':
			if (strcmp (cp, "size") == 0) {
			    OS	    os;

			    if (ot -> ot_range
				    && (os = ot -> ot_syntax)
				    && (!strcmp (os -> os_name, "OctetString")
					    || !strcmp (os -> os_name,
							"DisplayString")))
				goto do_range;
			    return TCL_OK;
			}
			goto null;

		    case 't':
		        if (strcmp (cp, "type") == 0) {
			    if (argc > 1)
				goto extra_args;

			    Tcl_SetResult (interp, ot -> ot_status
					               ? "OBJECT-TYPE" : "",
					   TCL_STATIC);
			    return TCL_OK;
			}
			goto null;

		    case 'v':
		        if (strcmp (cp, "value") == 0) {
			    if (argc > 1)
				goto extra_args;

			    Tcl_SetResult (interp, sprintoid (ot -> ot_name),
					   TCL_VOLATILE);
			    return TCL_OK;
			}
			goto null;

		    default:
null: ;
			return TCL_OK;
		}
	    }
	    goto none;

	case 'p':
	    if (strcmp (cp, "proclist") == 0) {
		register struct pHandle *p;
		Tcl_DString	dstring;

		if (argc > 1)
		    goto extra_args;

		Tcl_DStringInit (&dstring);
		for (p = pHead; p; p = p -> p_next)
		    if (p != pInternal)
			(void) Tcl_DStringAppendElement (&dstring,
							 p -> p_name);
		Tcl_DStringResult (interp, &dstring);
		Tcl_DStringFree (&dstring);

		return TCL_OK;
	    }
	    goto none;

	default: ;
none: ;
	    (void) sprintf (buffer, "expecting an operation, not %s", cp);
	    goto tcl_error;
    }

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_ERROR;
}

/*  */

#ifdef	MAGIC
/* ARGSUSED */

static int  SNMPmiscCommand (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    char   *cp,
	    buffer[BUFSIZ];

    if (argc < 2) {
	(void) sprintf (buffer, "argc=%d; expecting at least 2", argc);
	goto tcl_error;
    }
    argc--, argv++;

    Tcl_ResetResult (interp);
    switch (*(cp = *argv)) {
	case 'a':
	    if (strcmp (cp, "agentprop") == 0) {
		int	newPtr;
		char   *agent,
		       *ap;
		OID	oid;
		Tcl_HashEntry *hPtr,
			      *gPtr;
		Tcl_HashSearch search;
		Tcl_HashTable *acTable;

		argc--, argv++;

		if (argc == 0) {
		    for (gPtr = Tcl_FirstHashEntry (&agentTable, &search);
			     gPtr;
			     gPtr = Tcl_NextHashEntry (&search))
			Tcl_AppendElement (interp,
					   Tcl_GetHashKey (&agentTable, gPtr));

		    return TCL_OK;
		}

		if (!(oid = text2oid (agent = *argv))) {
		    (void) sprintf (buffer, "unable to resolve %s", *argv);
		    goto tcl_error;
		}
		argc--, argv++;

		hPtr = Tcl_CreateHashEntry (&agentTable, ap = sprintoid (oid),
					    &newPtr);
		oid_free (oid);

		if (!hPtr) {
no_mem: ;
		    (void) strcpy (buffer, "out of memory");
		    goto tcl_error;
		}
		if (newPtr) {
		    if (!(acTable = (Tcl_HashTable *)
						calloc (1, sizeof *acTable))) {
			(void) strcpy (buffer, "out of memory");
			goto tcl_error;
		    }
		    Tcl_SetHashValue (hPtr, (ClientData) acTable);
		    Tcl_InitHashTable (acTable, TCL_STRING_KEYS);
		}
		else
		    acTable = (Tcl_HashTable *) Tcl_GetHashValue (hPtr);
		if (argc == 0)
		    goto tcl_ok;

		if (strcmp (*argv, "supports") == 0) {
		    argc--, argv++;

		    if (argc % 2) {
			(void) strcpy
			          (buffer,
				   "expecting an even number of support args");
			goto tcl_error;
		    }

		    for (; argc-- > 0; argv++) {
			int	vecp;
			char   *module,
			      **vec,
			      **xvec;

			module = *argv;
			argc--, argv++;

			if (Tcl_SplitList (interp, *argv, &vecp, &vec)
			        != TCL_OK)
			    goto tcl_error;
			xvec = vec;

			if (!(gPtr = Tcl_CreateHashEntry (acTable, module,
							  &newPtr)))
			    goto no_mem;
			if (!(module = strdup (*vec)))
			    goto no_mem;
			Tcl_SetHashValue (gPtr, (ClientData) module);

			for (; vecp-- > 0; vec++)
			    if (gPtr = Tcl_FindHashEntry (&groupTable, *vec)) {
				OID    *oidp = (OID *) Tcl_GetHashValue (gPtr);

				if (!oidp) {
				    (void) sprintf (buffer,
						    "lost info on group %s",
						    *vec);
				    free ((char *) xvec);
				    goto tcl_error;
				}

				for (; *oidp; oidp++)
				    if (!Tcl_CreateHashEntry
						(acTable, sprintoid (*oidp),
						 &newPtr))
					goto no_mem;
			    }
			    else {
				OID	p;
				OT	ot = text2obj (*vec);

				if (!ot) {
				    (void) sprintf (buffer,
						    "unable to resolve %s",
						    *vec);
				    free ((char *) xvec);
				    goto tcl_error;
				}

				p = ot -> ot_name;
				for (ot = ot -> ot_next;
				         ot;
				         ot = ot -> ot_next) {
				    int	    i;
				    OID	    q = ot -> ot_name;

				    if (!ot -> ot_syntax
					    && ((i = strlen (ot -> ot_text))
						        <= 5
						    || strcmp (ot -> ot_text
							           +i - 5,
							       "Entry")))
					continue;
				    if (p -> oid_nelem < q -> oid_nelem
					    && bcmp ((char *) p ->oid_elements,
						     (char *) q ->oid_elements,
						     p -> oid_nelem
						         * sizeof *p
							    -> oid_elements)
						    == 0) {
					if (ot -> ot_status != OT_OBSOLETE
					        && !Tcl_CreateHashEntry
					    		    (acTable,
							     sprintoid (q),
							     &newPtr))
					    goto no_mem;
				    }
				    else
					break;
				}
			    }

			free ((char *) xvec);
		    }

		    goto tcl_ok;
		}

		if (strcmp (*argv, "variation") == 0) {
		    OT	    ot;

		    argc--, argv++;

		    if (argc != 3) {
			(void) sprintf (buffer,
					"argc=%d; expecting 3 for variation",
					argc);
			goto tcl_error;
		    }
		    argc--, argv++;

		    if (!(ot = text2obj (*argv))) {
			(void) sprintf (buffer, "unable to resolve %s", *argv);
			goto tcl_error;
		    }
		    if (!(gPtr = Tcl_FindHashEntry
			  	    (acTable, sprintoid (ot -> ot_name)))) {
			(void) sprintf (buffer, "%s not supported by agent",
					*argv);
			goto tcl_error;
		    }
		    argc--, argv++;

		    if (!(cp = strdup (*argv)))
			goto no_mem;
		    Tcl_SetHashValue (gPtr, (ClientData) cp);

		    goto tcl_ok;
		}

		if (strcmp (*argv, "dump") == 0) {
		    char   *bp,
			   *dp;
		    FILE   *fp;

		    argc--, argv++;

		    if (argc > 0) {
			(void) sprintf (buffer, "extra arguments");
			goto tcl_error;
		    }

		    if (!(fp = fopen (ap, "w"))) {
			(void) sprintf (buffer, "unable to create %s: %s",
					ap, Tcl_PosixError (interp));
			goto tcl_error;
		    }

		    fprintf (fp, "# AGENT-CAPABILITIES for %s (%s)\n\n",
			     agent, ap);

		    if (!(bp = strdup (ap))) {
			(void) fclose (fp);
			goto no_mem;
		    }
		    for (ap = bp; *ap; ap++)
			if (*ap == '.')
			    *ap = '_';
		    fprintf (fp, "global snmptcl_agentcap_%s\n\n", bp);

		    for (gPtr = Tcl_FirstHashEntry (acTable, &search);
			     gPtr;
			     gPtr = Tcl_NextHashEntry (&search)) {
			int	impl,
				vecp;
			char  **vec,
			      **xvec;

			cp = Tcl_GetHashKey (acTable, gPtr);
			if (!isdigit (*cp)
			        && (dp = (char *) Tcl_GetHashValue (gPtr))) {
			    fprintf (fp, "\n\
if {[catch { snmpinfo mibprop %s value }]} {\n\
    if {[file readable %s.defs]} {\n\
	snmpmisc mibload %s.defs\n\
    } else {\n\
	catch { snmpmisc mibload %s%s.defs }\n\
    }\n\
}\n\n", dp, cp, cp, DEFSDIR, cp);
			    continue;
			}

			if (!(dp = (char *) Tcl_GetHashValue (gPtr))) {
			    OT	    ot;

			    if (!(oid = str2oid (cp))) {
				(void) sprintf (buffer, "unable to convert %s",
						cp);
				goto losing_patience;
			    }
			    if (!(ot = name2obj (oid))) {
				(void) sprintf (buffer, "unable to resolve %s",
						cp);
				goto losing_patience;
			    }
			    if (ot -> ot_syntax)
				fprintf (fp,
					 "set snmptcl_agentcap_%s(%s,support) 1\n",
					 bp, cp);
			    continue;
			}

			if (Tcl_SplitList (interp, dp, &vecp, &vec) != TCL_OK){
losing_patience: ;
			    (void) fclose (fp);
			    goto tcl_error;
			}
			xvec = vec;

			impl = 1;
			for (; vecp-- > 0; vec++) {
			    int	    argp;
			    char  **args;

			    if (Tcl_SplitList (interp, *vec, &argp, &args)
				    != TCL_OK) {
really_losing_patience: ;
				free ((char *) xvec);
				goto losing_patience;
			    }
			    if (argp != 2) {
				(void) sprintf (buffer,
						"malformed variation %s",
						*vec);
really_really_losing_patience: ;
				free ((char *) args);
				goto really_losing_patience;
			    }

			    if (strcmp (args[0], "CREATION-REQUIRES") == 0) {
				int	zargp;
				char   *ep,
				      **zargs,
				      **zvec;

				if (Tcl_SplitList (interp, args[1], &zargp,
						   &zargs) != TCL_OK)
				    goto really_really_losing_patience;
				zvec = zargs;

				fprintf (fp, "set snmptcl_agentcap_%s(%s,%s) {",
					 bp, cp, args[0]);

				ep = "";
				for (; zargp-- > 0; zargs++) {
				    OT	    ot;

				    if (!(ot = text2obj (*zargs))) {
					(void) sprintf (buffer,
							"unable to resolve %s",
							*zargs);
					free ((char *) zvec);
					goto really_really_losing_patience;
				    }

				    fprintf (fp, "%s%s", ep,
					     sprintoid (ot -> ot_name));
				    ep = " ";
				}

				fprintf (fp, "}\n");

				free ((char *) zvec);

				impl = 0;
			    }
			    else
				fprintf (fp,
					 "set snmptcl_agentcap_%s(%s,%s) {%s}\n",
					 bp, cp, args[0], args[1]);

			    if (strcmp (args[0], "ACCESS") == 0
				    && strcmp (args[1], "not-implemented") ==0)
				impl = 0;

			    free ((char *) args);
			}

			free ((char *) xvec);

			if (impl)
			    fprintf (fp,
				     "set snmptcl_agentcap_%s(%s,support) 1\n",
				     bp, cp);
		    }

		    (void) fclose (fp);

		    goto tcl_ok;
		}

		cp = *argv;
		goto none;
	    }
	    goto none;

	case 'm':
	    if (strcmp (cp, "mibgroup") == 0) {
		int	newPtr;
		char   *group;
		register OID *op;
		OID    *oidp;
		Tcl_HashEntry *hPtr;

		argc--, argv++;

		if (argc < 2) {
		    (void) strcpy (buffer, "need a module and a group name");
		    goto tcl_error;
		}
		argc--, argv++;

		group = *argv;
		argc--, argv++;

		if (!(hPtr = Tcl_CreateHashEntry (&groupTable, group,
						  &newPtr)))
		    goto no_mem;
		if (!newPtr) {
		    (void) sprintf (buffer, "group %s already exists", group);
		    goto tcl_error;
		}

		if (!(oidp = (OID *) calloc ((unsigned) (argc + 1),
					     sizeof *oidp))) {
		    Tcl_DeleteHashEntry (hPtr);
		    goto no_mem;
		}
		Tcl_SetHashValue (hPtr, (ClientData) oidp);
		for (op = oidp; argc-- > 0; argv++, op++)
		    if (!(*op = text2oid (*argv))) {
			(void) sprintf (buffer, "unable to resolve %s", *argv);
			for (op--; op >= oidp; op--)
			    oid_free (*op);
			free ((char *) oidp);
			Tcl_DeleteHashEntry (hPtr);
			goto tcl_error;
		    }

		goto tcl_ok;
	    }
	    if (strcmp (cp, "mibload") == 0) {
		int	i;
		OT	ot;

		argc--, argv++;

		if (argc < 1) {
		    (void) strcpy (buffer, "need a file name");
		    goto tcl_error;
		}

		if (debug)
		    fprintf (stderr, "loading object definitions from %s\n",
			     *argv);
		if (loadobjects (*argv) == NOTOK) {
		    (void) sprintf (buffer, "loadobjects: %s", PY_pepy);
		    goto tcl_error;
		}

		for (ot = anchor; ot; ot = ot -> ot_next)
		    if (!ot -> ot_syntax
			    && (ot -> ot_index || ot -> ot_augments)
			    && (i = strlen (ot -> ot_text)) > 5
			    && !strcmp (ot -> ot_text + i - 5, "Entry")) {
			OT	ot2,
				otn;

			otn = ot;
			for (ot2 = ot -> ot_children;
			         ot2;
			         ot2 = ot2 -> ot_sibling) {
			    otn = ot2;
			    ot2 -> ot_iit = (caddr_t) ot;
			}

			ot = otn;
		    }

		goto tcl_ok;
	    }
	    goto none;

	default: ;
none: ;
	    (void) sprintf (buffer, "expecting an operation, not %s", cp);
	    goto tcl_error;
    }

tcl_ok: ;
    Tcl_ResetResult (interp);
    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_ERROR;
}
#endif

/*  */

/* ARGSUSED */

static int  SNMPoid2ode (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    char   *cp,
	    buffer[BUFSIZ];
    OID	    a;

    if (argc != 2) {
	(void) sprintf (buffer, "argc=%d; expecting 2", argc);
	goto tcl_error;
    }
	
    if (!(a = str2oid (cp = argv[1]))) {
	(void) sprintf (buffer, "bad OID %s", cp);
	goto tcl_error;
    }
    Tcl_SetResult (interp, oid2ode (a), TCL_VOLATILE);
    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_ERROR;
}

/*    SNMP */

static int  send_message (q, callback)
struct iHandle *q;
int	callback;
{
    struct timeval tvs;
    register struct request *rq = q -> i_current;
    struct snmp_party *agentParty = q -> i_acl -> sa_target;
    struct dHandle *d;

#ifdef	MAGIC
    magicommand = q -> i_magic;
#endif

    if (!(d = find_dHandle (q -> i_interp, agentParty)))
	goto losing;

#ifdef	MAGIC
    if (magicommand == MAGIC_BIGREQID)
	rq -> r_id = 0x7fffffff;
    else
#endif
    rq -> r_id = lastRequest;
    Tcl_ResetResult (q -> i_interp);
    if (agentParty -> party_tDomain == RFC1157DOMAIN) {
	struct type_SNMP_Snmp1157Msg *msg =
				(struct type_SNMP_Snmp1157Msg *) q -> i_msg;

	msg -> data -> un.get__request -> request__id = rq -> r_id;
	if (send_v1_message (q -> i_interp, d, msg) == NOTOK)  {
losing: ;
	    free_iHandle (q, callback ? "failed" : NULLCP, 1);
#ifdef	MAGIC
	    magicommand = MAGIC_NONE;
#endif
	    return NOTOK;
	}
    }
    else {
	q -> i_msg -> authData -> pdu -> un.get__request -> request__id =
							rq -> r_id;
	if (send_v2_message (q -> i_interp, d, q -> i_msg,
			     q -> i_acl -> sa_subject, agentParty,
			     q -> i_acl -> sa_resources) == NOTOK)
	    goto losing;
    }

    (void) gettimeofday (&tvs, (struct timezone *) 0);
    rq -> r_time = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L;
    rq -> r_token = Tk_CreateTimerHandler ((int) q -> i_timeout, lost_message,
					   (ClientData) rq -> r_id);
    rq -> r_waiting = 1;
#ifdef	MAGIC
    magicommand = MAGIC_NONE;
#endif

    return OK;
}

/*  */

static int  send_v1_message (interp, d, msg)
Tcl_Interp     *interp;
struct dHandle *d;
register struct type_SNMP_Snmp1157Msg *msg;
{
    int	    result = NOTOK;
    char    buffer[BUFSIZ];
    PE	    pe = NULL;
    PS	    ps = d -> d_ps;
#ifdef	MAGIC
    struct type_SNMP_PDU *old;
    struct type_SNMP_Trap__PDU trap;
    struct type_SNMP_TimeTicks stamp;
#endif

#ifdef	MAGIC
    switch (magicommand) {
	case MAGIC_BDSTPARTY:
	case MAGIC_NDSTMATCH:
	case MAGIC_BSRCPARTY:
	case MAGIC_NSRCMATCH:
	case MAGIC_BCONTEXT:
	    (void) sprintf (buffer, "V2 magic requested for V1 transaction");
	    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
	    goto out;

	default:
	    break;
    }
#endif

#ifdef	MAGIC
    if (msg -> data -> offset == type_SNMP_PDUs_snmpV2__trap) {
	old = msg -> data -> un.get__request;

	msg -> data -> offset = type_SNMP_PDUs_trap;
	bzero ((char *) &trap, sizeof trap);
	msg -> data -> un.trap = &trap;
	trap.enterprise = nullSpecific;
	if ((trap.agent__addr = str2qb ("\0\0\0\0", 4, 1)) == NULL)
	    goto losing_trap;
	trap.generic__trap = int_SNMP_generic__trap_coldStart;
	trap.specific__trap = 0;
	trap.time__stamp = &stamp;
	stamp.parm = 0;
	trap.variable__bindings = old -> variable__bindings;
    }
#endif

    if (encode_SNMP_Snmp1157Msg (&pe, 1, 0, NULLCP, msg) != NOTOK) {
#ifdef	MAGIC
	if (magicommand == MAGIC_NONUNITVR)
	    *pe -> pe_cons -> pe_prim = 1;
	else
	    if (magicommand == MAGIC_NULLCOMM)
		pe -> pe_cons -> pe_next -> pe_len = 0;
#endif

	if (*watch)
	    (void) print_SNMP_Snmp1157Msg (pe, 1, NULLIP, NULLVP, NULLCP);
	d -> d_errno = 0;
	if (pe2ps (ps, pe) == NOTOK) {
	    if (d -> d_errno)
		(void) strcpy (buffer, Tcl_PosixError (interp));
	    else
		(void) sprintf (buffer, "pe2ps loses {%s}",
				ps_error (ps -> ps_errno));
tcl_result: ;
	    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
	    goto out;
	}
	else
	    result = OK;
    }
    else {
#ifdef	MAGIC
losing_trap: ;
#endif
	(void) sprintf (buffer, "encode_SNMP_Snmp1157Msg loses {%s}", PY_pepy);
	goto tcl_result;
    }

out: ;
    if (pe)
	pe_free (pe);

#ifdef	MAGIC
    if (msg -> data -> offset == type_SNMP_PDUs_trap) {
	msg -> data -> offset = type_SNMP_PDUs_snmpV2__trap;
	msg -> data -> un.get__request = old;
	if (trap.agent__addr)
	    qb_free (trap.agent__addr);
    }
#endif

    return result;
}

/*  */

static int  send_v2_message (interp, d, msg, srcParty, dstParty, ourContext)
Tcl_Interp     *interp;
struct dHandle *d;
struct type_SNMP_SnmpAuthMsg *msg;
struct snmp_party *srcParty,
		  *dstParty;
struct snmp_context *ourContext;
{
    int	    result = NOTOK;
    char    buffer[BUFSIZ];
    struct type_SNMP_SnmpPrivMsg *privmsg = NULL;
    struct type_SNMP_SnmpMgmtCom *mgmtmsg = msg -> authData;
    PE	    pe = NULL;
    PS	    ps = d -> d_ps;

#ifdef	MAGIC
    switch (magicommand) {
	case MAGIC_NONUNITVR:
	case MAGIC_NULLCOMM:
	    (void) sprintf (buffer, "V1 magic requested for V2 transaction");
	    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
	    goto out;

	default:
	    break;
    }
#endif

    if (mgmtmsg -> srcParty)
	oid_free (mgmtmsg -> srcParty);
#ifdef	MAGIC
    if (magicommand == MAGIC_BSRCPARTY) {
	if (!(mgmtmsg -> srcParty = oid_cpy (nullSpecific)))
	    goto no_mem;
    }
    else
#endif
    if (!(mgmtmsg -> srcParty = oid_cpy (srcParty -> party_Identity))) {
no_mem: ;
	Tcl_SetResult (interp, "out of memory", TCL_STATIC);
	goto out;
    }
    if (mgmtmsg -> dstParty)
	oid_free (mgmtmsg -> dstParty);
#ifdef	MAGIC
    if (magicommand == MAGIC_NDSTMATCH) {
	if (!(mgmtmsg -> dstParty = oid_cpy (nullSpecific)))
	    goto no_mem;
    }
    else
#endif
    if (!(mgmtmsg -> dstParty = oid_cpy (dstParty -> party_Identity)))
	goto no_mem;
    if (mgmtmsg -> context)
	oid_free (mgmtmsg -> context);
#ifdef	MAGIC
    if (magicommand == MAGIC_BCONTEXT) {
	if (!(mgmtmsg -> context = oid_cpy (nullSpecific)))
	    goto no_mem;
    }
    else
#endif
    if (!(mgmtmsg -> context = oid_cpy (ourContext -> sc_identity)))
	goto no_mem;

    if (!(privmsg = (struct type_SNMP_SnmpPrivMsg *) calloc (1,
							     sizeof *privmsg)))
	goto no_mem;
#ifdef	MAGIC
    if (magicommand == MAGIC_BDSTPARTY) {
	if (!(privmsg -> privDst = oid_cpy (nullSpecific)))
	    goto no_mem;
    }
    else
#endif
    if (!(privmsg -> privDst = oid_cpy (dstParty -> party_Identity)))
	goto no_mem;
#ifdef	MAGIC
    if (magicommand == MAGIC_NSRCMATCH) {
	struct snmp_party *src,
			  *dst;
	struct snmp_access *sa;

	for (src = partyHead; src; src = src -> party_next)
	    if (src -> party_location == srcParty -> party_location
		    && src -> party_authProt > BADAUTH
		    && src -> party_authProt != srcParty -> party_authProt)
		break;
	if (!src) {
	    (void) sprintf (buffer,
			    "no source parties with alternative credentials");
	    goto tcl_result;
	}
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE
		    && sa -> sa_subject == src
		    && (dst = sa -> sa_target) -> party_tDomain
			    == dstParty -> party_tDomain
		    && dst -> party_TAddress_len
			    == dstParty -> party_TAddress_len
		    && bcmp ((char *) dst -> party_TAddress,
			     (char *) dstParty -> party_TAddress,
			     dst -> party_TAddress_len) == 0)
		break;
	if (!sa) {
	    (void) sprintf (buffer,
			    "no access control entries for alternative credentials");
	    goto tcl_result;
	}
	if (!(privmsg -> privData =
	          generate_outgoing_auth (msg, src, dst,
					  (caddr_t) (*watch ? -1 : 0))))
	    goto bad_auth;
    }
    else
#endif
    if (!(privmsg -> privData =
	     generate_outgoing_auth (msg, srcParty, dstParty,
				     (caddr_t) (*watch ? -1 : 0)))) {
#ifdef	MAGIC
bad_auth: ;
#endif
	(void) sprintf (buffer,
			"error generating outgoing authentication {%s}",
			PY_pepy);
tcl_result: ;
	Tcl_SetResult (interp, buffer, TCL_VOLATILE);
	goto out;
    }

    if (do_encryption (privmsg, dstParty) == NOTOK) {
	(void) sprintf (buffer, "error encrypting {%s}", PY_pepy);
	goto tcl_result;
    }

    if (encode_SNMP_SnmpPrivMsg (&pe, 1, 0, NULLCP, privmsg) != NOTOK) {
	if (*watch)
	    (void) print_SNMP_SnmpPrivMsg (pe, 1, NULLIP, NULLVP, NULLCP);

	d -> d_errno = 0;
	if (pe2ps (ps, pe) == NOTOK) {
	    if (d -> d_errno)
		(void) strcpy (buffer, Tcl_PosixError (interp));
	    else
		(void) sprintf (buffer, "pe2ps loses {%s}",
				ps_error (ps -> ps_errno));
	    goto tcl_result;
	}
	else
	    result = OK;
    }
    else {
	(void) sprintf (buffer, "encode_SNMP_SnmpPrivMsg loses {%s}", PY_pepy);
	goto tcl_result;
    }

out: ;
    if (pe)
	pe_free (pe);

    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);

    return result;
}

/*  */

/* ARGSUSED */

static void  recv_message (clientData, mask)
ClientData	clientData;
int		mask;
{
    u_long  now;
    char    buffer[BUFSIZ];
    struct type_SNMP_SnmpV1orV2 *msg = NULL;
    struct type_SNMP_SnmpAuthMsg *authmsg = NULL;
    struct type_SNMP_PDUs *pdu;
    register struct type_SNMP_PDU *parm;
    register struct type_SNMP_VarBindList *vp;
    struct snmp_party *dstParty,
		      *srcParty;
    struct dHandle *d = (struct dHandle *) clientData;
    struct qbuf *qb = NULL;
    struct timeval tvs;
    PE	    pe;
    PS	    ps = d -> d_ps;
    PS	    qbufPS = NULL;
    register struct pHandle *p;
    register struct iHandle *q;
    Tcl_DString     dstring,
		    ystring,
		    zstring;

    (void) gettimeofday (&tvs, (struct timezone *) 0);
    now = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L;

    Tcl_DStringInit (&dstring);
    Tcl_DStringInit (&ystring);
    Tcl_DStringInit (&zstring);

    if (!(pe = ps2pe (ps))) {
bad_parse: ;
	if (debug || *watch)
	    fprintf (stderr, "ps2pe loses: %s\n", ps_error (ps -> ps_errno));
	if (debug && d -> d_pktsize >= 0) {
	    register char *dp,
			  *ep;

	    fprintf (stderr, "read %d octets\n", d -> d_pktsize);
	    for (ep = (dp = d -> d_packet) + d -> d_pktsize; dp < ep; ) {
		char    *cp;
		register char *fp;

		if ((fp = dp + 16) > ep)
		    fp = ep;
		for (cp = ""; dp < fp; dp++, cp = " ")
		    fprintf (stderr, "%s%02x", cp, *dp & 0xff);
		fprintf (stderr, "\n");
	    }
	}
	goto out;
    }

    if (decode_SNMP_SnmpV1orV2 (pe, 1, NULLIP, NULLVP, &msg) == NOTOK) {
	if (debug || *watch)
	    fprintf (stderr, "decode_SNMP_SnmpV1orV2 loses: %s\n", PY_pepy);
bad_syntax: ;
	if (debug) {
	    PS	    pl;

	    if (!(pl = ps_alloc (std_open)))
		fprintf (stderr, "ps_alloc: loses\n");
	    else {
		if (std_setup (pl, stdout) == NOTOK)
		    fprintf (stderr, "std_setup: %s\n",
			     ps_error (pl -> ps_errno));
		else
		    if (pe2pl (pl, pe) == NOTOK)
			fprintf (stderr, "pe2pl: %s\n",
				 ps_error (pl -> ps_errno));
		ps_free (pl);
	    }
	}
	goto out;
    }

    if (*watch)
	(void) print_SNMP_SnmpV1orV2 (pe, 1, NULLIP, NULLVP, NULLCP);

    if (msg -> offset == type_SNMP_SnmpV1orV2_snmpV1)
	pdu = msg -> un.snmpV1 -> data;
    else {
	struct type_SNMP_SnmpPrivMsg *privmsg = msg -> un.snmpV2;

	if (!(dstParty = lookup_party (privmsg -> privDst, MANAGER))) {
	    if (debug || *watch)
		fprintf (stderr, "badParty: destination=%s\n",
			 oid2ode (privmsg -> privDst));
	    goto out;
	}
	if (!(qb = do_decryption (privmsg, dstParty))) {
	    if (debug || *watch)
		fprintf (stderr, "error decrypting: %s\n", PY_pepy);
	    goto out;
	}
	if (!(qbufPS = ps_alloc (qbuf_open))
	        || qbuf_setup (qbufPS, qb) == NOTOK) {
	    if (qbufPS) {
		if (debug || *watch)
		    fprintf (stderr, "qbuf_setup: %s\n",
			     ps_error (ps -> ps_errno));
		ps_free (qbufPS);
	    }
	    else {
		if (debug || *watch)
		    fprintf (stderr, "ps_alloc(qbuf_open) loses\n");
	    }
	    goto out;
	}

	pe_free (pe);
	if (!(pe = ps2pe (qbufPS))) {
	    if (debug)
		d -> d_pktsize = -1;
	    goto bad_parse;
	}
	(void) do_decryption_fixup (privmsg, qb);

	if (decode_SNMP_SnmpAuthMsg (pe, 1, NULLIP, NULLVP, &authmsg)
	        == NOTOK) {
	    if (debug || *watch)
		(void) fprintf (stderr, "decode_SNMP_SnmpAuthMsg loses: %s\n",
				PY_pepy);
	    goto bad_syntax;
	}

	if (*watch)
	    (void) print_SNMP_SnmpAuthMsg (pe, 1, NULLIP, NULLVP, NULLCP);

	if (!(srcParty = lookup_party (authmsg -> authData -> srcParty,
				       AGENT))) {
	    if (debug || *watch)
		fprintf (stderr, "badParty: source=%s\n",
			 oid2ode (authmsg -> authData -> srcParty));
	    goto out;
	}

	if (authenticate_incoming (privmsg, authmsg, srcParty, dstParty,
				   NULLIP) == NOTOK) {
	    if (debug || *watch)
		fprintf (stderr, "authentication failure: %s\n", PY_pepy);
	    goto out;
	}

	ps_free (qbufPS), qbufPS = NULL;
	free_SNMP_SnmpV1orV2 (msg), msg = NULL;

	pdu = authmsg -> authData -> pdu;
    }

    switch (pdu -> offset) {
	case type_SNMP_PDUs_inform__request:
	case type_SNMP_PDUs_snmpV2__trap:
	    {
		int	gotone = 0;

		for (p = pHead; p; p = p -> p_next) {
		    register struct snmp_context *sc = p -> p_context;
		    
		    if (sc -> sc_identity) {
			if (oid_cmp (sc -> sc_identity,
				     authmsg -> authData -> context))
			    continue;

try_selected: ;
			if (p -> p_selected =
			            (pdu -> offset
				             == type_SNMP_PDUs_inform__request)
					? (p -> p_inform ? 1 : 0)
			                : (p -> p_trap ? 1 : 0))
				gotone = 1;
		    }
		    else {
			int	len;
			u_char *base;
			register struct snmp_context *sc2;

			base = sc -> sc_name, len = sc -> sc_name_len;
			for (sc2 = contextHead; sc2; sc2 = sc2 -> sc_next)
			    if (sc2 -> sc_identity
				    && sc2 -> sc_name
				    && sc2 -> sc_name_len >= len
				    && lexnequ ((char *) sc2 -> sc_name,
						(char *) base, len) == 0)
				goto try_selected;
		    }
		}

		if (!gotone) {
		    if (debug || *watch)
			fprintf (stderr, "unexpected %s...\n",
				 pdu -> offset
				         == type_SNMP_PDUs_inform__request
				     ? "inform" : "snmpV2-trap");
		    goto out;
		}
	    }
	    break;

	case type_SNMP_PDUs_trap:
	    {
		int	gotone = 0;
		register struct snmp_access *sa;
		register struct snmp_context *sc;
		register struct snmp_party *sp;

		for (sa = accessHead; sa; sa = sa -> sa_next)
		    if (sa -> sa_status == ROW_ACTIVE
			    && (sp = sa -> sa_subject)
			    && sp -> party_location == AGENT
			    && sp -> party_tDomain == RFC1157DOMAIN
			    && sp -> party_AuthPrivate_len ==
					msg -> un.snmpV1 -> community -> qb_len
			    && !bcmp ((char *) sp -> party_AuthPrivate,
				      msg -> un.snmpV1 -> community -> qb_data,
				      sp -> party_AuthPrivate_len)
			    && (sc = sa -> sa_resources))
			for (p = pHead; p; p = p -> p_next)
			    if (p -> p_context == sc && p -> p_trap)
				gotone = p -> p_selected = 1;

		if (!gotone) {
		    if (debug || *watch)
			fprintf (stderr, "unexpected snmpV1-trap...\n");
		    goto out;
		}
	    }
	    break;

	case type_SNMP_PDUs_response:
	    {
		int	rid = (parm = pdu -> un.response) -> request__id;
		u_long  curtt;
		register struct iHandle **qp;
		register struct request *rq;
		struct snmp_party *agentParty;

		for (p = pHead; p; p = p -> p_next) {
		    for (q = p -> p_invokes; q; q = q -> i_next) {
			for (rq = q -> i_attempts; rq <= q -> i_current; rq++)
			    if (rq -> r_id == rid)
				break;
			if (rq <= q -> i_current)
			    break;
		    }
		    if (q)
			break;
		}
		if (!p) {
		    if (debug || *watch)
			fprintf (stderr, "orphaned response %d...\n", rid);
		    goto out;
		}

		Tk_DeleteTimerHandler (rq -> r_token);
		rq -> r_waiting = 0;
		agentParty = q -> i_acl -> sa_target;
		curtt = (now > rq -> r_time) ? now - rq -> r_time : 1;
		if ((debug || *watch)
		        && q -> i_current != q -> i_attempts) {
		    fprintf (stderr, "retry #%d returned in %U, ",
			     rq - q -> i_attempts, curtt);
		    fprintf (stderr, "min=%U est=%U max=%U next=%U\n",
			     agentParty -> party_timemin,
			     agentParty -> party_timeout,
			     agentParty -> party_timemax,
			     (curtt > agentParty -> party_timeout)
			         ? (curtt * 3) >> 1
			     	 : (agentParty -> party_timemax + curtt) >> 1);
		}
		if (agentParty -> party_timemin > curtt)
		    agentParty -> party_timemin = curtt;
		if (agentParty -> party_timemax < curtt)
		    agentParty -> party_timemax = curtt;
		if (curtt > agentParty -> party_timeout)
		    agentParty -> party_timeout = (curtt * 3) >> 1;
		else
		    agentParty -> party_timeout =
			(agentParty -> party_timemax + curtt) >> 1;

		for (qp = &q -> i_pHandle -> p_invokes;
		         *qp != q;
		         qp = &((*qp) -> i_next))
		    continue;
		*qp = q -> i_next;

		if (strcmp (q -> i_callback, GET_CLOCKS) == 0) {
		    get_clocks_callback (parm -> error__status,
					 parm -> variable__bindings,
					 q -> i_acl, q -> i_zero,
					 q -> i_update, q -> i_sync);
		    free_iHandle (q, NULLCP, 0);
		    goto out;
		}
	    }

	    (void) Tcl_DStringAppendElement (&dstring, q -> i_callback);
	    (void) Tcl_DStringAppendElement (&dstring, q -> i_pname);
	    (void) sprintf (buffer, "%U", q -> i_id);
	    (void) Tcl_DStringAppendElement (&dstring, buffer);
	    (void) Tcl_DStringAppendElement (&dstring,
					     snmp_error (parm->error__status));
	    (void) sprintf (buffer, "%d", (int) parm -> error__index);
	    (void) Tcl_DStringAppendElement (&dstring, buffer);
	    break;

	default:
	    if (debug || *watch)
		fprintf (stderr, "unexpected message type %d\n",
			 pdu -> offset);
	    goto out;
    }

    Tcl_DStringStartSublist (&dstring);
    if (pdu -> offset == type_SNMP_PDUs_trap) {
	struct type_SNMP_Trap__PDU *trap = pdu -> un.trap;

	Tcl_DStringStartSublist (&dstring);	/* sysUpTime.0 */
	(void) Tcl_DStringAppendElement (&dstring, "1.3.6.1.2.1.1.3.0");
	(void) Tcl_DStringAppendElement (&dstring, "TimeTicks");
	(void) sprintf (buffer, "%d", (int) trap -> time__stamp -> parm);
	(void) Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);	/* snmpTrapOID.0 */
	(void) Tcl_DStringAppendElement (&dstring, "1.3.6.1.6.3.1.1.4.1.0");
	(void) Tcl_DStringAppendElement (&dstring, "ObjectID");
	if (trap -> generic__trap
	        != int_SNMP_generic__trap_enterpriseSpecific) {
	    (void) sprintf (buffer, "1.3.6.1.6.3.1.1.5.%d", /* snmpTraps.n+1 */
			    ((int) trap -> generic__trap) + 1);
	    (void) Tcl_DStringAppendElement (&dstring, buffer);
	}
	else {
	    (void) Tcl_DStringAppendElement (&dstring,
					     sprintoid (trap -> enterprise));
	    (void) sprintf (buffer, ".0.%d", (int) trap -> specific__trap);
	    (void) Tcl_DStringAppend (&dstring, buffer, -1);
	}
	Tcl_DStringEndSublist (&dstring);
    }
    for (vp = parm -> variable__bindings; vp; vp = vp -> next) {
	char	buf2[BUFSIZ],
	       *vec[4];
	struct type_SNMP_VarBind *v = vp -> VarBind;
	PE	vv = v -> value;

	Tcl_DStringAppend (&ystring, sprintoid (v -> name), -1);
	vec[0] = Tcl_DStringValue (&ystring);

	vec[2] = buf2;

	if (vv -> pe_form != PE_FORM_PRIM) {
	    vec[1] = "error";
	    (void) strcpy (buf2,
			   snmpV2_exception (int_SNMP_error__status_wrongEncoding));
	}
	else if (vv -> pe_class == PE_CLASS_CONT) {
	    vec[1] = "exception";
	    (void) strcpy (buf2, snmpV2_exception (vv -> pe_id));
	}
	else
	switch (PE_ID (vv -> pe_class, vv -> pe_id)) {
	    case PE_ID (PE_CLASS_UNIV, PE_PRIM_INT):
		vec[1] = "Integer32";
		{
		    integer	i = prim2num (vv);

		    if (i == NOTOK && vv -> pe_errno != PE_ERR_NONE) {
do_error: ;
			(void) strcpy (buf2, pe_error (vv -> pe_errno));
do_error_aux: ;
			vec[1] = "error";
		    }
		    else
			(void) sprintf (buf2, "%d", i);
		}
		break;


	    case PE_ID (PE_CLASS_UNIV, PE_PRIM_BITS):
		vec[1] = "BitString";
		if (!prim2bit (vv))
		    goto do_error;
		else {
		    register int    i;

		    for (i = 0; i < vv -> pe_nbits; i++)
			(void) Tcl_DStringAppend (&zstring,
						  bit_test (vv, i) == 1 ? "1"
									: "0",
						  1);

		    vec[2] = Tcl_DStringValue (&zstring);
		}
		break;

	    case PE_ID (PE_CLASS_UNIV, PE_PRIM_OCTS):
		{
		    int	    notnvt;
		    register char  *dp,
				   *ep;
		    register struct qbuf *pb;
		    struct qbuf *rb;

		    if (!(rb = prim2qb (vv)))
			goto do_error;

		    notnvt = 0;
		    for (pb = rb -> qb_forw; pb != rb; pb = pb -> qb_forw)
			for (ep = (dp = pb -> qb_data) + pb -> qb_len;
			         dp < ep;
			         dp++)
			    if (!isspace (*dp) && !isprint (*dp)) {
				notnvt = 1;
				break;
			    }

		    if (notnvt) {
			char   *cp = "0x";

			vec[1] = "OctetString";

			for (pb = rb -> qb_forw;
			         pb != rb;
			         pb = pb -> qb_forw) {
			    for (ep = (dp = pb -> qb_data) + pb -> qb_len;
			             dp < ep;
			             dp++, cp = ":") {
				(void) sprintf (buffer, "%s%02x", cp,
						*dp & 0xff);
				(void) Tcl_DStringAppend (&zstring, buffer,-1);
			    }
			}
		    }
		    else {
			vec[1] = "DisplayString";

			for (pb = rb -> qb_forw;
			         pb != rb;
			         pb = pb -> qb_forw)
			    (void) Tcl_DStringAppend (&zstring,
						      (char *) pb -> qb_data,
						      pb -> qb_len);
		    }

		    qb_free (rb);

		    vec[2] = Tcl_DStringValue (&zstring);
		}
		break;

#ifdef	notdef
	    case PE_ID (PE_CLASS_UNIV, PE_PRIM_NULL):
		vec[1] = "NULL";
		buf2[0] = NULL;
		break;
#endif

	    case PE_ID (PE_CLASS_UNIV, PE_PRIM_OID):
		vec[1] = "ObjectID";
		{
		    OID	    oid;

		    if (!(oid = prim2oid (vv)))
			goto do_error;
		    else
			vec[2] = sprintoid (oid);
		}
		break;

	    case PE_ID (PE_CLASS_APPL, 0):
		vec[1] = "IpAddress";
		{
		    struct type_SNMP_IpAddress *ip;
		    register struct qbuf *ib;

		    if (!(ip = prim2qb (vv)))
			goto do_error;
		    if (qb_pullup (ip) == NOTOK) {
			(void) strcpy (buf2, "out of memory");
			free_SNMP_IpAddress (ip);
			goto do_error_aux;
		    }
		    else
			if ((ib = ip -> qb_forw) -> qb_len != 4) {
			    (void) sprintf (buf2, "wrong length %d",
					    ib -> qb_len);
			    free_SNMP_IpAddress (ip);
			    goto do_error_aux;
			}
		        else {
			    (void) sprintf (buf2, "%d.%d.%d.%d",
					    ib -> qb_data[0] & 0xff,
					    ib -> qb_data[1] & 0xff,
					    ib -> qb_data[2] & 0xff,
					    ib -> qb_data[3] & 0xff);
			}
		    free_SNMP_IpAddress (ip);
		}
		break;

	    case PE_ID (PE_CLASS_APPL, 1):
		vec[1] = "Counter32";
		goto do_unsigned;
	    case PE_ID (PE_CLASS_APPL, 2):
		vec[1] = "Gauge32";
		goto do_unsigned;
	    case PE_ID (PE_CLASS_APPL, 3):
		vec[1] = "TimeTicks";
		goto do_unsigned;
	    case PE_ID (PE_CLASS_APPL, 6):
		vec[1] = "Counter64";
		goto do_unsigned;
	    case PE_ID (PE_CLASS_APPL, 7):
		vec[1] = "UInteger32";
		/* and fall... */
do_unsigned: ;
		{
		    register int    i;
		    register char  *dp;
		    double	dd;
		    counter64	c64;

		    if (_prim2ulong (vv, &c64, (vv -> pe_id == 6 ? 2 : 1)
						    * sizeof (u_long))
			    == NOTOK)
			goto do_error;
		    else
			if (c64.hiword == 0)
			    (void) sprintf (buf2,
					    (c64.loword & 0x80000000)
					    	? "%U.0" : "%U",
					    c64.loword);
			else {
			    dd = c64.hiword;
			    for (i = sizeof c64.loword * 8; i > 0; i--)
				dd *= 2.0;

			    if (c64.hiword <= 0x1fffff) {
				dd += c64.loword;
				(void) sprintf (buf2, "%.f", dd);
			    }
			    else {
				int	carry;
				register char  *ep;
				char	last5[6];

				dd += c64.loword & 0xfffff000;
				(void) sprintf (buf2, "%.f", dd);
				(void) sprintf (last5, "%5.5d",
						c64.loword & 0xfff);
				for (i = carry = 0,
				             dp = buf2 + strlen (buf2) - 1,
				     	     ep = last5 + 4;
				         i < 5;
				         i++, dp--, ep--) {
				    int	    j = carry + (*dp - '0')
						      + (*ep - '0');

				    if (j > 9)
					j -= 10, carry = 1;
				    else
					carry = 0;
				    *dp = j + '0';
				}
			    }
			    for (dp = buf2; *dp; dp++)
				if (*dp == '.' || !isdigit (*dp))
				    break;
			    if (!*dp)
				(void) strcpy (dp, ".0");
			}
		}
		break;

	    case PE_ID (PE_CLASS_APPL, 4):
		vec[1] = "Opaque";
		buf2[0] = NULL;
		break;

	    case PE_ID (PE_CLASS_APPL, 5):
		vec[1] = "NsapAddress";
		{
		    int	    len;
		    struct type_SNMP_NsapAddress *np;
		    register struct qbuf *nb;

		    if (!(np = prim2qb (vv)))
			goto do_error;
		    if (qb_pullup (np) == NOTOK) {
			(void) strcpy (buf2, "out of memory");
		        free_SNMP_NsapAddress (np);
			goto do_error_aux;
		    }
		    else
			if ((len = (nb = np -> qb_forw) -> qb_len) < 1
			        || len == 2
			        || len == 3
			        || len > 21) {
			    (void) sprintf (buf2, "wrong length %d", len);
		            free_SNMP_NsapAddress (np);
			    goto do_error_aux;
			}
		        else {
			    register char *bp,
					  *dp,
					  *ep;
			    char   *cp = "0x";

			    if ((len = nb -> qb_data[0] & 0xff) >=nb -> qb_len)
				len = nb -> qb_len - 1;
			    bp = buf2;
			    for (ep = (dp = nb -> qb_data + 1) + len;
				     dp < ep;
				     dp++, cp = ":") {
				(void) sprintf (bp, "%s%02x", cp, *dp & 0xff);
				bp += strlen (bp);
			    }
			}
		    free_SNMP_NsapAddress (np);
		}
		break;

	    default:
		vec[1] = "unknown";
		(void) sprintf (buf2, "%s %u",
				pe_classlist[((int) vv -> pe_class) & 03],
				(unsigned int) vv -> pe_id);
		break;
	}

	Tcl_DStringStartSublist (&dstring);
	(void) Tcl_DStringAppendElement (&dstring, vec[0]);
	(void) Tcl_DStringAppendElement (&dstring, vec[1]);
	(void) Tcl_DStringAppendElement (&dstring, vec[2]);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringTrunc (&ystring, 0);

	if (vec[2] != buf2) {
	    Tcl_DStringTrunc (&zstring, 0);
	    vec[2] = buf2;
	}
    }
    if (pdu -> offset == type_SNMP_PDUs_trap) {
	Tcl_DStringStartSublist (&dstring);	/* snmpTrapEnterprise.0 */
	(void) Tcl_DStringAppendElement (&dstring, "1.3.6.1.6.3.1.1.4.3.0");
	(void) Tcl_DStringAppendElement (&dstring, "ObjectID");
	(void) Tcl_DStringAppendElement (&dstring,
					 sprintoid (pdu -> un.trap
							    -> enterprise));
	Tcl_DStringEndSublist (&dstring);
    }
    Tcl_DStringEndSublist (&dstring);

    if (pdu -> offset == type_SNMP_PDUs_response) {
/*
	struct timeval tvs2;

	(void) gettimeofday (&tvs2, (struct timezone *) 0);
 */
	if (Tcl_Eval (q -> i_interp, Tcl_DStringValue (&dstring)) != TCL_OK)
	    Tk_BackgroundError (q -> i_interp);
/*
	(void) gettimeofday (&tvs, (struct timezone *) 0);
	if (tvs2.tv_usec>tvs.tv_usec)
	tvs.tv_sec--, tvs.tv_usec += 1000000;
	fprintf (stderr, "%2U.%06U %d\n", tvs.tv_sec - tvs2.tv_sec,
		 tvs.tv_usec - tvs2.tv_usec, Tcl_DStringLength (&dstring));
 */
	free_iHandle (q, NULLCP, 0);
    }
    else {
	for (p = pHead; p; p = p -> p_next)
	    if (p -> p_selected) {
		register struct nHandle *n;

		p -> p_selected = 0;

		n = pdu -> offset == type_SNMP_PDUs_inform__request
						? p -> p_inform : p -> p_trap;
		for (; n; n = n -> n_next) {
		    Tcl_DString    nstring;

		    Tcl_DStringInit (&nstring);

		    (void) Tcl_DStringAppendElement (&nstring, n -> n_name);
		    (void) Tcl_DStringAppendElement (&nstring, p -> p_name);
		    (void) sprintf (buffer, "%U", q -> i_id);
		    (void) Tcl_DStringAppendElement (&nstring, buffer);
		    (void) Tcl_DStringAppendElement (&nstring,
						     snmp_error (parm -> error__status));
		    (void) sprintf (buffer, "%d", (int) parm -> error__index);
		    (void) Tcl_DStringAppendElement (&nstring, buffer);
		    (void) Tcl_DStringAppendElement (&nstring,
						     Tcl_DStringValue (&dstring));

		    if (Tcl_Eval (p -> p_interp, Tcl_DStringValue (&nstring))
			    != TCL_OK)
			Tk_BackgroundError (p -> p_interp);

		    Tcl_DStringFree (&nstring);
		}
	    }

	if (pdu -> offset == type_SNMP_PDUs_inform__request) {
	    bcopy ((char *) &d -> d_lastrecv, (char *) &d -> d_agent,
		   sizeof d -> d_agent);

	    pdu -> offset = type_SNMP_PDUs_response;
	    (void) send_v2_message (p -> p_interp, d, authmsg, dstParty,
				    srcParty, p -> p_context);
	}
    }

out: ;
    Tcl_DStringFree (&zstring);
    Tcl_DStringFree (&ystring);
    Tcl_DStringFree (&dstring);
    if (pe)
	pe_free (pe);
    if (qbufPS)
	ps_free (qbufPS);
    if (qb)
	qb_free (qb);
    if (msg)
	free_SNMP_SnmpV1orV2 (msg);
    if (authmsg)
	free_SNMP_SnmpAuthMsg (authmsg);
}

/*  */

/* ARGSUSED */

static void  lost_message (clientData)
ClientData	clientData;
{
    int	    rid = (int) clientData;
    register struct pHandle *p;
    register struct iHandle *q;
    register struct snmp_party *r;

    if (debug || *watch)
	fprintf (stderr, "timeout for %d...\n", rid);
    for (p = pHead; p; p = p -> p_next) {
	for (q = p -> p_invokes; q; q = q -> i_next)
	    if (q -> i_current -> r_id == rid)
		break;
	if (q)
	    break;
    }
    if (p) {
	if (debug || *watch) {
	    r = q -> i_acl -> sa_target;
	    fprintf (stderr, "min=%U est=%U(%U retry #%d) max=%U\n",
		     r -> party_timemin,
		     q -> i_timeout, r -> party_timeout,
		     q -> i_current - q -> i_attempts, r -> party_timemax);
	}
    }
    else {
	if (debug || *watch)
	    fprintf (stderr, "spurious timeout...\n");
	return;
    }

    q -> i_current -> r_waiting = 0;
    if (q -> i_retries-- > 0) {
	q -> i_current++;

	if (++lastRequest & 0x80000000)
	    lastRequest = 1;
	q -> i_timeout = rto (q -> i_acl -> sa_target -> party_timeout,
			      q -> i_current - q -> i_attempts, q -> i_rtrqos);
	(void) send_message (q, 1);
	return;
    }

    if ((r = q -> i_acl -> sa_subject) -> party_authProt == MD5AUTHPROT) {
/* We might want to check for clock-skew.  However, this is an unlikely
   event, because:

    	- we periodically execute a clock synchronization algorithm; and,

	- subsequent traffic warps the clock appropriately.

   Hence, if the clock is skewed it means either:

    	- someone changed the secret and reset the clock; or,

	- the agent rebooted and advanced the clock a large amount.

   So, we schedule a clock sync, but return a timeout to the user...
*/

	if (r -> party_clocksync) {
	    if (!r -> party_new_storage) {
		if (debug || *watch)
		    fprintf (stderr, "clock sync already in progress\n");
	    }
	    else {
		if (debug || *watch)
		    fprintf (stderr,
			     "performing clock sync using knowledge\n");

		r -> party_new_storage = 0;
		get_clocks (r -> party_clocksync, 0, 1, r, NULLCP);
	    }
	}
	else {
	    if (debug || *watch)
		fprintf (stderr, "no knowledge, so scheduling clock sync\n");

	    start_clocks ((ClientData) 1);
	}
    }

    if (debug || *watch)
	fprintf (stderr, "timeout: no response within %d attempt%s\n",
		 (q -> i_current - q -> i_attempts) + 1,
		 q -> i_current != q -> i_attempts ? "s" : "");

    Tcl_ResetResult (q -> i_interp);
    free_iHandle (q, "timeout", 1);
}

/*  */

static char *errors[] = {
    "noError",
    "tooBig",
    "noSuchName",
    "badValue",
    "readOnly",
    "genErr",
    "noAccess",
    "wrongType",
    "wrongLength",
    "wrongEncoding",
    "wrongValue",
    "noCreation",
    "inconsistentValue",
    "resourceUnavailable",
    "commitFailed",
    "undoFailed",
    "authorizationError",
    "notWritable",
    "inconsistentName"
};

static char  *snmp_error (i)
integer	i;
{
    static char buffer[BUFSIZ];

    if (0 <= i && i < sizeof errors / sizeof errors[0])
	return errors[i];
    (void) sprintf (buffer, "%d", i);

    return buffer;
}


static char *exceptions[] = {
    "noSuchObject",
    "noSuchInstance",
    "endOfMibView"
};


static char  *snmpV2_exception (i)
PElementID	i;
{
    static char buffer[BUFSIZ];

    if (i < sizeof exceptions / sizeof exceptions[0])
	return exceptions[i];
    (void) sprintf (buffer, "%u", (unsigned int) i);

    return buffer;
}

/*    INVOCATIONS */

static void  free_iHandle (q, callmsg, remove)
struct iHandle *q;
char   *callmsg;
int	remove;
{
    if (remove) {
	register struct iHandle **qp;

	for (qp = &q -> i_pHandle -> p_invokes;
	         *qp != q;
	         qp = &((*qp) -> i_next))
	    continue;
	*qp = q -> i_next;
    }

    if (callmsg) {
	char    buffer[BUFSIZ];

	if (strcmp (q -> i_callback, GET_CLOCKS) == 0) {
	    register struct snmp_party *p;

	    if (debug || *watch)
		fprintf (stderr, "%s timeout\n",
			 q -> i_sync ? "clock sync" : "get_clocks");

	    wantsync = 1;
	    if (p = q -> i_sync) {
		if (!p -> party_new_storage) {
		    if (debug || *watch)
			fprintf (stderr, "invalidating knowledge for %s\n",
				 oid2ode (p -> party_Identity));

		    p -> party_clocksync = NULL;
		}
	    }
	    else
		find_clocks (q -> i_zero);
	}
	else {
	    Tcl_DString	dstring;

	    Tcl_DStringInit (&dstring);
	    (void) Tcl_DStringAppendElement (&dstring, q -> i_callback);
	    (void) Tcl_DStringAppendElement (&dstring, q -> i_pname);
	    (void) sprintf (buffer, "%U", q -> i_id);
	    (void) Tcl_DStringAppendElement (&dstring, buffer);
	    (void) Tcl_DStringAppendElement (&dstring, callmsg);
	    (void) Tcl_DStringAppendElement (&dstring, "0");
	    (void) Tcl_DStringAppendElement (&dstring, q -> i_interp -> result
							    ? q -> i_interp
								-> result :"");

	    if (Tcl_Eval (q -> i_interp, Tcl_DStringValue (&dstring))
		    != TCL_OK)
		Tk_BackgroundError (q -> i_interp);

	    Tcl_DStringFree (&dstring);
	}
    }
    if (q -> i_callback)
	free (q -> i_callback);
    if (q -> i_pname)
	free (q -> i_pname);
    
    if (q -> i_attempts) {
	register struct request *rq = q -> i_current;

	if (rq -> r_waiting)
	    Tk_DeleteTimerHandler (rq -> r_token);

	free ((char *) q -> i_attempts);
    }

    if (q -> i_msg) {
	if (q -> i_acl -> sa_target -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)q -> i_msg);
	else
	    free_SNMP_SnmpAuthMsg (q -> i_msg);
    }

    free ((char *) q);
}

/*  */

static u_long	rto (timeout, retry, rtrqos)
u_long	timeout;
int	retry,
	rtrqos;
{
    u_long	curtt;

    switch (rtrqos) {
	case RTR_EXPONENTIAL:
	    curtt = timeout << retry;
	    break;

	case RTR_SUPERLINEAR:
	default:
	    curtt = timeout;
	    while (retry-- > 0)
		curtt += timeout + ((retry * timeout) >> 1);
	    break;
    }

    return curtt;
}

/*    OPTIONS */

struct optpair {
    char   *o_string;
    int	    o_value;
};

static struct optpair rtrqos[] = {	/* array position is IMPORTANT... */
    "exponential", RTR_EXPONENTIAL,
    "superlinear", RTR_SUPERLINEAR,

    NULL
};

static struct optpair secqos[] = {
    "none",	   SEC_NONE,
    "auth",	   SEC_AUTH,
    "authpriv",	   SEC_AUTH | SEC_PRIV,

    NULL
};

#ifdef	MAGIC
static struct optpair magics[] = {
    "none",      MAGIC_NONE,
    "notminlen", MAGIC_NOTMINLEN,
    "constring", MAGIC_CONSTRING,
    "indflen",   MAGIC_INDFLEN,
    "nonunitvr", MAGIC_NONUNITVR,
    "nullcomm",  MAGIC_NULLCOMM,
    "bigreqid",  MAGIC_BIGREQID,
    "nonzeroes", MAGIC_NONZEROES,
    "nonzeroei", MAGIC_NONZEROEI,
    "bdstparty", MAGIC_BDSTPARTY,
    "ndstmatch", MAGIC_NDSTMATCH,
    "bsrcparty", MAGIC_BSRCPARTY,
    "nsrcmatch", MAGIC_NSRCMATCH,
    "bcontext",	 MAGIC_BCONTEXT,
    "berlesudp", MAGIC_BERLESUDP,
    "udplesber", MAGIC_UDPLESBER,

    NULL
};
#endif

/*  */

static int  make_oHandle (interp, o, argc, argv)
Tcl_Interp     *interp;
struct oHandle *o;
int		argc;
char	      **argv;
{
    char    buffer[BUFSIZ];
    register struct optpair *m;

    for (; argc > 0; argc--) {
	register char *cp = *argv++;

	if (*cp != '-') {
	    (void) sprintf (buffer, "options should start with -, not %s", cp);
	    goto losing;
	}

	switch (*++cp) {
	    case 'm':
#ifdef	MAGIC
	        if (strcmp (cp, "magic") == 0) {
		    if (--argc < 1)
			goto need_string;
		    for (m = magics; m -> o_string; m++)
			if (strcmp (*argv, m -> o_string) == 0) {
			    o -> o_magic = m -> o_value;
			    break;
			}
		    if (!m -> o_string) {
			(void) sprintf (buffer, "unknown magic %s", *argv);
			goto losing;
		    }
		    argv++;
		    break;
		}
#endif
	        if (strcmp (cp, "maxrept") == 0) {
		    if (--argc < 1) {
need_integer: ;
			(void) sprintf (buffer, "option %s needs an integer",
					cp);
			goto losing;
		    }
		    if (Tcl_GetInt (interp, *argv++, &o -> o_maxrept)
			    != TCL_OK)
			return TCL_ERROR;
		    if (o -> o_maxrept < 1)
			o -> o_maxrept = 1;
		    continue;
		}
		goto none;

	    case 'n':
	        if (strcmp (cp, "nonrptr") == 0) {
		    if (--argc < 1)
			goto need_integer;
		    if (Tcl_GetInt (interp, *argv++, &o -> o_nonrptr)
			    != TCL_OK)
			return TCL_ERROR;
		    if (o -> o_nonrptr < 0)
			o -> o_nonrptr = 0;
		    continue;
		}
		goto none;

	    case 'r':
	        if (strcmp (cp, "retries") == 0) {
		    if (--argc < 1)
			goto need_integer;
		    if (Tcl_GetInt (interp, *argv++, &o -> o_retries)
			    != TCL_OK)
			return TCL_ERROR;
		    if (o -> o_retries < 0)
			o -> o_retries = 0;
		    continue;
		}
		if (strcmp (cp, "rtrqos") == 0) {
		    if (--argc < 1) {
need_string: ;
			(void) sprintf (buffer, "option %s needs a string",
					cp);
			goto losing;
		    }
		    for (m = rtrqos; m -> o_string; m++)
			if (strcmp (*argv, m -> o_string) == 0) {
			    o -> o_rtrqos = m -> o_value;
			    break;
			}
		    if (!m -> o_string) {
			(void) sprintf (buffer, "unknown rtrqos %s", *argv);
			goto losing;
		    }
		    argv++;
		    break;
		}
		goto none;

	    case 's':
		if (strcmp (cp, "secqos") == 0) {
		    if (--argc < 1)
			goto need_string;
		    for (m = secqos; m -> o_string; m++)
			if (strcmp (*argv, m -> o_string) == 0) {
			    o -> o_secqos = m -> o_value;
			    break;
			}
		    if (!m -> o_string) {
			(void) sprintf (buffer, "unknown secqos %s", *argv);
			goto losing;
		    }
		    argv++;
		    break;
		}
		goto none;

	    case 't':
	        if (strcmp (cp, "timeout") == 0) {
		    int	    timeout;

		    if (--argc < 1)
			goto need_integer;
		    if (Tcl_GetInt (interp, *argv++, &timeout) != TCL_OK)
			return TCL_ERROR;
		    if (timeout < 1)
			timeout = 1;
		    o -> o_timeout = timeout * 1000L;
		    continue;
		}
		goto none;

	    default:
none: ;
		(void) sprintf (buffer, "unknown option %s", cp);
		goto losing;
	}
    }
    return OK;

losing: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return NOTOK;
}

/*  */

static int  conf_pHandle (interp, option, newvalue, p)
Tcl_Interp     *interp;
char	       *option,
	       *newvalue;
struct pHandle *p;
{
    char    buffer[BUFSIZ];
    register struct optpair *m;
    struct snmp_context *sc = p -> p_context;

    if (!option) {
	Tcl_DString	dstring;

	Tcl_DStringInit (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "active");
	(void) sprintf (buffer, "%d", sc -> sc_status == ROW_ACTIVE ? 1 : 0);
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);
	    
	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "identity");
	if (sc -> sc_identity)
	    Tcl_DStringAppendElement (&dstring, sprintoid (sc -> sc_identity));
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "maxrept");
	(void) sprintf (buffer, "%d", p -> p_maxrept);
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "name");
	if (sc -> sc_name) {
	    char   *cp = u_char2str (sc -> sc_name, sc -> sc_name_len);

	    if (*cp == '"') {
		cp++;
		cp [strlen (cp) - 1] = NULL;
	    }
	    (void) strcpy (buffer, cp);
	    Tcl_DStringAppendElement (&dstring, buffer);
	}
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "nonrptr");
	(void) sprintf (buffer, "%d", p -> p_nonrptr);
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "resources");
	if (p -> p_resources)
	    Tcl_DStringAppendElement (&dstring, p -> p_resources);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "retries");
	(void) sprintf (buffer, "%d", p -> p_retries);
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "rtrqos");
	Tcl_DStringAppendElement (&dstring, rtrqos[p -> p_rtrqos].o_string);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "secqos");
	if (p -> p_secqos & SEC_AUTH)
	    (void) strcpy (buffer,
			   p -> p_secqos & SEC_PRIV ? "authpriv" : "auth");
	else
	    (void) strcpy (buffer, "none");
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringStartSublist (&dstring);
	Tcl_DStringAppendElement (&dstring, "timeout");
	(void) sprintf (buffer, "%U", p -> p_timeout / 1000);
	Tcl_DStringAppendElement (&dstring, buffer);
	Tcl_DStringEndSublist (&dstring);

	Tcl_DStringResult (interp, &dstring);
	Tcl_DStringFree (&dstring);

	return TCL_OK;
    }

    Tcl_ResetResult (interp);
    switch (*option) {
        case 'a':
	    if (strcmp (option, "active") == 0) {
		if (newvalue) {
read_only: ;
		    (void) sprintf (buffer, "option %s is read-only", option);
		    goto tcl_error;
		}

		(void) sprintf (buffer, "%d",
				sc -> sc_status == ROW_ACTIVE ? 1 : 0);
		break;
	    }
	    goto none;

	case 'i':
	    if (strcmp (option, "identity") == 0) {
		if (newvalue)
		    goto read_only;

		if (sc -> sc_identity)
		    (void) strcpy (buffer, sprintoid (sc -> sc_identity));
		break;
	    }
	    goto none;

	case 'm':
	    if (strcmp (option, "maxrept") == 0) {
		(void) sprintf (buffer, "%d", p -> p_maxrept);

		if (newvalue) {
		    if (Tcl_GetInt (interp, newvalue, &p -> p_maxrept)
			    != TCL_OK)
			return TCL_ERROR;
		    if (p -> p_maxrept < 1)
			p -> p_maxrept = 1;
		}
		break;
	    }
	    goto none;

	case 'n':
	    if (strcmp (option, "name") == 0) {
		if (newvalue)
		    goto read_only;

		if (sc -> sc_name) {
		    char   *cp = u_char2str (sc -> sc_name, sc -> sc_name_len);

		    if (*cp == '"') {
			cp++;
			cp [strlen (cp) - 1] = NULL;
		    }
		    (void) strcpy (buffer, cp);
		}
		break;
	    }
	    if (strcmp (option, "nonrptr") == 0) {
		(void) sprintf (buffer, "%d", p -> p_nonrptr);

		if (newvalue) {
		    if (Tcl_GetInt (interp, newvalue, &p -> p_nonrptr)
			    != TCL_OK)
			return TCL_ERROR;
		    if (p -> p_nonrptr < 0)
			p -> p_nonrptr = 0;
		}
		break;
	    }
	    goto none;

	case 'r':
	    if (strcmp (option, "resources") == 0) {
		if (p -> p_resources)
		    Tcl_SetResult (interp, p -> p_resources, TCL_VOLATILE);

		if (newvalue) {
		    char   *cp = strdup (newvalue);

		    if (!cp) {
			(void) strcpy (buffer, "out of memory");
			goto tcl_error;
		    }
		    if (p -> p_resources)
			free (p -> p_resources);
		    p -> p_resources = cp;
		}
		return TCL_OK;
	    }
	    if (strcmp (option, "retries") == 0) {
		(void) sprintf (buffer, "%d", p -> p_retries);

		if (newvalue) {
		    if (Tcl_GetInt (interp, newvalue, &p -> p_retries)
			    != TCL_OK)
			return TCL_ERROR;
		    if (p -> p_retries < 0)
			p -> p_retries = 0;
		}
		break;
	    }
	    if (strcmp (option, "rtrqos") == 0) {
		(void) strcpy (buffer, rtrqos[p -> p_rtrqos].o_string);

		if (newvalue) {
		    for (m = rtrqos; m -> o_string; m++)
			if (strcmp (newvalue, m -> o_string) == 0) {
			    p -> p_rtrqos = m -> o_value;
			    break;
			}
		    if (!m -> o_string) {
			(void) sprintf (buffer, "unknown rtrqos %s", newvalue);
			goto tcl_error;
		    }
		}
		break;
	    }
	    goto none;

	case 's':
	    if (strcmp (option, "secqos") == 0) {
		if (p -> p_secqos & SEC_AUTH)
		    (void) strcpy (buffer, p -> p_secqos & SEC_PRIV
				   		? "authpriv" : "auth");
		else
		    (void) strcpy (buffer, "none");

		if (newvalue) {
		    for (m = secqos; m -> o_string; m++)
			if (strcmp (newvalue, m -> o_string) == 0) {
			    p -> p_secqos = m -> o_value;
			    break;
			}
		    if (!m -> o_string) {
			(void) sprintf (buffer, "unknown secqos %s", newvalue);
			goto tcl_error;
		    }
		}
		break;
	    }
	    goto none;

	case 't':
	    if (strcmp (option, "timeout") == 0) {
		(void) sprintf (buffer, "%U", p -> p_timeout / 1000);

		if (newvalue) {
		    int		timeout;

		    if (Tcl_GetInt (interp, newvalue, &timeout) != TCL_OK)
			return TCL_ERROR;
		    if (timeout < 1)
			timeout = 1;
		    p -> p_timeout = timeout * 1000L;
		}
		break;
	    }
	    goto none;

	default:
none: ;
	    (void) sprintf (buffer, "unknown option %s", option);
	    goto tcl_error;
    }
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_OK;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_ERROR;
}

/*    ACCESS CONTROL */

static struct snmp_access *get_sa (ourContext, mask, sq)
struct snmp_context *ourContext;    
int	mask,
	sq;
{
    int	    qos;
    register struct snmp_access *sa;
    struct snmp_party *p;

    for (qos = sq; qos <= SEC_ALL; qos <<= 1, qos |= 1)
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE
	            && sa -> sa_resources == ourContext
	            && (sa -> sa_privileges & mask)
		    && (p = sa -> sa_target)
		    && p -> party_location == AGENT
	            && p -> party_tDomain == SNMPUDPDOMAIN
	            && ((qos & SEC_PRIV) ? p -> party_privProt > NOPRIV
				         : p -> party_privProt <= NOPRIV)
	            && (p = sa -> sa_subject)
		    && p -> party_location == MANAGER
	            && ((qos & SEC_AUTH) ? p -> party_authProt > NOAUTH
				         : p -> party_authProt <= NOAUTH))
	    return sa;

    if (sq != SEC_NONE)
	return NULL;

    for (sa = accessHead; sa; sa = sa -> sa_next)
	if (sa -> sa_status == ROW_ACTIVE
	        && sa -> sa_resources == ourContext
	        && (sa -> sa_privileges & mask)
	        && (p = sa -> sa_target)
	        && p -> party_location == AGENT
	        && p -> party_tDomain == RFC1157DOMAIN)
	    return sa;

    return NULL;
}

/*    CONTEXT */

static struct snmp_context *find_context (interp, cp)
Tcl_Interp     *interp;
char	       *cp;
{
    int	    i;
    register struct snmp_context *sc;
    OID	    oid;

    i = strlen (cp);
    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_name
	        && sc -> sc_name_len == i
	        && !lexequ ((char *) sc -> sc_name, cp))
	    return sc;

    if (!(oid = text2oid (cp)))
	return fuzzy_context (interp, cp);

    sc = lookup_context_aux (oid, 0);
    oid_free (oid);

    if (!sc)
	return fuzzy_context (interp, cp);
    return sc;
}

/*  */

static struct snmp_context *fuzzy_context (interp, cp)
Tcl_Interp     *interp;
char	       *cp;
{
    int	    i;
    register char *dp;
    char    buffer[BUFSIZ],
	    error[BUFSIZ];
    register struct hostent *hp;
    struct sockaddr_in sins;
    register struct sockaddr_in *sin = &sins;
    register struct snmp_party *p1,
    			       *p2;
    register struct snmp_context *sc;
    register struct snmp_access *sa;
    OID	    oid = NULLOID;

    i = strlen (cp);
    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_name
	        && sc -> sc_name_len >= i
	        && !lexnequ ((char *) sc -> sc_name, cp, i))
	    return make_context (interp, cp);

/* not good as a wildcard, make it real... */
    p1 = p2 = NULL, sa = NULL;

    if (hp = gethostbystring (cp))
	dp = NULL;
    else
	if (dp = index (cp, '/')) {
	    *dp = NULL;
	    hp = gethostbystring (cp);
#ifndef	wrongdef
	    *dp = '/';
#endif
	}

    if (!hp) {
#ifdef	wrongdef
	if (dp)
	    *dp = '/';
#endif
	(void) sprintf (error, "%s: unknown host", cp);

losing: ;
	Tcl_SetResult (interp, error, TCL_VOLATILE);
losing_aux: ;
#ifdef	wrongdef
	if (dp)
	    *dp = '/';
#endif
	if (p1)
	    free_party (p1);
	if (p2)
	    free_party (p2);
	if (sc)
	    free_context (sc);
	if (sa)
	    free_access (sa);
	if (oid)
	    oid_free (oid);

	return NULL;
    }
    bzero ((char *) sin, sizeof *sin);
    sin -> sin_family = hp -> h_addrtype;
    inaddr_copy (hp, sin);
    sin -> sin_port = htons ((u_short) 161);

    (void) sprintf (buffer, "initialContextId.%s.1",
		    inet_ntoa (sin -> sin_addr));
    if (!(oid = text2oid (buffer))) {
oops: ;
	(void) sprintf (error, "text2oid (\"%s\") failed!", buffer);
	goto losing;
    }
    while (lookup_context_aux (oid, 0))
	oid -> oid_elements[oid -> oid_nelem - 1]++;
    if (!(sc = make_context (interp, cp)))
	goto losing;
    if (!(sc -> sc_identity = oid_cpy (oid))) {
no_mem: ;
	(void) strcpy (error, "out of memory");
	goto losing;
    }
    oid_free (oid), oid = NULL;

    (void) sprintf (buffer, "initialPartyId.%s.1",
		    inet_ntoa (sin -> sin_addr));
    if (!(oid = text2oid (buffer)))
	goto oops;
    while (lookup_party_aux (oid, EITHER, 0))
	oid -> oid_elements[oid -> oid_nelem - 1]++;
    if (!(p1 = make_party (interp, oid, dp ? dp + 1 : NULLCP, AGENT)))
	goto losing_aux;
    bcopy ((char *) &sin -> sin_addr,
	   (char *) p1 -> party_TAddress,
	   sizeof sin -> sin_addr);
    bcopy ((char *) &sin -> sin_port,
	   (char *) (p1 -> party_TAddress + sizeof sin -> sin_addr),
	   sizeof sin -> sin_port);
    oid_free (oid), oid = NULL;

    if (!dp) {
	(void) sprintf (buffer, "initialPartyId.%s.2",
			inet_ntoa (sin -> sin_addr));
	if (!(oid = text2oid (buffer)))
	    goto oops;
	while (lookup_party_aux (oid, EITHER, 0))
	    oid -> oid_elements[oid -> oid_nelem - 1]++;
	if (!(p2 = make_party (interp, oid, NULLCP, MANAGER)))
	    goto losing_aux;
	oid_free (oid), oid = NULL;
    }

    if (!(sa = (struct snmp_access *) calloc (1, sizeof *sa)))
	goto no_mem;
    sa -> sa_target_index = (sa -> sa_target = p1) -> party_index;
    sa -> sa_subject_index = (sa -> sa_subject = p2 ? p2 : p1) -> party_index;
    sa -> sa_resources_index = (sa -> sa_resources = sc) -> sc_index;
    {
	OID inst = &sa -> sa_instance;

	inst -> oid_nelem = 3;
	if (!(inst -> oid_elements = (unsigned int *)
			calloc ((unsigned int) inst -> oid_nelem,
				sizeof *inst -> oid_elements)))
	    goto no_mem;
	inst -> oid_elements[0] = sa -> sa_target_index;
	inst -> oid_elements[1] = sa -> sa_subject_index;
	inst -> oid_elements[2] = sa -> sa_resources_index;
    }
    sa -> sa_privileges = sa -> sa_new_privileges = PARTY_PRIV_GET
						  | PARTY_PRIV_NEXT
						  | PARTY_PRIV_SET  /* XXX */
						  | PARTY_PRIV_BULK;
    sa -> sa_status = ROW_ACTIVE;
    sa -> sa_storage = STORAGE_VOLATILE;

    sa -> sa_next = accessHead;
    accessHead = sa;

    if (dp)
	*dp = '/';

    return sc;
}

/*  */

static struct snmp_context *make_context (interp, cp)
Tcl_Interp     *interp;
char	       *cp;
{
    int	    i;
    char    buffer[BUFSIZ];
    register struct snmp_context  *sc,
				 **pc;
    if (!(sc = (struct snmp_context *) calloc (1, sizeof *sc))) {
no_mem: ;
	(void) strcpy (buffer, "out of memory");
	goto losing;
    }
    sc -> sc_name = (u_char *) strdup (cp);
    sc -> sc_name_len = strlen (cp);
    sc -> sc_location = AGENT;
    sc -> sc_view = 1;
    sc -> sc_entity = (u_char *) calloc (1, sizeof *sc -> sc_entity);
    sc -> sc_time = oid_cpy (currentTime);
    sc -> sc_dst_oid = oid_cpy (nullSpecific);
    sc -> sc_src_oid = oid_cpy (nullSpecific);
    sc -> sc_ctx = oid_cpy (nullSpecific);
    sc -> sc_status = ROW_ACTIVE;
    sc -> sc_storage = STORAGE_VOLATILE;

    if (!sc -> sc_name
	    || !sc -> sc_entity
	    || !sc -> sc_time
	    || !sc -> sc_dst_oid
	    || !sc -> sc_src_oid
	    || !sc -> sc_ctx) {
	free_context_aux (sc, 0);
	goto no_mem;
    }

    i = 1;
    for (pc = &contextHead; *pc; pc = &((*pc) -> sc_next))
	if ((*pc) -> sc_index >= i)
	    i = (*pc) -> sc_index + 1;
    *pc = sc;
    sc -> sc_index = i;

    return sc;

losing: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return NULL;
}

/*    PARTY */

/* ARGSUSED */

static void  start_clocks (clientData)
ClientData	clientData;
{
    register struct snmp_access *sa;


    if (debug || *watch)
	fprintf (stderr, "start_clocks %d invoked...\n", (int) clientData);

    if (!synctime) {
	if (!clientData) {
	    if (debug || *watch)
		fprintf (stderr, "spurious start_clocks...\n");
	    return;
	}

/* in progress, so tell find_clocks to schedule RSN */
	wantsync = 1;
	return;
    }

    if (clientData) {
/* authenticated transaction timeout... */
	if (synctoken)
	    Tk_DeleteTimerHandler (synctoken);
    }

    for (sa = accessHead; sa; sa = sa -> sa_next)
	sa -> sa_fresh = 0;
    wantsync = 0, synctime = 0L;
    find_clocks (0);
}

/*  */

static void  find_clocks (zero)
int	zero;
{
    int	    perms;
    u_long  msec;
    struct timeval tvs;
    register struct snmp_party *p,
			       *q,
			       *r;
    register struct snmp_access *sa,
				*qa;

again: ;
    perms = PARTY_PRIV_BULK;
    if (zero)
	perms |= PARTY_PRIV_SET;

    for (p = partyHead; p; p = p -> party_next) {
	if (p -> party_status != ROW_ACTIVE
	        || p -> party_location != AGENT
		|| (zero ? p -> party_authProt <= NOAUTH
			 : p -> party_authProt != NOAUTH))
	    continue;

	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE
		    && !sa -> sa_fresh
		    && sa -> sa_target == p
		    && (sa -> sa_privileges & perms) == perms)
		for (q = partyHead; q; q = q -> party_next)
		    if (p != q
			    && q -> party_status == ROW_ACTIVE
			    && q -> party_location == AGENT
			    && q -> party_authProt == MD5AUTHPROT
			    && p -> party_tDomain == q -> party_tDomain
			    && p -> party_TAddress_len
						    == q -> party_TAddress_len
			    && bcmp ((char *) p -> party_TAddress,
				     (char *) q -> party_TAddress,
				     p -> party_TAddress_len) == 0) {
			for (qa = accessHead; qa; qa = qa -> sa_next)
			    if (qa -> sa_status == ROW_ACTIVE
				    && ((qa -> sa_target == q
					        && (r = qa -> sa_subject)
					        && r -> party_location
								    == MANAGER)
					    || (qa -> sa_subject == q
						    && (r = qa -> sa_target)
						    && r -> party_location
								== MANAGER))
				    && (zero ? (!q -> party_clockzero
						    && !r -> party_clockzero)
					    : !r -> party_clocksync))
				break;
			if (qa) {
			    sa -> sa_fresh = 1;
			    get_clocks (sa, zero, 0, NULLPARTY, NULLCP);
			    return;
			}
		    }
    }

    if (!zero) {
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    sa -> sa_fresh = 0;

	zero = 1;
	goto again;
    }

    msec = (wantsync ? 5 : 6 * 60) * 60 * 1000L;
    synctoken = Tk_CreateTimerHandler ((int) msec, start_clocks,
				       (ClientData) 0);
    (void) gettimeofday (&tvs, (struct timezone *) 0);
    synctime = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L + msec;

    if (debug || *watch)
	fprintf (stderr, "clock sync scheduled in %u secs\n", msec / 1000L);
}

/*  */

static void  get_clocks (sa, zero, update, sync, suffix)
struct snmp_access *sa;
int	zero,
	update;
struct snmp_party *sync;
char   *suffix;
{
    int	    maxrept,
	    vecp;
    char    buffer[BUFSIZ],
	   *vec[10];
    Tcl_DString	dstring;
    register struct snmp_party *p;

    Tcl_DStringInit (&dstring);
    Tcl_DStringStartSublist (&dstring);
    (void) Tcl_DStringAppendElement (&dstring,
				     sprintoid (partyAProtocol -> ot_name));
    if (suffix)
	(void) Tcl_DStringAppend (&dstring, suffix, -1);
    Tcl_DStringEndSublist (&dstring);
    Tcl_DStringStartSublist (&dstring);
    (void) Tcl_DStringAppendElement (&dstring,
				     sprintoid (partyAClock -> ot_name));
    if (suffix)
	(void) Tcl_DStringAppend (&dstring, suffix, -1);
    Tcl_DStringEndSublist (&dstring);
    if (zero) {
	Tcl_DStringStartSublist (&dstring);
	(void) Tcl_DStringAppendElement (&dstring,
					 sprintoid (partyAPrivate -> ot_name));
	if (suffix)
	    (void) Tcl_DStringAppend (&dstring, suffix, -1);
	Tcl_DStringEndSublist (&dstring);
	Tcl_DStringStartSublist (&dstring);
	(void) Tcl_DStringAppendElement (&dstring, sprintoid (partyMMessageSize
							          -> ot_name));
	if (suffix)
	    (void) Tcl_DStringAppend (&dstring, suffix, -1);
	Tcl_DStringEndSublist (&dstring);
    }

    maxrept = 1;
    for (p = partyHead; p; p = p -> party_next)
	maxrept++;
    maxrept <<= 1;
    (void) sprintf (buffer, "%d", maxrept);

    vecp = 0;
    vec[vecp++] = pInternal -> p_name;
    vec[vecp++] = "bulk";
    vec[vecp++] = GET_CLOCKS;
    vec[vecp++] = Tcl_DStringValue (&dstring);
    vec[vecp++] = "-maxrept";
    vec[vecp++] = buffer;
    vec[vecp++] = "-nonrptr";
    vec[vecp++] = "0";
    vec[vecp] = NULL;

    pInternal -> p_acl = sa;
    if (SNMPhandleCommand ((ClientData) pInternal, tclInterp, vecp, vec)
	    != TCL_OK) {
	Tk_BackgroundError (tclInterp);	/* XXX: should probably die... */
    }
    else {	/* should read return value, but we know how it works... */
	struct iHandle *i = pInternal -> p_invokes;

	i -> i_zero = zero, i -> i_update = update, i -> i_sync = sync;
    }

    Tcl_DStringFree (&dstring);
}

/*  */

static void  get_clocks_callback (status, vp, sa, zero, update, sync)
integer	status;
register struct type_SNMP_VarBindList *vp;
struct snmp_access *sa;
int	zero,
	update;
struct snmp_party *sync;
{
    u_long  now;
    register struct snmp_party *p;
    OID	    next = NULL;
    struct OIDentifier oids;

    if (status != int_SNMP_error__status_noError)
	goto out;

    now = party_time_now ();
    for (p = partyHead; p; p = p -> party_next) {
	p -> party_fresh = 0, p -> party_new_AuthStart = 0;
	p -> party_old_AuthStart = p -> party_AuthStart;
	p -> party_new_MaxMessageSize = 0;
	p -> party_old_clone = 0;
    }

    for (; vp; vp = vp -> next) {
	caddr_t	    value;
	register struct type_SNMP_VarBind *v = vp -> VarBind;
	OID	vn = v -> name;
	register PE	vv = v -> value;
	register OI	oi;
	register OS	os;
	register OT	ot;

	if (vv -> pe_class == PE_CLASS_CONT) {
	    if (debug || *watch)
		fprintf (stderr, "get_clocks: %s=%s", oid2ode (vn),
			 snmpV2_exception (v -> value -> pe_id));
	    continue;
	}

	if (!(oi = name2inst (vn))
	        || !(ot = oi -> oi_type)
	        || !(os = oi -> oi_type -> ot_syntax))
	    continue;
	if (ot == partyAProtocol)
	    next = vn;
	else
	    if (ot == partyAClock) {
	    }
	    else
		if (zero && (ot == partyAPrivate || ot == partyMMessageSize)) {
		}
		else {
		    next = NULL;
		    continue;
		}
	oids.oid_nelem = vn -> oid_nelem
			     - partyAProtocol -> ot_name -> oid_nelem;
	oids.oid_elements = vn -> oid_elements
			        + partyAProtocol -> ot_name -> oid_nelem;
	if (!(p = lookup_party_aux (&oids, EITHER, 0))) {
	    if (debug)
		fprintf (stderr,
			 "ignoring information about unknown party %s\n",
			 oid2ode (&oids));
	    continue;
	}
	if (!zero && p -> party_location != MANAGER) {
	    if (debug)
		fprintf (stderr,
			 "don't need information about agent party %s\n",
			 oid2ode (p -> party_Identity));
	    continue;
	}
	if (!update
	        && (zero ? p -> party_clockzero : p -> party_clocksync)) {
	    if (debug)
		fprintf (stderr, "already know about party %s\n",
			 oid2ode (p -> party_Identity));
	    continue;
	}

	if ((*os -> os_decode) (&value, vv) == NOTOK)
	    continue;

	if (ot == partyAProtocol) {
	    if (oid_cmp ((OID) value, md5Auth)) {
		if (p -> party_authProt == MD5AUTHPROT) {
		    char    buffer[BUFSIZ];

wrong_auth: ;
		    (void) strcpy (buffer, oid2ode (p -> party_Identity));
		    if (debug)
			fprintf (stderr,
				 "get_clocks: agent thinks party %s is using %s",
				 buffer, oid2ode ((OID) value));
		}
	    }
	    else
		if (p -> party_authProt != MD5AUTHPROT)
		    goto wrong_auth;
		else
		    p -> party_fresh = 1;
	}
	else
	    if (ot == partyAClock) {
		if (p -> party_authProt == MD5AUTHPROT) {
		    u_long newclock = now - *((u_long *) value);

		    p -> party_new_AuthStart = newclock;
		}
	    }
	    else
		if (ot == partyAPrivate) {
		    if (p -> party_authProt == MD5AUTHPROT)
			p -> party_old_clone = 1;
		}
		else
		    p -> party_new_MaxMessageSize = *((integer *) value);

	(*os -> os_free) (value);
    }

    for (p = partyHead; p; p = p -> party_next) {
	if (p -> party_fresh)
	    p -> party_new_storage = 1;
	
	if (p -> party_fresh && p -> party_new_AuthStart)
	    if (zero) {
		if (p -> party_old_clone
		        && sa -> sa_target != p
		        && sa -> sa_subject != p)
		    p -> party_clockzero = sa;

		if (p -> party_new_MaxMessageSize >= MAXSNMP
		        && p -> party_new_MaxMessageSize <= BIGSNMP
		        && (p -> party_new_MaxMessageSize
			        != p -> party_MaxMessageSize)) {
		    if (debug)
			fprintf (stderr,
				 "party %s MTU changed from %d to %d\n",
				 oid2ode (p -> party_Identity),
				 p -> party_MaxMessageSize,
				 p -> party_new_MaxMessageSize);
		    p -> party_new_MaxMessageSize = p -> party_MaxMessageSize;
		}
	    }
	    else {
		p -> party_clocksync = sa;
		if (p -> party_AuthStart > p -> party_new_AuthStart) {
		    if (debug)
			fprintf (stderr,
				 "party %s clock changed from %ld to %ld\n",
				 oid2ode (p -> party_Identity),
				 now - p -> party_AuthStart,
				 now - p -> party_new_AuthStart);
		    p -> party_AuthStart = p -> party_new_AuthStart;
		    party_changed = 1;
		}
	    }
    }

out: ;
    if (next) {
	Tcl_DString	dstring;

	oids.oid_nelem = next -> oid_nelem
			     - partyAProtocol -> ot_name -> oid_nelem;
	oids.oid_elements = next -> oid_elements
			        + partyAProtocol -> ot_name -> oid_nelem;

	Tcl_DStringInit (&dstring);
	Tcl_DStringAppend (&dstring, ".", -1);
	Tcl_DStringAppend (&dstring, sprintoid (&oids), -1);

	get_clocks (sa, zero, update, sync, Tcl_DStringValue (&dstring));

	Tcl_DStringFree (&dstring);
    }
    else
	if (p = sync) {
	    if (!p -> party_new_storage) {
		if (debug || *watch)
		    fprintf (stderr, "invalidating knowledge for %s\n",
			     oid2ode (p -> party_Identity));

		p -> party_clocksync = NULL;
	    }
	}
        else
	    find_clocks (zero);
}

/*  */

static struct snmp_party *make_party (interp, oid, community, who)
Tcl_Interp     *interp;
OID		oid;
char	       *community;
int		who;
{
    int	    i;
    register struct snmp_party  *p,
			       **pp;

    if (!(p = (struct snmp_party *) calloc (1, sizeof *p))) {
no_mem: ;
	Tcl_SetResult (interp, "out of memory", TCL_STATIC);
	return NULL;
    }
    p -> party_Identity = oid_cpy (oid);
    if (community) {
	p -> party_TDomain = oid_cpy (rfc1157Domain);
	p -> party_tDomain = RFC1157DOMAIN;
	p -> party_AuthPrivate = (u_char *) strdup (community);
	p -> party_AuthPrivate_len = strlen (community);
    }
    else {
	p -> party_TDomain = oid_cpy (snmpUDPDomain);
	p -> party_tDomain = SNMPUDPDOMAIN;
    }
    p -> party_TAddress = (u_char *) calloc ((unsigned) p -> party_TAddress_len = 6,
					     sizeof *p -> party_TAddress);
    p -> party_MaxMessageSize = MAXDGRAM;
    p -> party_location = who;
    if (community) {
	p -> party_AuthProtocol = oid_cpy (rfc1157noAuth);
	p -> party_authProt = RFC1157NOAUTH;
    }
    else {
	p -> party_AuthProtocol = oid_cpy (noAuth);
	p -> party_authProt = NOAUTH;
    }
    p -> party_AuthStart = party_time_now ();
    p -> party_AuthLifetime = 300;
    p -> party_PrivProtocol = oid_cpy (noPriv);
    p -> party_privProt = NOPRIV;
    p -> party_status = ROW_ACTIVE;
    p -> party_storage = STORAGE_VOLATILE;

    if (!p -> party_Identity
	    || !p -> party_TDomain
	    || !p -> party_TAddress
	    || !p -> party_AuthProtocol
	    || (p -> party_tDomain == RFC1157DOMAIN
		    && !p -> party_AuthPrivate)
	    || !p -> party_PrivProtocol) {
	free_party_aux (p, 0);
	goto no_mem;
    }

    i = 1;
    for (pp = &partyHead; *pp; pp = &((*pp) -> party_next))
	if ((*pp) -> party_index >= i)
	    i = (*pp) -> party_index + 1;
    *pp = p;
    p -> party_index = i;

    return p;
}

/*    DGRAM */

static struct dHandle *find_dHandle (interp, p)
Tcl_Interp *interp;
register struct snmp_party *p;
{
    char    buffer[BUFSIZ];
    register struct dHandle *d;
#ifdef	TCP
    register struct sockaddr_in *isock;
#endif

    for (d = dHead; d; d = d -> d_next)
	if (d -> d_tdomain == p -> party_tDomain) {
	    if (d -> d_lastsend == p)
		return d;

	    break;
	}
/* could do special check to see if party_tDomain is RFC1157DOMAIN, in that
   case a d_tdomain of SNMPUDPDOMAIN would work, but why not just have a
   dHandle for each instead of having to check each time... */

    if (!d && !(d = make_dHandle (interp, p, 1)))
	return NULL;
    
    switch (d -> d_tdomain) {
#ifdef	TCP
        case SNMPUDPDOMAIN:
        case RFC1157DOMAIN:
	    isock = (struct sockaddr_in *) &d -> d_agent;
	    isock -> sin_family = AF_INET;
	    bcopy ((char *) p -> party_TAddress, (char *) &isock -> sin_addr,
		   sizeof isock -> sin_addr);
	    bcopy ((char *) p -> party_TAddress + sizeof isock -> sin_addr,
		   (char *) &isock -> sin_port, sizeof isock -> sin_port);
	    break;
#endif

	default:
	    (void) sprintf (buffer, "unsupported transport domain %s",
			    oid2ode (p -> party_TDomain));
	    goto tcl_error;
    }

    d -> d_lastsend = p;
    return d;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return NULL;
}

/*  */

static int  make_dHandles (interp)
Tcl_Interp *interp;
{
    register struct snmp_access *sa;
    register struct snmp_party *p;
    register struct dHandle *d;

    for (sa = accessHead; sa; sa = sa -> sa_next)
	if (sa -> sa_status == ROW_ACTIVE
	        && (p = sa -> sa_target)
	        && p -> party_location == MANAGER
		&& (sa -> sa_privileges
		        & (PARTY_PRIV_TRAP | PARTY_PRIV_INFO |PARTY_PRIV_NTFY))
	        && !make_dHandle (interp, p, 0))
	    goto tcl_error;

    return TCL_OK;

tcl_error: ;
    while (d = dHead) {
	if (d -> d_sd != NOTOK)
	    (void) close (d -> d_sd);
	if (d -> d_ps)
	    ps_free (d -> d_ps);
	dHead = d -> d_next;

	free ((char *) d);
    }
    return TCL_ERROR;
}

/*  */

static struct dHandle *make_dHandle (interp, p, new)
Tcl_Interp *interp;
register struct snmp_party *p;
int	new;
{
    char    buffer[BUFSIZ];
    struct sockaddr sas;
#ifdef	TCP
    register struct sockaddr_in *isock = (struct sockaddr_in *) &sas;
#endif
    register struct dHandle *d;

    bzero ((char *) &sas, sizeof sas);
    switch (p -> party_tDomain) {
#ifdef	TCP
        case SNMPUDPDOMAIN:
        case RFC1157DOMAIN:
	    isock -> sin_family = AF_INET;
	    if (!new) {
		bcopy ((char *) p -> party_TAddress,
		       (char *) &isock -> sin_addr,
		       sizeof isock -> sin_addr);
		bcopy ((char *) p -> party_TAddress + sizeof isock -> sin_addr,
		       (char *) &isock -> sin_port, sizeof isock -> sin_port);
		for (d = dHead; d; d = d -> d_next)
		    if (d -> d_tdomain == p -> party_tDomain
		            && bcmp ((char *) &d -> d_manager, (char *) isock,
				     sizeof *isock) == 0)
			break;
		if (d) {
		    if (debug)
			fprintf (stderr, "already listening for party %s\n",
				 oid2ode (p -> party_Identity));
		    break;
		}
	    }
	    if (!(d = make_udp_dHandle (interp, p, isock)))
		return NULL;
	    break;
#endif

	default:
	    (void) sprintf (buffer, "unsupported transport domain %s",
			    oid2ode (p -> party_TDomain));
	    goto tcl_error;
    }
    return d;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return NULL;
}

/*  */

#ifdef	TCP
static struct dHandle *make_udp_dHandle (interp, p, isock)
Tcl_Interp *interp;
register struct snmp_party *p;
struct sockaddr_in *isock;
{
    int	    onoff = 1,
	    maxsize,
	    sd;
    char    buffer[BUFSIZ];
    register struct dHandle *d;
    PS	    ps;

    if ((sd = socket (AF_INET, SOCK_DGRAM, 0)) == NOTOK) {
	(void) sprintf (buffer, "unable to create UDP socket: %s",
			Tcl_PosixError (interp));
	goto tcl_error;
    }
    if (isock -> sin_port) {
	if (bind (sd, (struct sockaddr *) isock, sizeof *isock) == NOTOK) {
	    (void) sprintf (buffer, "unable to bind for party %s: %s",
			    oid2ode (p -> party_Identity),
			    Tcl_PosixError (interp));
	    goto tcl_error;
	}

	if (debug)
	    fprintf (stderr, "listening for party %s at %s\n",
		     oid2ode (p -> party_Identity), party_address (p));
    }
    else {
	int	port;

	for (port = IPPORT_RESERVED;; port++) {
	    isock -> sin_port = htons ((u_short) port);
	    if (bind (sd, (struct sockaddr *) isock, sizeof *isock) != NOTOK)
		break;
	    if (errno != EADDRINUSE) {
		(void) sprintf (buffer, "unable to bind to port %d: %s",
				port, Tcl_PosixError (interp));
		goto tcl_error;
	    }
	}

	if (debug)
	    fprintf (stderr, "listening for party %s on port %d\n",
		     oid2ode (p -> party_Identity),
		     ntohs ((u_short) isock -> sin_port));
	isock -> sin_port = htons ((u_short) 0);
    }

    if ((maxsize = p -> party_MaxMessageSize) < MAXDGRAM)
	maxsize = MAXDGRAM;

    if ((d = (struct dHandle *) calloc (1, (unsigned) (sizeof *d + maxsize)))
	    == NULL) {
	(void) strcpy (buffer, "out of memory");
	goto tcl_error;
    }
    d -> d_sd = NOTOK;

    d -> d_tdomain = p -> party_tDomain;
    bcopy ((char *) isock, (char *) &d -> d_manager, sizeof *isock);

    Tk_CreateFileHandler (sd, TK_READABLE, recv_message, (ClientData) d);
#ifdef	FIONBIO
    (void) ioctl (sd, FIONBIO, (char *) &onoff);
#else
#ifdef	O_NDELAY
    (void) fcntl (sd, F_SETFL, O_NDELAY);
#endif
#endif

    if (!(ps = ps_alloc (dg_open))
	    || dg_setup (ps, (int) d, d -> d_maxsize = maxsize, read_dHandle,
			 write_dHandle, NULLIFP) == NOTOK) {
	if (ps)
	    (void) sprintf (buffer, "dg_setup: %s", ps_error (ps -> ps_errno));
	else
	    (void) strcpy (buffer, "ps_alloc(dg_open): failed");

	free ((char *) d);
	goto tcl_error;
    }
    d -> d_sd = sd, d -> d_ps = ps;

    d -> d_next = dHead, dHead = d;
    return d;

tcl_error: ;
    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    if (sd != NOTOK)
	(void) close (sd);
    return NULL;
}
#endif

/*  */

static int  read_dHandle (fd, q)
int	fd;
struct qbuf **q;
{
    int	    cc,
	    slen;
    register struct dHandle *d = (struct dHandle *) fd;
    register struct qbuf *qb;

    if ((qb = (struct qbuf *) malloc ((unsigned) (sizeof *qb
						      + d -> d_maxsize)))
	     == NULL)
	return NOTOK;
    qb -> qb_data = qb -> qb_base;

    slen = sizeof d -> d_lastrecv;
    if ((cc = recvfrom (d -> d_sd, qb -> qb_data, d -> d_maxsize, 0,
			&d -> d_lastrecv, &slen)) == NOTOK) {
	d -> d_errno = errno;
	free ((char *) qb);
	return NOTOK;
    }
#ifdef	BSD44
    d -> d_lastrecv.sa_len = slen;
#endif
    qb -> qb_len = cc;
    qb -> qb_forw = qb -> qb_back = qb;

    if (debug)
	bcopy (qb -> qb_data, d -> d_packet, d -> d_pktsize = cc);

    *q = qb;

    return cc;
}

/*  */

static int  write_dHandle (fd, qb)
int	fd;
struct qbuf *qb;
{
    int	    result;
    register struct dHandle *d = (struct dHandle *) fd;

    if ((result = sendto (d -> d_sd, qb -> qb_data, qb -> qb_len, 0,
			  &d -> d_agent, sizeof d -> d_agent)) == NOTOK)
	d -> d_errno = errno;

    if (debug) {
	register char  *dp,
		       *ep;

	fprintf (stderr, "write %d octets\n", qb -> qb_len);
	for (ep = (dp = qb -> qb_data) + qb -> qb_len; dp < ep; ) {
	    char    *cp;
	    register char *fp;

	    if ((fp = dp + 16) > ep)
		fp = ep;
	    for (cp = ""; dp < fp; dp++, cp = " ")
		fprintf (stderr, "%s%02x", cp, *dp & 0xff);
	    fprintf (stderr, "\n");
	}
    }

    return result;
}

/*    MIBS */

static int  resolve_iit (interp, ot)
Tcl_Interp *interp;
OT	ot;
{
    int	    vecp;
    char    buffer[BUFSIZ],
	  **vec,
	  **xvec;
    register struct iinfo *iit;
    OT	    ot2;

    if (ot -> ot_augments) {
	if (!(ot2 = text2obj (ot -> ot_augments))) {
	    (void) sprintf (buffer, "unable to resolve %s", ot -> ot_augments);
	    goto tcl_error;
	}
	if (!ot2 -> ot_index) {
	    (void) sprintf (buffer,
			    "MIB error, %s doesn't have an INDEX clause",
			    ot2 -> ot_text);
	    goto tcl_error;
	}
	if (!ot2 -> ot_iit && resolve_iit (interp, ot2) != TCL_OK)
	    return TCL_ERROR;
	ot -> ot_iit = ot2 -> ot_iit;

	return TCL_OK;
    }

    if (Tcl_SplitList (interp, ot -> ot_index, &vecp, &vec) != TCL_OK)
	goto tcl_error;
    xvec = vec;

    if (!(iit = (struct iinfo *) calloc ((unsigned) (vecp + 1), sizeof *iit))){
	(void) strcpy (buffer, "out of memory");
	free ((char *) xvec);
	goto tcl_error;
    }

    for (ot -> ot_iit = (caddr_t) iit; vecp-- > 0; vec++, iit++) {
	int	impl;
	char   *cp;

	if (impl = (*(cp = *vec) == '*'))
	    cp++;
	    
	if (!(ot2 = text2obj (cp))) {
	    (void) sprintf (buffer, "unable to resolve %s", cp);
	    free (ot -> ot_iit), ot -> ot_iit = NULL;
	    free ((char *) xvec);
	    goto tcl_error;
	}
	iit -> ii_ot = ot2;

	if (!impl) {
	    char   *dp;
	    OS	    os;

	    if (!(os = ot2 -> ot_syntax)) {
		(void) sprintf (buffer, "unknown syntax for %s", cp);
losing: ;
		iit = (struct iinfo *) ot -> ot_iit;
		iit -> ii_ot = NULL;
		free ((char *) xvec);
		goto tcl_error;
	    }

	    dp = os -> os_textual ? os -> os_textual : os -> os_name;	    
	    {
		register struct pair *pp;

		for (pp = pairs; pp -> p_old; pp++)
		    if (strcmp (pp -> p_old, dp) == 0) {
			dp = pp -> p_new;
			break;
		    }
	    }
	    switch (*dp) {
		case 'D':
		    if (strcmp (dp, "DisplayString") == 0) {
do_string: ;
			if (ot2 -> ot_range
			        && ot2 -> ot_lendpoint == ot2 -> ot_rendpoint)
			    iit -> ii_length = ot2 -> ot_lendpoint;
			else
			    iit -> ii_length = -1;
			break;
		    }
		    goto none;

		case 'G':
		    if (strcmp (dp, "Gauge32") == 0) {
			iit -> ii_length = 1;
			break;
		    }
		    goto none;

		case 'I':
		    if (strcmp (dp, "Integer32") == 0) {
			iit -> ii_length = 1;
			break;
		    }
		    if (strcmp (dp, "IpAddress") == 0) {
			iit -> ii_length = 4;
			break;
		    }
		    goto none;

		case 'N':
		    if (strcmp (dp, "NetworkAddress") == 0) {
			iit -> ii_length = 5;
			break;
		    }
		    if (strcmp (dp, "NsapAddress") == 0) {
			iit -> ii_length = -1;
			break;
		    }
		    goto none;

		case 'O':
		    if (strcmp (dp, "OctetString") == 0)
			goto do_string;
		    if (strcmp (dp, "ObjectID") == 0) {
			iit -> ii_length = -1;
			break;
		    }
		    goto none;
		    
		case 'U':
		    if (strcmp (dp, "UInteger32") == 0) {
			iit -> ii_length = 1;
			break;
		    }
		    goto none;

		default:
none: ;
		    (void) sprintf (buffer, "invalid syntax for %s, %s",
				    cp, dp);
		    goto losing;
	    }
	}
    }

    free ((char *) xvec);

    return TCL_OK;

tcl_error: ;
    if (interp)
	Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_ERROR;
}

/*  */

static void append_iit (interp, ot, oid)
Tcl_Interp *interp;
OT	ot;
OID	oid;
{
    char   *cp = "";
    register struct iinfo *iit =
				(struct iinfo *) ((OT) ot -> ot_iit) -> ot_iit;
    Tcl_DString	dstring;

    Tcl_DStringInit (&dstring);

    for (; iit -> ii_ot; iit++) {
	register int j;
	int	i = iit -> ii_length;
	register unsigned int *ip;
	char   *dp,
		buffer[BUFSIZ];
	OS	os = iit -> ii_ot -> ot_syntax;

	if (i < 0) {
	    i = oid -> oid_elements[0];
	    oid -> oid_elements++, oid -> oid_nelem--;
	}
	else
	    if (i == 0) {
		register struct iinfo *ii2;

		j = 0;
		for (ii2 = iit + 1; ii2 -> ii_ot; ii2++) {
		    int	    k = ii2 -> ii_length;

		    if (k <= 0)
			goto out;
		    j += k;
		}
		i = oid -> oid_nelem - j;
	    }
	if (oid -> oid_nelem < i)
	    goto out;

	ip = oid -> oid_elements, j = i;

	dp = os -> os_textual ? os -> os_textual : os -> os_name;
	{
	    register struct pair *pp;

	    for (pp = pairs; pp -> p_old; pp++)
		if (strcmp (pp -> p_old, dp) == 0) {
		    dp = pp -> p_new;
		    break;
		}
	}
	switch (*dp) {
	    case 'D':
	        if (strcmp (dp, "DisplayString") == 0) {
		    int	    notnvt;
		    Tcl_DString	nstring;

do_string: ;
		    notnvt = 0;
		    for (; j-- > 0; ip++)
			if (!isspace ((u_char) *ip)
			        && !isprint ((u_char) *ip)) {
			    notnvt = 1;
			    break;
			}
		    ip = oid -> oid_elements, j = i;

		    Tcl_DStringInit (&nstring);

		    if (notnvt) {
do_raw: ;
			dp = "0x";
			for (; j-- > 0; ip++, dp = ":") {
			    (void) sprintf (buffer, "%s%02x", dp, *ip & 0xff);
			    (void) Tcl_DStringAppend (&nstring, buffer, -1);
			}
		    }
		    else
			for (; j-- > 0; ip++) {
			    char    c = *ip & 0xff;

			    (void) Tcl_DStringAppend (&nstring, &c, 1);
			}

		    Tcl_DStringAppendElement (&dstring,
					      Tcl_DStringValue (&nstring));

		    Tcl_DStringFree (&nstring);
		    break;
		}
		goto do_string;

	    case 'G':
		if (strcmp (dp, "Gauge32") == 0) {
		    (void) sprintf (buffer, "%U", *ip);
		    (void) Tcl_DStringAppendElement (&dstring, buffer);
		    break;
		}
		goto do_string;

	    case 'I':
		if (strcmp (dp, "Integer32") == 0) {
		    (void) sprintf (buffer, "%d", *ip);
		    (void) Tcl_DStringAppendElement (&dstring, buffer);
		    break;
		}
		if (strcmp (dp, "IpAddress") == 0) {
do_ipaddr: ;
		    (void) sprintf (buffer, "%d.%d.%d.%d",
				    ip[0], ip[1], ip[2], ip[3]);
		    (void) Tcl_DStringAppendElement (&dstring, buffer);
		    break;
		}
		goto do_string;

	    case 'N':
		if (strcmp (dp, "NetworkAddress") == 0) {
		    ip++;
		    goto do_ipaddr;
		}
		if (strcmp (dp, "NsapAddress") == 0) {
		    ip++, j--;
		    (void) Tcl_DStringAppendElement (&dstring, "");
		    goto do_raw;
		}
		goto do_string;

	    case 'O':
		if (strcmp (dp, "OctetString") == 0)
		    goto do_string;
		if (strcmp (dp, "ObjectID") == 0) {
		    struct OIDentifier oids;
		    oids.oid_elements = ip, oids.oid_nelem = j;
		    (void) Tcl_DStringAppendElement (&dstring,
						     sprintoid (&oids));
		    break;
		}
		goto do_string;

	    case 'U':
		if (strcmp (dp, "UInteger32") == 0) {
		    (void) sprintf (buffer, "%U", *ip);
		    (void) Tcl_DStringAppendElement (&dstring, buffer);
		    break;
		}
		goto do_string;

	    default:
		goto do_string;
	}

	oid -> oid_elements += i, oid -> oid_nelem -= i;
    }
    if (oid -> oid_nelem > 0)
	goto out;

    cp = Tcl_DStringValue (&dstring);

out: ;
    Tcl_AppendElement (interp, cp);
    Tcl_DStringFree (&dstring);
    return;
}

/*    MISC */

char   *sprintoid (oid)
register OID	oid;
{
    register int    i;
    register unsigned int  *ip;
    register char  *cp;
    char    buffer[BUFSIZ];
    static int once_only = 0;
    static Tcl_DString	dstring;

    if (oid == NULLOID || oid -> oid_nelem < 1)
	return "";

    if (!once_only) {
	once_only = 1;

	Tcl_DStringInit (&dstring);
    }
    else
	Tcl_DStringFree (&dstring);
    
    for (ip = oid -> oid_elements, i = oid -> oid_nelem, cp = "";
	    i-- > 0;
	    ip++, cp = ".") {
	(void) sprintf (buffer, "%s%u", cp, *ip);
	(void) Tcl_DStringAppend (&dstring, buffer, -1);
    }

    return Tcl_DStringValue (&dstring);
}

/*    MAGIC */

#ifdef	MAGIC
/* taken from psap/pe2ps.c */


static PElement pe_eoc = { PE_CLASS_UNIV, PE_FORM_PRIM, PE_UNIV_EOC, 0 };
static int  pe2ps_aux2 ();
static int  ps_write_id ();
static int  ps_write_len ();

/*  */

static int  pe2ps_aux (ps, pe, eval)
register PS	ps;
register PE	pe;
int	eval;
{
    int     result;

    if (magicommand == MAGIC_INDFLEN)
	ps_len_strategy = PS_LEN_INDF;

    if (eval > 0)
	switch (pe -> pe_form) {
	    case PE_FORM_PRIM: 
	    case PE_FORM_ICONS: 
		break;

	    case PE_FORM_CONS: 
		(void) ps_get_abs (pe);
		break;
	}

    if (magicommand == MAGIC_INDFLEN)
	ps_len_strategy = PS_LEN_LONG;

    if (magicommand == MAGIC_UDPLESBER)
	pe -> pe_len++;
    if ((result = pe2ps_aux2 (ps, pe, eval)) != NOTOK) {
	if (magicommand == MAGIC_BERLESUDP)
	    (void) ps_write_aux (ps, (PElementData) "surprise!",
				 (PElementLen) strlen ("surprise!"), 1);

	result = ps_flush (ps);
    }

    return result;
}


static int  pe2ps_aux2 (ps, pe, eval)
register PS	ps;
register PE	pe;
int	eval;
{
    register PE	    p;

    if (pe -> pe_form == PE_FORM_ICONS) {
	if (ps_write_aux (ps, pe -> pe_prim, pe -> pe_len, 1) == NOTOK)
	    return NOTOK;

	return OK;
    }
    
    if (ps_write_id (ps, pe) == NOTOK || ps_write_len (ps, pe) == NOTOK)
	return NOTOK;

    switch (pe -> pe_form) {
	case PE_FORM_PRIM: 
	    if (ps_write_aux (ps, pe -> pe_prim, pe -> pe_len, 1) == NOTOK)
		return NOTOK;
	    break;

	case PE_FORM_CONS: 
	    if (eval < 0)
		    return OK;
	    if (pe -> pe_len) {
		for (p = pe -> pe_cons; p; p = p -> pe_next)
		    if (pe2ps_aux2 (ps, p, 0) == NOTOK)
			return NOTOK;

		if (pe -> pe_len == PE_LEN_INDF
			&& pe2ps_aux2 (ps, &pe_eoc, 0) == NOTOK)
		    return NOTOK;
	    }
	    break;
    }

    return OK;
}

/*  */

static int  ps_write_id (ps, pe)
register PS	ps;
register PE	pe;
{
    byte    buffer[1 + sizeof (PElementID)];
    register byte  *bp = buffer;
    PElementForm    form;
    register PElementID id;

    if ((form = pe -> pe_form) == PE_FORM_ICONS)
	form = PE_FORM_CONS;
    *bp = ((pe -> pe_class << PE_CLASS_SHIFT) & PE_CLASS_MASK)
		| ((form << PE_FORM_SHIFT) & PE_FORM_MASK);

    if ((id = pe -> pe_id) < PE_ID_XTND)
	*bp++ |= id;
    else {
	register byte *ep;
	register PElementID jd;

	*bp |= PE_ID_XTND;

	ep = buffer;
	for (jd = id; jd != 0; jd >>= PE_ID_SHIFT)
	    ep++;

	for (bp = ep; id != 0; id >>= PE_ID_SHIFT)
	    *bp-- = id & PE_ID_MASK;
	for (bp = buffer + 1; bp < ep; bp++)
	    *bp |= PE_ID_MORE;

	bp = ++ep;
    }

    if (ps_write (ps, buffer, bp - buffer) == NOTOK)
	return NOTOK;

    return OK;
}

/*  */

/* probably should integrate the non-PE_LEN_SMAX case with the algorithm in
   num2prim() for a single, unified routine */

static int  ps_write_len (ps, pe)
register PS	ps;
register PE	pe;
{
    byte    buffer[1 + sizeof (PElementLen)];
    register byte  *bp = buffer,
		   *ep;
    register PElementLen len;

    if ((len = pe -> pe_len) == PE_LEN_INDF)
	*bp++ = PE_LEN_XTND;
    else
	if (len <= PE_LEN_SMAX && magicommand != MAGIC_NOTMINLEN)
	    *bp++ = len & 0xff;
	else {
	    ep = buffer + sizeof buffer - 1;
	    for (bp = ep; len != 0 && buffer < bp; len >>= 8)
		*bp-- = len & 0xff;
	    if (magicommand == MAGIC_NOTMINLEN) {
		if (pe -> pe_len == 0)
		    *bp-- = 0;
		if ((ep - bp) < sizeof (PElementLen))
		    *bp-- = 0;
	    }
	    *bp = PE_LEN_XTND | ((ep - bp) & 0xff);
	    if (ps_write (ps, bp, ep - bp + 1) == NOTOK)
		return NOTOK;

	    return OK;
	}

    if (ps_write (ps, buffer, bp - buffer) == NOTOK)
	return NOTOK;

    return OK;
}

/*  */

/* taken from psap/ps_get_abs.c */

static  int	ps_get_abs (pe)
register PE	pe;
{
    register    PElementLen len;
    register    PE p;

    switch (pe -> pe_form) {
	case PE_FORM_PRIM: 
	    len = pe -> pe_len;
	    break;

	case PE_FORM_CONS: 
	    len = 0;
	    for (p = pe -> pe_cons; p; p = p -> pe_next)
		len += ps_get_abs (p);

	    switch (ps_len_strategy) {
		case PS_LEN_SPAG: 
		default: 
		    if (len <= PE_LEN_SMAX) {
			pe -> pe_len = len;
			break;
		    }		/* else fall */

		case PS_LEN_INDF: 
		    pe -> pe_len = PE_LEN_INDF;
		    len += 2;	/* for EOC */
		    break;

		case PS_LEN_LONG: 
		    pe -> pe_len = len;
		    break;
	    }
	    break;

	case PE_FORM_ICONS:
	    return pe -> pe_len;
    }

    return (ps_get_id (pe) + ps_get_len (pe) + len);
}

/*  */

static int  ps_get_id (pe)
register PE	pe;
{
    register int    i;
    register PElementID id;

    if ((id = pe -> pe_id) < PE_ID_XTND)
	return 1;

    for (i = 1; id != 0; id >>= PE_ID_SHIFT)
	i++;
    return i;
}

/*  */

static int  ps_get_len (pe)
register PE	pe;
{
    register int    i;
    register PElementLen len;

    if ((len = pe -> pe_len) == PE_LEN_INDF || len <= PE_LEN_SMAX)
	return ((magicommand == MAGIC_NOTMINLEN) ? 3 : 1);

    for (i = 1; len > 0; len >>= 8)
	i++;
    if ((magicommand == MAGIC_NOTMINLEN) && (i <= sizeof (PElementLen)))
	i++;

    return i;
}
#endif
