#!/usr/local/bin/snmptcl -f

if {![info exists snmptcl_loaded(polling.tcl)]} {
    if {[file readable polling.tcl]} {
	source			polling.tcl
    } else {
	source			$snmptcl_library/polling.tcl
    }
}

if {[info exists snmptcl_loaded(browsing.tcl)]} {
    return
}
set	snmptcl_loaded(browsing.tcl)	1


if {![file readable [set snmptcl_browsing_icon browsing.icon]]} {
    set	snmptcl_browsing_icon	$snmptcl_library/browsing.icon
}

if {[info commands FSBox] == ""} {
    if {[file readable FSBox.tcl]} {
	source			FSBox.tcl
    } else {
	source			$snmptcl_library/FSBox.tcl
    }
}


#
# browse instances
#
#     pHandle - context procedure
#         ask - ask about objects, regardless of A-C
#		-1 = never ask, 0 = ask if A-C info available, 1 = always ask
#       title - title for window
#       iname - name for icon
#        objs - list of instances to browse, each element may be a list,
#		if so, first is descriptor/OID and second is label
#    interval - polling interval (in seconds)
#     command - command to call on each iteration
#		(see snmptcl_poll_callback for calling procedure)
#

proc	snmptcl_proc_browse \
	{pHandle ask title iname objs interval {command snmptcl_browse_command}} {
    global	snmptcl_poll_cookie snmptcl_browsing_icon

    if {[catch { $pHandle config active }]} {
	blt_bell	100
	snmptcl_proc_mkmsg \
			"first, select a management context"
	return
    }
    snmptcl_proc_mkmsg		""

    set		w		.$pHandle
    while {[winfo exists $w$snmptcl_poll_cookie]} {
	incr	snmptcl_poll_cookie
    }
    append	w		$snmptcl_poll_cookie
    incr	snmptcl_poll_cookie

    if {($ask >= 0) \
	    && ([llength [set resources [$pHandle config resources]]] > 0) \
	    && (![catch { snmpinfo agentprop [lindex $resources 0] sysUpTime \
			      support }])} {
	set	avail		1
	set	ask		1
    } else {
	set	avail		0
    }

    if {$ask < 1} {
	snmptcl_proc_poll	$pHandle $w $title $iname \
				@$snmptcl_browsing_icon $objs $interval \
				$command
	return
    }

    toplevel	$w

    wm title	$w		$title
    wm iconname	$w		$iname
    wm iconbitmap \
		$w		@$snmptcl_browsing_icon

    frame	$w.frame	-borderwidth 10

    set		array		snmptcl_poll_data[join [split $w "."] "_"]
    global	$array
    upvar #0	$array		varinfo
    catch { unset varinfo }

    set		ovars		""
    foreach	obj	$objs {
	if {[llength $obj] > 1} {
	    set		ovar	[lindex $obj 0]
	} else {
	    set		ovar	$obj
	}
	if {[catch { set descr [snmpinfo mibprop $ovar descriptor] } result]} {
	    destroy	$w
	    snmptcl_dialog	$title $result questhead 0 Abort
	    unset	varinfo
	    return
	}
	if {[info exists varinfo($descr)]} {
	    continue
	}
	if {!$avail} {
	    set		varinfo($descr)	1
	} else {
	    set		varinfo($descr)	0
	    foreach	resource	$resources {
		if {[catch { set support [snmpinfo agentprop $resource $descr \
					      support] }]} {
		    continue
		}
		if {$support} {
		    set	    varinfo($descr)	1
		    break
		}
	    }
	}
	pack append \
		$w.frame	[checkbutton $w.frame.$descr \
					-relief flat \
					-text $descr \
					-variable "${array}($descr)"] \
				    {top frame w}
    }

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[button $w.bot.ok \
					-command "destroy $w; snmptcl_proc_browse_aux [list $pHandle $w $title $iname $objs $interval $command]" \
					-text OK] \
				    {left expand padx 5 pady 5} \
				[button $w.bot.cancel \
					-command "destroy $w; unset $array" \
					-text Cancel] \
				    {left expand padx 5 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}

proc	snmptcl_proc_browse_aux \
	{pHandle w title iname objs interval command} {
    global	snmptcl_browsing_icon

    set		array		snmptcl_poll_data[join [split $w "."] "_"]
    global	$array
    upvar #0	$array		varinfo

    set		ovars		""
    foreach	obj		$objs {
	if {[llength $obj] > 1} {
	    set		ovar	[lindex $obj 0]
	    set		olabel	[lindex $obj 1]
	} else {
	    set		olabel	[set ovar $obj]
	}
	set 	descr		[snmpinfo mibprop $ovar descriptor]
	if {$varinfo($descr)} {
	    lappend	ovars	[list $ovar $olabel]
	}
    }

    unset	varinfo

    if {[llength $ovars] <= 0} {
	snmptcl_dialog	$title "nothing selected" info 0 OK
	return
    }
    snmptcl_proc_poll		$pHandle $w $title $iname \
				@$snmptcl_browsing_icon $ovars $interval \
				$command
}



#
# default browse instances command
#

proc	snmptcl_browse_command	{w array} {
    global	snmptcl_print_object snmptcl_print_syntax

    global	${array}_info
    upvar #0	${array}_info	ainfo

    global	${array}_var
    upvar #0	${array}_var	avar

    set		frame		$ainfo(frame)
    if {![info exists ainfo(onceonly)]} {
	set	ainfo(onceonly)	1

	catch { $frame configure -background lightskyblue }

	set	width		0
	foreach	obj		$ainfo(objs) {
	    if {[set owid [string length $obj]] > $width} {
		set	width	$owid
	    }
	}

	set	i		0
	foreach	obj		$ainfo(objs) {
	    pack append \
		    $frame	[frame $frame.$i \
					-borderwidth 2] \
				    {top fillx}
	    catch { $frame.$i configure -background lightskyblue }
	
	    pack append \
		    $frame.$i	[label $frame.$i.obj \
					-text [format "%*s" $width $obj] \
					-width $width] \
				    {left padx 5 pady 5} \
				[label $frame.$i.val] \
				    {left fillx}

	    catch { $frame.$i.obj configure -background lightskyblue }
	    catch { $frame.$i.val configure -background lightskyblue }

	    incr    i
	}
    }

    set		i		0
    foreach	var		$ainfo(vars) {
	if {[info exists avar($var)]} {
	    set		data	$avar($var)

	    set		value	[lindex $data 4]
	    set		syntax	[lindex $data 6]
	    if {[set object [lindex $data 7]] == ""} {
		lappend	avar($var)	\
				[set object [snmpinfo mibprop $var descriptor]]
	    }
	    if {[info exists snmptcl_print_object($object)]} {
		set	value	[$snmptcl_print_object($object) $var \
								 avar($var)]
	    } elseif {[info exists snmptcl_print_syntax($syntax)]} {
		set	value	[$snmptcl_print_syntax($syntax) $var \
								 avar($var)]
	    }

	    if {"[$frame.$i.val configure -text]" != $value} {
		$frame.$i.val	configure \
					-text $value
	    }
	} else {
	    $frame.$i.val \
			configure \
				-text ""
	}
	incr	i
    }
}


#
# default graph instances command
#

proc	snmptcl_graph_command {xtitle ytitle w array} {
    global	snmptcl_graph_object snmptcl_graph_syntax

    global	${array}_info
    upvar #0	${array}_info	ainfo

    global	${array}_var
    upvar #0	${array}_var	avar

    set		frame		$ainfo(frame)
    if {![info exists ainfo(onceonly)]} {
	set	ainfo(onceonly)	1

	catch { $frame configure -background lightskyblue }

	pack append	$frame		[blt_graph $frame.graph \
						-font -*-courier-medium-r-normal-*-120-* \
						-plotrelief raised \
						-xtitle $xtitle \
						-ytitle $ytitle] \
					    {top expand}
	pack    after	$w.bot.polling	[button $w.bot.save \
						-command "FSBox {} {} \"snmptcl_save_graph $frame.graph\"" \
						-text Save...] \
					    {left padx 5 pady 5 expand}

	set	graph			$frame.graph
	set	ainfo(graph)		$graph
	catch { $graph configure -background lightskyblue }
	bind	$graph		<Any-Motion>	"%W crosshairs configure -position @%x,%y"
	bind	$graph		<B1-ButtonRelease> \
						"%W crosshairs toggle"

	if {[info exists ainfo(objs)]} {
	set symbols		[list line square circle plus diamond cross]
	} else {
	    set		symbols	[list square circle plus diamond cross line]

	    foreach	var	[lsort -command snmptcl_proc_oidcmp \
				     [array names avar]] {
		lappend		ainfo(vars)	$var
		set		full		[snmpinfo mibprop $var full]
		lappend		ainfo(objs)	"[lindex $full 0] [lindex $full 4]"
	    }
	}
	set	symindex	0
	set	symsize		[llength $symbols]

	set	i		0
	foreach obj		$ainfo(objs) {
	    $graph	element create	$i.0 \
					    -label $obj \
					    -linewidth 1 \
					    -symbol [lindex $symbols $symindex]

	    incr		i
	    if {[incr symindex] >= $symsize} {
		set		symindex	0
	    }
	}
	if {$i == 1} {
	    $graph	legend configure	-mapped false
	} else {
	    $graph	legend configure	-font -*-courier-medium-r-normal-*-120-* \
						-relief flat
	    catch { $graph legend configure -background lightskyblue }
	}

	$graph	xaxis configure	-font -*-courier-medium-r-normal-*-100-* \
				-min 0
						
	$graph	yaxis configure	-font -*-courier-medium-r-normal-*-100-* \
				-min 0
    } else {
	set	graph			$ainfo(graph)
    }

    set		iter		$ainfo(iter)
    set		i		0
    set		value	""
    foreach	var		$ainfo(vars) {
	if {[info exists avar($var)]} {
	    set		data		$avar($var)

	    set		syntax	[lindex $data 6]
	    if {[set object [lindex $data 7]] == ""} {
		lappend	avar($var)	\
				[set object [snmpinfo mibprop $var descriptor]]
	    }
	    if {[set start [lindex $data 8]] == ""} {
		lappend avar($var)	[set start 0]
	    }
	    if {[info exists snmptcl_graph_object($object)]} {
		set	value	[$snmptcl_graph_object($object) $var \
								 avar($var)]
	    } elseif {[info exists snmptcl_graph_syntax($syntax)]} {
		set	value	[$snmptcl_graph_syntax($syntax) $var \
								 avar($var)]
	    }
	}

	if {$value != ""} {
	    $graph	element	append	$i.$start [list $iter $value]
	} elseif {([lindex [$graph element configure $i.0 -xdata] 4] != "") \
		      && ([lindex [$graph element configure $i.$start -xdata] 4] != "")} {
	    set		symbol		[$graph element configure $i.0 -symbol]
	    incr	start
	    set		avar($var)	[lreplace $avar($var) 8 8 $start]
	    $graph	element create	$i.$start \
					    -label "" \
					    -linewidth 1 \
					    -symbol [lindex $symbol 4]
	}

	incr		i
    }
}

proc	snmptcl_save_graph	{w}	{
    global	fsBox

    $w postscript	$fsBox(path)/$fsBox(name)
}

#
# browse a table
#
#     pHandle - context procedure
#	  ask - ask about columns, regardless of A-C
#		-1 = never ask, 0 = ask if A-C info available, 1 = always ask
#	title - title for window
#	iname - name for icon
#	 cols - list of columnars to poll, each element may be a list,
#		if so, first is descriptor/OID and second is label
#    interval - polling interval (in seconds)
#     command - command to call on each iteration
#		(see snmptcl_polltable_callback for calling procedure)
#

proc	snmptcl_proc_browsetable \
	{pHandle ask title iname cols interval command} {
    global	snmptcl_poll_cookie snmptcl_browsing_icon

    if {[catch { $pHandle config active }]} {
	blt_bell	100
	snmptcl_proc_mkmsg \
			"first, select a management context"
	return
    }
    snmptcl_proc_mkmsg		""

    set		w		.$pHandle
    while {[winfo exists $w$snmptcl_poll_cookie]} {
	incr	snmptcl_poll_cookie
    }
    append	w		$snmptcl_poll_cookie
    incr	snmptcl_poll_cookie

    set		avail		0
    if {($ask >= 0) \
	    && ([llength [set resources [$pHandle config resources]]] > 0) \
	    && (![catch { snmpinfo agentprop [lindex $resources 0] sysUpTime \
			      support }])} {
	set	avail		1
	set	ask		1
    } else {
	set	avail		0
    }

    if {$ask < 1} {
	snmptcl_proc_polltable	$pHandle $w $title $iname \
				@$snmptcl_browsing_icon $cols $interval \
				$command
	return
    }

    toplevel	$w

    wm title	$w		$title
    wm iconname	$w		$iname
    wm iconbitmap \
		$w		@$snmptcl_browsing_icon

    frame	$w.frame	-borderwidth 10

    set		array		snmptcl_polltable_data[join [split $w "."] "_"]
    global	$array
    upvar #0	$array		varinfo
    catch { unset varinfo }

    set		cvars		""
    foreach	col	$cols {
	if {[llength $col] > 1} {
	    set		cvar	[lindex $col 0]
	} else {
	    set		cvar	$col
	}
	if {[catch { set descr [snmpinfo mibprop $cvar descriptor] } result]} {
	    destroy	$w
	    snmptcl_dialog	$title $result questhead 0 Abort
	    unset	varinfo
	    return
	}
	if {!$avail} {
	    set		varinfo($descr)	1
	} else {
	    set		varinfo($descr)	0
	    foreach	resource	$resources {
		if {[catch { set support [snmpinfo agentprop $resource $descr \
					      support] }]} {
		    continue
		}
		if {$support} {
		    set	    varinfo($descr)	1
		    break
		}
	    }
	}
	pack append \
		$w.frame	[checkbutton $w.frame.$descr \
					-relief flat \
					-text $descr \
					-variable "${array}($descr)"] \
				    {top frame w}
    }

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[button $w.bot.ok \
					-command "destroy $w; snmptcl_proc_browsetable_aux [list $pHandle $w $title $iname $cols $interval $command]" \
					-text OK] \
				    {left expand padx 5 pady 5} \
				[button $w.bot.cancel \
					-command "destroy $w; unset $array" \
					-text Cancel] \
				    {left expand padx 5 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}

proc	snmptcl_proc_browsetable_aux \
	{pHandle w title iname cols interval command} {
    global	snmptcl_browsing_icon

    set		array		snmptcl_polltable_data[join [split $w "."] "_"]
    global	$array
    upvar #0	$array		varinfo

    set		cvars		""
    foreach	col		$cols {
	if {[llength $col] > 1} {
	    set		cvar	[lindex $col 0]
	    set		clabel	[lindex $col 1]
	} else {
	    set		clabel	[set cvar $col]
	}
	set 	descr		[snmpinfo mibprop $cvar descriptor]
	if {$varinfo($descr)} {
	    lappend	cvars	[list $cvar $clabel]
	}
    }

    unset	varinfo

    if {[llength $cvars] <= 0} {
	snmptcl_dialog	$title "nothing selected" info 0 OK
	return
    }
    snmptcl_proc_polltable	$pHandle $w $title $iname \
				@$snmptcl_browsing_icon $cvars $interval \
				$command
}


#
# default browse table command
#

proc	snmptcl_browsetable_command	{instlabel instfmt w array}	{
    global	snmptcl_print_object snmptcl_print_syntax

    global	${array}_info
    upvar #0	${array}_info	ainfo

    global	${array}_var
    upvar #0	${array}_var	avar

    global	${array}_ncol
    upvar #0	${array}_ncol	ancol

    set		frame		$ainfo(frame)
    if {![info exists ainfo(onceonly)]} {
	set	ainfo(onceonly)	1

	catch { $frame configure -background lightskyblue }

	pack append	$frame		[scrollbar $frame.yscroll \
						-command "$frame.list yview" \
						-orient vertical] \
				    	    {right filly} \
					[scrollbar $frame.xscroll \
						-command "$frame.list xview" \
						-orient horizontal] \
				    	    {bottom fillx} \
					[listbox $frame.list \
						-font -*-courier-medium-r-normal-*-120-* \
						-xscrollcommand "$frame.xscroll set" \
						-yscrollcommand "$frame.yscroll set"] \
				    	    {top expand}

	set		table		$frame.list
	set		ainfo(table)	$table
	catch { $table configure -background lightskyblue }
#	bind		$table		<1> ""
#	bind		$table		<B1-Motion> ""
#	bind		$table		<Shift-1> ""
#	bind		$table		<Shift-B1-Motion> ""

	set		ncol		[list $instlabel]
	foreach		col		$ainfo(cols) {
	    lappend	ncol		$col
	}
	$table		insert		end [set ancol("") $ncol]

	set		ainfo(rows)	[set rows [list ""]]
    } else {
	set		table		$ainfo(table)
	set		rows		$ainfo(rows)
    }

    if {[catch { array size avar }]} {
	$table		configure	-geometry [expr [string length $ancol("")]+1]x1

	return
    }

    set		iter		$ainfo(iter)
    foreach	var		[array names avar] {
	set	data		$avar($var)
	set	col		[expr [lindex $data 7]+1]

	set	full		[snmpinfo mibprop $var full]
	set	inst		[lindex $full 3]
	set	row		[lsearch -exact $rows $inst]

	if {[lindex $data 0] != $iter} {
	    unset	avar($var)

	    if {$row < 0} {
		continue
	    }

	    set		ancol($inst)	[lreplace $ancol($inst) $col $col ""]

	    if {[string length [join [lrange $ancol($inst) 1 end] ""]] == 0} {
		$table	delete		$row
		set	rows		[lreplace $rows $row $row]
		unset	ancol($inst)
	    }

	    continue
	}

	if {$row < 0} {
	    set		row		[llength [lappend rows $inst]]

	    set		value		[set iinfo [lindex $full 4]]
	    if {$instfmt != ""} {
		set	value		[$instfmt $iinfo]
	    }
	    set		ncol		[list $value]
	    foreach	value		$ainfo(cols) {
		lappend	ncol		""
	    }

	    $table	insert		end [set ancol($inst) $ncol]
	}

	set		descr		[lindex $full 0]
	set		value		[lindex $data 4]
	set		syntax		[lindex $data 6]
	if {[info exists snmptcl_print_object($descr)]} {
	    set		value		[$snmptcl_print_object($descr) \
					     $var avar($var)]
	} elseif {[info exists snmptcl_print_syntax($syntax)]} {
	    set		value		[$snmptcl_print_syntax($syntax) \
					     $var avar($var)]
	}
	switch $syntax {
	    Integer32
		-
	    TimeTicks
		-
	    UInteger32 {
		if {[string match "*\[A-Za-z\]*" $value]} {
		    set		anchor		""
		} else {
		    set		anchor		" "
		}
	    }

	    Counter32
		-
	    Gauge32 {
		set		anchor		" "
	    }

	    default {
		set		anchor		""
	    }
	}

	set		ancol($inst)	[lreplace $ancol($inst) $col $col \
					     $anchor$value]
    }

    for {set ncols 0; set cols [expr [llength $ainfo(cols)]+1]} \
	    {$ncols < $cols} \
	    {incr ncols} {
	set	width($ncols)	0
    }

    foreach	inst		[array names ancol] {
	set	value		$ancol($inst)

	for {set i 0} {$i < $ncols} {incr i} {
	    set	    wid		[string length [set val [lindex $value $i]]]
	    if {[string index $val 0] == " "} {
		incr		wid		-1
	    }
	    if {$wid > $width($i)} {
		set		width($i)	$wid
	    }
	}
    }

    set		x		0
    for {set i 0} {$i < $ncols} {incr i} {
	incr	x		$width($i)
	incr	x		2
    }
    if {$x > 132} {
	set	x		132
    }
    if {[set y [llength $rows]] > 45} {
	set	y		45
    }
    $table	configure	-geometry ${x}x$y

    set		value		$ancol("")
    set		text		[format "%-$width(0)s" [lindex $value 0]]
    for {set i 1} {$i < $ncols} {incr i} {
	set	val		[lindex $value $i]
	set	diff		[expr $width($i)-[string length $val]]
	append	text		[format "  %*s%s%*s" [expr $diff/2] "" $val \
				     [expr $diff-($diff/2)] ""]
    }
    if {[$table get 0] != "$text"} {
	$table	delete		0
	$table	insert		0 $text
    }

    set		rows		[lsort -command snmptcl_proc_oidcmp $rows]
    set		ainfo(rows)	$rows
			     
# this VERY conveniently skips $ancol("")...
    foreach	inst		[array names ancol] {
	if {[set row [lsearch -exact $rows $inst]] < 0} {
	    continue
	}

	set	value		$ancol($inst)
	set	text		[format "%-$width(0)s" [lindex $value 0]]
	for {set i 1} {$i < $ncols} {incr i} {
	    set		val	[lindex $value $i]
	    if {[string index $val 0] == " "} {
		set	lor	""
		set	val	[string range $val 1 end]
	    } else {
		set	lor	"-"
	    }
	    append	text	[format "  %$lor$width($i)s" $val]
	}
	if {[$table get $row] != "$text"} {
	    $table	delete	$row
	    $table	insert	$row $text
	}
    }
}


#
# default pretty-printers
#

set	snmptcl_print_syntax(Counter32)		snmptcl_print_Counter32
set	snmptcl_print_syntax(Counter64)		snmptcl_print_Counter64
set	snmptcl_print_syntax(Integer32)		snmptcl_print_Integer32
set	snmptcl_print_syntax(NsapAddress)	snmptcl_print_Octets
set	snmptcl_print_syntax(ObjectID)		snmptcl_print_ObjectID
set	snmptcl_print_syntax(OctetString)	snmptcl_print_Octets
set	snmptcl_print_syntax(TimeTicks)		snmptcl_print_TimeTicks

proc	snmptcl_print_Counter32	{var data} {
    upvar	$data	info

    if {[set prevalue [lindex $info 1]] == ""} {
	return	""
    }
    set		value		[lindex $info 4]
    set		initvalue	[lindex $info 3]
    if {$prevalue > $value} {
	set	initvalue	[expr $initvalue-4294967295.0]
	set	info		[lreplace $info 3 3 $initvalue]
    }
    return	[expr $value-$initvalue]
}

proc	snmptcl_print_Counter64	{var data} {
    upvar	$data	info

    if {[set prevalue [lindex $info 1]] == ""} {
	return	""
    }
    set		value		[lindex $info 4]
    set		initvalue	[lindex $info 3]
    if {$prevalue > $value} {
	set	initvalue	[expr $initvalue-18446744073709551615.0]
	set	info		[lreplace $info 3 3 $initvalue]
    }
    return	[expr $value-$initvalue]
}

proc	snmptcl_print_Integer32	{var data} {
    upvar	$data	info

    set		value		[lindex $info 4]
    if {([set enums [snmpinfo mibprop $var enums]] != "") \
	    && ([set enum [lsearch -glob $enums "$value *"]] >= 0)} {
	return	[lindex [lindex $enums $enum] 1]
    }

    return $value
}

proc	snmptcl_print_ObjectID	{var data} {
    upvar	$data	info

    set		value		[lindex $info 4]
    if {$value == "0.0"} {
	return	"null"
    }
    return	[snmptcl_proc_oid2ode $value]
}

proc	snmptcl_print_Octets	{var data} {
    upvar	$data	info

    set		value		[lindex $info 4]
    return	[string range $value 2 end]
}

proc	snmptcl_print_TimeTicks	{var args}	{
    if {$args != ""} {
	upvar	$args		info

	set	value		[lindex $info 4]
    } else {
	set	value		$var
    }

    set         secs            [expr $value/100]
    set         value           [expr $value%100]
 
    set         mins            [expr $secs/60]
    set         secs            [expr $secs%60]
 
    set         hours           [expr $mins/60]
    set         mins            [expr $mins%60]
 
    set         days            [expr $hours/24]
    set         hours           [expr $hours%24]
 
    set         str             ""
 
    if {$days > 0} {
        append  str             [format "%d day" $days]
        if {$days != 1} {
	    append	str	"s"
	}
        append  str             ", "
    }
 
    if {($days > 0) || ($hours > 0)} {
        append  str             [format "%d hour" $hours]
        if {$hours != 1} {
	    append	str	"s"
	}
        append  str             ", "
    }
 
    if {($days > 0) || ($hours > 0) || ($mins > 0)} {
        append  str             [format "%d minute" $mins]
        if {$mins != 1} {
	    append	str	"s"
	}
        append  str             ", "
    }
 
    append      str             [format "%d" $secs]
    if {$value > 0} {
	append	str		[format ".%02d" $value]
    }
    append      str             " second"
    if {$secs != 1 || $value > 0} {
	append	str		"s"
    }
 
    return	$str
}

# these two must be explicitly referenced by snmptcl_print_object...
#	(maybe do DateAndTime later!)

proc	snmptcl_print_TimeStamp {var data}	{
    upvar	$data	info

    set		value		[lindex $info 4]
    set		uptime		[lindex $info 5]
    if {[set time [expr $uptime-$value]] < 0} {
	return	$value
    }
    return	"[snmptcl_print_TimeTicks $time] ago"
}

proc	snmptcl_print_TimeInterval {var data}	{
    upvar	$data	info

    return	[snmptcl_print_TimeStamp $var info]
}

#
# default graphers
#

set	snmptcl_graph_syntax(Counter32)		snmptcl_graph_Counter
set	snmptcl_graph_syntax(Counter64)		snmptcl_graph_Counter
set	snmptcl_graph_syntax(Gauge32)		snmptcl_graph_Number
set	snmptcl_graph_syntax(Integer32)		snmptcl_graph_Number
set	snmptcl_graph_syntax(TimeTicks)		snmptcl_graph_TimeTicks
set	snmptcl_graph_syntax(UInteger32)	snmptcl_graph_Number

proc	snmptcl_graph_Counter {var data} {
    upvar	$data	info

    if {[set prevalue [lindex $info 1]] == ""} {
	return	""
    }
    if {$prevalue > [set value [lindex $info 4]]} {
	set	diff		[expr $value+(4294967295.0-prevalue)]
    } else {
	set	diff		[expr $value-$prevalue]
    }
    if {[set delta [expr [lindex $info 5]-[lindex $info 2]]] == 0} {
	set	delta		1
    } else {
	set	delta		[expr $delta]
    }
    return	[expr ($diff*100)/$delta]
}

proc	snmptcl_graph_Number {var data} {
    upvar	$data	info

    return	[lindex $info 4]
}

proc	snmptcl_graph_TimeTicks {var data} {
    upvar	$data	info

    return	[expr [lindex $info 4]/100]
}


#
# set a scalar
#
#     pHandle - context procedure
#       title - title for window
#       iname - name for icon
#       descr - scalar variable to set
#      defval - default value for set
#

proc	snmptcl_proc_setscalar {pHandle title iname descr {defval {}}} {
    global	snmptcl_poll_cookie snmptcl_browsing_icon

    if {[catch { $pHandle config active }]} {
	blt_bell	100
	snmptcl_proc_mkmsg \
			"first, select a management context"
	return
    }
    snmptcl_proc_mkmsg		""

    if {[catch { set var [snmpinfo mibprop $descr value] } result]} {
	snmptcl_dialog	$title $result questhead 0 Abort
	return
    }

    if {$defval == ""} {
	global	snmptcl_info

	set	iHandle		[$pHandle get snmptcl_getscalar_callback \
				     [list [list $var.0]]]
	set	snmptcl_info($iHandle) \
				[list $title $iname $descr]
	return
    }

    set		w		.$pHandle
    while {[winfo exists $w$snmptcl_poll_cookie]} {
	incr	snmptcl_poll_cookie
    }
    append	w		$snmptcl_poll_cookie
    incr	snmptcl_poll_cookie

    set		srs		""
    if {[set resources [$pHandle config resources]] != ""} {
	foreach	resource	$resources {
	    catch { set srs [snmpinfo agentprop $resources $var WRITE-SYNTAX] }
	    if {$srs == ""} {
		catch { set srs [snmpinfo agentprop $resources $var SYNTAX] }
	    }
	    if {$srs != ""} {
		break
	    }
	}
    }

    if {([set enums [snmpinfo mibprop $var enums]] != "") \
	    && ([lindex $srs 0] == "enum")} {
	set	enew		""
	foreach	sr		$srs {
	    if {[set enum [lsearch -glob $enums "$sr *"]] >= 0} {
		lappend	enew	[list $sr [lindex [lindex $enums $enum] 1]]
	    } else {
		lappend	enew	[list $sr $sr]
	    }
	    set		enums	$enew
	}
	set	enums		[lsort -command snmptcl_sort_enums $enums]
    }

    set		range		[snmpinfo mibprop $var range]
    if {[lindex $srs 0] == "range"} {
	set	range		[lrange $srs 1 end]
    }

    set		array		snmptcl_setscalar_data[join [split $w "."] "_"]
    global	$array
    upvar #0	$array		varinfo
    catch { unset varinfo }

    set		varinfo(pHandle) \
				$pHandle
    set		varinfo(w)	$w
    set		varinfo(title)	$title
    set		varinfo(iname)	$iname
    set		varinfo(descr)	$descr
    set		varinfo(var)	$var
    set		varinfo(syntax)	[snmpinfo mibprop $var SYNTAX]
    set		varinfo(value)	$defval
    set		varinfo(enums)	$enums

    toplevel	$w

    wm title	$w		$title
    wm iconname	$w		$iname
    wm iconbitmap \
		$w		@$snmptcl_browsing_icon

    frame	$w.frame	-borderwidth 10
    if {$enums == ""} {
	label	$w.frame.value	-text "$descr: "
    } else {
	menubutton \
		$w.frame.value	-menu $w.frame.value.menu \
				-text "$descr: "
	menu	$w.frame.value.menu
    }
    pack append	$w.frame	$w.frame.value \
				    {left} \
				[entry $w.frame.en \
					-borderwidth 2 \
					-font -*courier-medium-r-normal-*-120-* \
					-relief sunken \
					-textvariable ${array}(value) \
					-width 60] \
				    {left}
    foreach	enum	$enums {
	$w.frame.value.menu add command \
				-command "set ${array}(value) [lindex $enum 0];snmptcl_proc_setscalar_aux $w ${array} 1" \
				-label [lindex $enum 1]
    }

    bind	$w.frame.en	<Return>	"$w.bot.confirm flash; $w.bot.confirm configure -state active; $w.bot.confirm invoke"
    bind	$w		<Any-Enter>	"focus $w.frame.en"

    frame	$w.bot		-borderwidth 5
    if {$range != ""} {
	pack append \
		$w.bot		[scale $w.bot.range \
					-command "set ${array}(value)" \
					-from [lindex $range 0] \
					-to [lindex $range 1] \
					-label $descr \
					-orient horizontal] \
				    {left expand padx 5 pady 5} \
    }
    pack append	$w.bot		[button	$w.bot.confirm \
					-command "snmptcl_proc_setscalar_aux $w ${array} 1" \
					-text Confirm] \
				    {left expand padx 5 pady 5} \
				[button	$w.bot.cancel \
					-command "snmptcl_proc_setscalar_aux $w ${array} 0" \
					-text Cancel] \
				    {left expand padx 5 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}

proc	snmptcl_proc_setscalar_aux	{w array deuxit} {
    global	snmptcl_info

    global	$array
    upvar #0	$array		varinfo
    
    if {!$deuxit} {
	destroy $w
	unset	varinfo
	return
    }

    set		value	$varinfo(value)
    foreach	enum	$varinfo(enums) {
	if {[lindex $enum 1] == "$value"} {
	    set	varinfo(value)	[lindex $enum 0]
	    break
	}
    }

    if {[catch { set iHandle [$varinfo(pHandle) set \
			          snmptcl_setscalar_callback \
				  [list [list $varinfo(var).0 \
					      $varinfo(syntax) \
					      $varinfo(value)]]] } result]} {
	snmptcl_dialog	$varinfo(title) $result questhead 0 Abort
	destroy		$w
	unset		varinfo
	return
    }
    set	    snmptcl_info($iHandle) \
			$array

    destroy	$w
}

proc	snmptcl_getscalar_callback {pHandle iHandle eStatus eIndex vBinds} {
    global	snmptcl_info

    set		data	$snmptcl_info($iHandle)
    unset	snmptcl_info($iHandle)

    if {($eStatus == "noError") \
	    && ([set syntax [lindex [set vb [lindex $vBinds 0]] 1]] \
		    != "exception") \
	    && ($syntax != "error")} {
	set	value	[lindex $vb 2]
	if {([set enums [snmpinfo mibprop [lindex $vb 0] enums]] != "") \
		&& ([set enum [lsearch -glob $enums "$value *"]] >= 0)} {
	    set		value	[lindex [lindex $enums $enum] 1]
	}
    } else {
	set	value	" "
    }

    snmptcl_proc_setscalar	$pHandle [lindex $data 0] [lindex $data 1] \
				[lindex $data 2] $value
}

proc	snmptcl_setscalar_callback {pHandle iHandle eStatus eIndex vBinds} {
    global	snmptcl_info snmptcl_browsing_icon

    set		data	$snmptcl_info($iHandle)
    unset	snmptcl_info($iHandle)
    set		array	[lindex $data 0]

    global	$array
    upvar #0	$array		varinfo

    switch -- $eStatus {
	noError {
	    set		w	$varinfo(w)
	    toplevel	$w

	    wm title	$w	$varinfo(title)
	    wm iconname $w	$varinfo(iname)
	    wm iconbitmap \
			$w	@$snmptcl_browsing_icon

	    frame	$w.frame \
				-borderwidth 10
	    pack append $w.frame \
				[message $w.frame.msg \
					-font -adobe-times-medium-r-normal-*-180-* \
					-text "successful set of $varinfo(descr)" \
					-width 3i] \
				    {right fill} \
				[label $w.frame.label \
					-bitmap info] \
				    {left padx 5 pady 5}
				
	    frame	$w.bot	-borderwidth 5
	    pack append	$w.bot	[button $w.bot.dismiss \
					-command "destroy $w" \
					-text Dismiss] \
				    {left expand padx 5 pady 5}


	    pack append	$w	$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}

	    unset	varinfo
	    return
	}

	genErr
	    -
	resourceUnavailable
	    -
	commitFailed
	    -
	undoFailed
	    -
	timeout {
	    set		offset	0
	}

	wrongValue {
	    set		offset	1
	}

	default {
	    set		offset	2
	}
    }

    switch [snmptcl_dialog "Agent Error for $varinfo(title)" \
	        $eStatus question $offset Retry "Try New Value" Abort] {
	0 {
	    set	    iHandle	[$varinfo(pHandle) set \
			             snmptcl_setscalar_callback \
				     [list [list $varinfo(var).0 \
						 $varinfo(syntax) \
			                         $varinfo(value)]]]
	    set	    snmptcl_info($iHandle) \
				$array
	}

	1 {
	    set		value	$varinfo(value)
	    foreach	enum	$varinfo(enums) {
		if {[lindex $enum 0] == "$value"} {
		    set	value	[lindex $enum 1]
		    break
		}
	    }
	    snmptcl_proc_setscalar \
				$varinfo(pHandle) $varinfo(title) \
				$varinfo(iname) $varinfo(descr) $value
	    unset	varinfo
	}

	default {
	    unset	varinfo
	}
    }
}


#
# sorts enums
#

proc	snmptcl_sort_enums	{a b} {
    if {[set a [lindex $a 0]] < [set b [lindex $b 0]]} {
	return		-1
    }
    if {$a > $b} {
	return		1
    }
    return		0
}
