#!/usr/local/bin/snmptcl -f

if {![info exists snmptcl_loaded(collecting.tcl)]} {
    if {[file readable collecting.tcl]} {
	source			collecting.tcl
    } else {
	source			$snmptcl_library/collecting.tcl
    }
}

if {[info exists snmptcl_loaded(mib2-collect.tcl)]} {
    return
}
set	snmptcl_loaded(mib2-collect.tcl)	1


#
# turn on "Applications" menu
#

.title.apps	configure	-state normal

.title.apps.menu \
		add cascade	-label "MIB-II collect" \
				-menu [set menu .title.apps.menu.collectmib2]
menu		$menu


#
# Interfaces table
#

$menu		add command	-command {snmptcl_proc_collectable $snmptcl_agent 1 "Interfaces table for $snmptcl_context" iftable {
{ ifDescr			Name			}
{ ifType			Type			}
{ ifMtu				MTU			}
{ ifSpeed			Speed			}
{ ifPhysAddress			"Physical Address"	}
{ ifAdminStatus			Admin			}
{ ifOperStatus			Oper			}
{ ifLastChange			Change			}
{ ifInOctets			IOcts			}
{ ifInUcastPkts			IUPkts			}
{ ifInNUcastPkts		INUPkts			}
{ ifInDiscards			IDiscs			}
{ ifInErrors			IErrs			}
{ ifInUnknownProtos	 	UnkProts		}
{ ifOutOctets			OOcts			}
{ ifOutUcastPkts	 	OUPkts			}
{ ifOutNUcastPkts	 	ONUPkts			}
{ ifOutDiscards			OutDiscs		}
{ ifOutErrors			OErrs			}
{ ifOutQLen			Queue			}
} 60 {snmptcl_collectable_command "#" {}}} \
				-label "Interfaces table"

set	snmptcl_print_object(ifSpeed)	snmptcl_print_ifSpeed

proc	snmptcl_print_ifSpeed	{var data} {
    upvar	$data	info

    set		value		[lindex $info 4]
    for {set i 0} {$value > 9999} {set value [expr $value>>10]} {
	incr	i
    }
    switch $i {
	0
	    -
	1
	    -
	2
	    -
	3
	    -
	4 {
	    return [format "%s%s" $value [lindex [list "" "K" "M" "G" "T"] $i]]
	}

	default {
	    return "huge"
	}
    }
}

$menu		add separator


#
# IP scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "IP scalars for $snmptcl_context" ip {
{ ipForwarding.0		ipForwarding		}
{ ipDefaultTTL.0		ipDefaultTTL		}
{ ipInReceives.0		ipInReceives		}
{ ipInHdrErrors.0		ipInHdrErrors		}
{ ipInAddrErrors.0		ipInAddrErrors		}
{ ipForwDatagrams.0		ipForwDatagrams		}
{ ipInUnknownProtos.0		ipInUnknownProtos	}
{ ipInDiscards.0		ipInDiscards		}
{ ipInDelivers.0		ipInDelivers		}
{ ipOutRequests.0		ipOutRequests		}
{ ipOutDiscards.0		ipOutDiscards		}
{ ipOutNoRoutes.0		ipOutNoRoutes		}
{ ipReasmTimeout.0		ipReasmTimeout		}
{ ipReasmReqds.0		ipReasmReqds		}
{ ipReasmOKs.0			ipReasmOKs		}
{ ipReasmFails.0		ipReasmFails		}
{ ipFragOKs.0			ipFragOKs		}
{ ipFragFails.0			ipFragFails		}
{ ipFragCreates.0		ipFragCreates		}
{ ipRoutingDiscards.0		ipRoutingDiscards	}
} 60} \
				-label "IP scalars"


#
# IP mappings
#

$menu		add command	-command {snmptcl_proc_collectable $snmptcl_agent 1 "IP mappings for $snmptcl_context" ipmap {
{ ipNetToMediaPhysAddress	"Physical Address"	}
{ ipNetToMediaType		Mapping			}
} 60 {snmptcl_collectable_command "if#   IP Address" snmptcl_print_ipmapping}} \
				-label "IP mappings"

proc	snmptcl_print_ipmapping	{inst}	{
    if {[llength $inst] != 2} {
	return		$inst
    }

    return	[format "%3s %-15s" [lindex $inst 0] [lindex $inst 1]]
}


#
# IP routes
#

$menu		add command	-command {snmptcl_proc_collectable $snmptcl_agent 1 "IP routes for $snmptcl_context" iproute {
{ ipRouteMask			Mask			}
{ ipRouteIfIndex		"if#"			}
{ ipRouteNextHop		Gateway			}
{ ipRouteType			Type			}
{ ipRouteProto			Proto			}
{ ipRouteAge			Age			}
{ ipRouteMetric1		M1			}
{ ipRouteMetric2		M2			}
{ ipRouteMetric3		M3			}
{ ipRouteMetric4		M4			}
{ ipRouteMetric5		M5			}
{ ipRouteInfo			c.f.			}
} 120 {snmptcl_collectable_command destination {}}} \
				-label "IP routes"

$menu		add separator


#
# ICMP input scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "ICMP input scalars for $snmptcl_context" icmpin {
{ icmpInMsgs.0			icmpInMsgs		}
{ icmpInErrors.0		icmpInErrors		}
{ icmpInDestUnreachs.0		icmpInDestUnreachs	}
{ icmpInTimeExcds.0		icmpInTimeExcds		}
{ icmpInParmProbs.0		icmpInParmProbs		}
{ icmpInSrcQuenchs.0		icmpInSrcQuenchs	}
{ icmpInRedirects.0		icmpInRedirects		}
{ icmpInEchos.0			icmpInEchos		}
{ icmpInEchoReps.0		icmpInEchoReps		}
{ icmpInTimestamps.0		icmpInTimestamps	}
{ icmpInTimestampReps.0		icmpInTimestampReps	}
{ icmpInAddrMasks.0		icmpInAddrMasks		}
{ icmpInAddrMaskReps.0		icmpInAddrMaskReps	}
} 60} \
				-label "ICMP input scalars"

#
# ICMP output scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "ICMP output scalars for $snmptcl_context" icmpout {
{ icmpOutMsgs.0			icmpOutMsgs		}
{ icmpOutErrors.0		icmpOutErrors		}
{ icmpOutDestUnreachs.0		icmpOutDestUnreachs	}
{ icmpOutTimeExcds.0		icmpOutTimeExcds	}
{ icmpOutParmProbs.0		icmpOutParmProbs	}
{ icmpOutSrcQuenchs.0		icmpOutSrcQuenchs	}
{ icmpOutRedirects.0		icmpOutRedirects	}
{ icmpOutEchos.0		icmpOutEchos		}
{ icmpOutEchoReps.0		icmpOutEchoReps		}
{ icmpOutTimestamps.0		icmpOutTimestamps	}
{ icmpOutTimestampReps.0	icmpOutTimestampReps	}
{ icmpOutAddrMasks.0		icmpOutAddrMasks	}
{ icmpOutAddrMaskReps.0		icmpOutAddrMaskReps	}
} 60} \
				-label "ICMP output scalars"

$menu		add separator


#
# TCP scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "TCP scalars for $snmptcl_context" tcp {
{ tcpRtoAlgorithm.0		tcpRtoAlgorithm		}
{ tcpRtoMin.0			tcpRtoMin		}
{ tcpRtoMax.0			tcpRtoMax		}
{ tcpMaxConn.0			tcpMaxConn		}
{ tcpActiveOpens.0		tcpActiveOpens		}
{ tcpPassiveOpens.0		tcpPassiveOpens		}
{ tcpAttemptFails.0		tcpAttemptFails		}
{ tcpEstabResets.0		tcpEstabResets		}
{ tcpCurrEstab.0		tcpCurrEstab		}
{ tcpInSegs.0			tcpInSegs		}
{ tcpOutSegs.0			tcpOutSegs		}
{ tcpRetransSegs.0		tcpRetransSegs		}
{ tcpInErrs.0			tcpInErrs		}
{ tcpOutRsts.0			tcpOutRsts		}
} 120} \
				-label "TCP scalars"

$menu		add separator


#
# UDP scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "UDP scalars for $snmptcl_context" udp {
{ udpInDatagrams.0		udpInDatagrams		}
{ udpNoPorts.0			udpNoPorts		}
{ udpInErrors.0			udpInErrors		}
{ udpOutDatagrams.0		udpOutDatagrams		}
} 120} \
				-label "UDP scalars"

$menu		add separator


#
# SNMPv1 input scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "SNMPv1 input scalars for $snmptcl_context" snmpv1in {
{ snmpInPkts.0			snmpInPkts		}
{ snmpInBadVersions.0		snmpInBadVersions	}
{ snmpInBadCommunityNames.0	snmpInBadCommunityNames	}
{ snmpInBadCommunityUses.0	snmpInBadCommunityUses	}
{ snmpInASNParseErrs.0		snmpInASNParseErrs	}
{ snmpInTooBigs.0		snmpInTooBigs		}
{ snmpInNoSuchNames.0		snmpInNoSuchNames	}
{ snmpInBadValues.0		snmpInBadValues		}
{ snmpInReadOnlys.0		snmpInReadOnlys		}
{ snmpInGenErrs.0		snmpInGenErrs		}
{ snmpInTotalReqVars.0		snmpInTotalReqVars	}
{ snmpInTotalSetVars.0		snmpInTotalSetVars	}
{ snmpInGetRequests.0		snmpInGetRequests	}
{ snmpInGetNexts.0		snmpInGetNexts		}
{ snmpInSetRequests.0		snmpInSetRequests	}
{ snmpInGetResponses.0		snmpInGetResponses	}
{ snmpInTraps.0			snmpInTraps		}
} 120} \
				-label "SNMPv1 input scalars"


#
# SNMPv1 output scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "SNMPv1 output scalars for $snmptcl_context" snmpv1out {
{ snmpOutPkts.0			snmpOutPkts		}
{ snmpOutTooBigs.0		snmpOutTooBigs		}
{ snmpOutNoSuchNames.0		snmpOutNoSuchNames	}
{ snmpOutBadValues.0		snmpOutBadValues	}
{ snmpOutGenErrs.0		snmpOutGenErrs		}
{ snmpOutGetRequests.0		snmpOutGetRequests	}
{ snmpOutGetNexts.0		snmpOutGetNexts		}
{ snmpOutSetRequests.0		snmpOutSetRequests	}
{ snmpOutGetResponses.0		snmpOutGetResponses	}
{ snmpOutTraps.0		snmpOutTraps		}
} 120} \
				-label "SNMPv1 output scalars"


#
# SNMPv2 scalars
#

$menu		add command	-command {snmptcl_proc_collect $snmptcl_agent 1 "SNMPv2 scalars for $snmptcl_context" snmpv2 {
{ snmpStatsPackets.0		snmpStatsPackets	}
{ snmpStats30Something.0	snmpStats30Something	}
{ snmpStatsEncodingErrors.0	snmpStatsEncodingErrors	}
{ snmpStatsUnknownDstParties.0	snmpStatsUnknownDstParties }
{ snmpStatsDstPartyMismatches.0	snmpStatsDstPartyMismatches }
{ snmpStatsUnknownSrcParties.0	snmpStatsUnknownSrcParties }
{ snmpStatsBadAuths.0		snmpStatsBadAuths	}
{ snmpStatsNotInLifetimes.0	snmpStatsNotInLifetimes	}
{ snmpStatsWrongDigestValues.0	snmpStatsWrongDigestValues }
{ snmpStatsUnknownContexts.0	snmpStatsUnknownContexts }
{ snmpStatsBadOperations.0	snmpStatsBadOperations	}
{ snmpStatsSilentDrops.0	snmpStatsSilentDrops	}
{ snmpV1BadCommunityNames.0	snmpV1BadCommunityNames	}
{ snmpV1BadCommunityUses.0	snmpV1BadCommunityUses	}
} 120} \
				-label "SNMPv2 scalars"

