#! /usr/local/bin/snmptcl -f

if {![info exists snmptcl_loaded]} {
    if {[file readable $snmptcl_myname.tcl]} {
	source			$snmptcl_myname.tcl
    } else {
	source			$snmptcl_library/$snmptcl_myname.tcl
    }
}
set	snmptcl_verbose	1
foreach p { oidcmp oid2ode } {
    rename	snmptcl_proc_$p	$p
}

if {![file readable [set snmptcl_testing_icon testing.icon]]} {
    set	snmptcl_testing_icon	$snmptcl_library/testing.icon
}

if {[info commands FSBox] == ""} {
    if {[file readable FSBox.tcl]} {
	source			FSBox.tcl
    } else {
	source			$snmptcl_library/FSBox.tcl
    }
}


#
# first, add some stuff to the top-level window
#

frame	.test		-borderwidth 2

pack append \
	.test		[label .test.system \
				-font -*-courier-medium-r-normal-*-120-* \
				-text ""] \
			    {top fillx} \
			[label .test.label \
				-font -*-courier-medium-r-normal-*-120-* \
				-relief sunken \
				-text [format "%-74.74s" \
				           [format "%-10s %-12s %s" \
				                "Test" "Status" "Remarks"]]] \
			    {top fillx} \
			[scrollbar .test.yscroll \
				-command ".test.list yview" \
				-orient vertical \
				-relief sunken] \
			    {right filly} \
			[scrollbar .test.xscroll \
				-command ".test.list xview" \
				-orient horizontal \
				-relief sunken] \
			    {bottom fillx} \
			[listbox .test.list \
				-font -*-courier-medium-r-normal-*-120-* \
				-geometry 70x24 \
				-relief sunken \
				-xscroll ".test.xscroll set" \
				-yscroll ".test.yscroll set"] \
			    {top expand}

catch { .test.label	configure	-background lightskyblue }

bind	.test.list	<Double-1>	"runTest"
bind	.test.list	<Shift-Double-1> "runTest"
bind	.test.list	<2>		{%W select from [%W nearest %y]}
bind	.test.list	<Double-2>	"resultsTest"
bind	.test.list	<3>		{%W select from [%W nearest %y]}
bind	.test.list	<Double-3>	"talkTest"
catch {	.test.list	configure	-background lightskyblue }

pack append \
	.		.test \
			    {bottom fill}


#
# list of tests we can run
#

#     identity	preqs
#	description
set     testset \
{
    { 1.1.1.1   {}
"1.1.1   P: Agent performs lexicographic ordering from beginning of OID tree\n        1.1.1.1 T: Issue a get-next from 0.0\n                E: Agent returns first variable"
    }
    { 1.1.1.2   {}
"1.1.1   P: Agent performs lexicographic ordering from beginning of OID tree\n        1.1.1.2 T: Issue a get-next from 1.0\n                E: Agent returns a variable, probably the first one"
    }
    { 1.1.2     {}
"1.1.2   P: Agent performs lexicographic ordering throughout entire OID tree\n        T: Get-next entire MIB from agent\n        E: Each variable should be lexi-greater than previous variable,\n           and should have correct SYNTAX and instance-ID\n        X: Agent returns noSuchName after final variable is returned\n        Y: Agent returns endOfMibView after final variable is returned"
    }
    { 1.1.3     { 1.1.2 }
"1.1.3   P: Agent performs lexicographic ordering with small OID\n        T: If a variable lexi-greater than 2.0 is returned in test 1.1.2,\n           issue a get-next from 2.0\n        E: Agent returns first variable greater than 2.0"
    }
    { 1.1.4     { 1.1.2 }
"1.1.4   P: Agent performs lexicographic ordering with non-variable arguments\n        T: For each variable returned in test 1.1.2:\n             remove a sub-identifier and issue a get-next\n             repeat until only two sub-identifiers remaining\n        E: Agent returns first variable greater than argument"
    }
    { 1.1.5     { 1.1.2 }
"1.1.5   P: Agent performs lexicographic ordering with large instance-IDs\n        T: For each variable returned in test 1.1.2:\n             replace each sub-identifier in the instance-ID, one at a time,\n             with 4G-1, and issue a get-next\n        E: Agent returns first variable greater than argument"
    }
    { 1.1.6     { 1.1.2 }
"1.1.6   P: Agent performs lexicographic ordering with long instance-IDs\n        T: For each variable returned in test 1.1.2, pad to 128 sub-identifiers\n           (using { 0 127 128 255 256 64K-1 64K 16M-1 16MG 4G-1 }) and issue a\n           get-next\n        E: Agent returns first variable greater than argument"
    }
    { 1.1.7.1   { 1.1.2 }
"1.1.7   P: Agent performs lexicographic ordering between unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n        1.1.7.1 T: issue a get-next of variables from unrelated tables\n                E: Agent returns first variable greater than each argument"
    }
    { 1.1.7.2   { 1.1.2 }
"1.1.7   P: Agent performs lexicographic ordering between unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n        1.1.7.2 T: issue a get-next of variables containing both scalar and\n                   columnar objects"
    }
    { 1.1.7.3   { 1.1.2 }
"1.1.7   P: Agent performs lexicographic ordering between unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n        1.1.7.3 T: issue a get-next of multiple columnar objects from the same\n                   table"
    }
    { 1.1.8     { 1.1.2 }
"1.1.8   P: Agent performs direct lookup consistently with lexicographic\n           ordering\n        T: For each variable returned in test 1.1.2: issue a get\n        E: Agent returns same variable (static table), or (for dynamic tables)\n        X: Agent returns noSuchName\n        Y: Agent returns noSuchInstance"
    }
    { 1.1.9     { 1.1.2 }
"1.1.9   P: Agent performs direct lookup with long instance-IDs\n        T: For each variable returned in test 1.1.2, pad to 128 sub-identifiers\n           (using { 0 127 128 255 256 64K-1 64K 16M-1 16M 4G-1 }) and issue a\n        get\n        X: Agent returns noSuchName\n        Y: Agent returns noSuchInstance"
    }
    { 1.1.10    { 1.1.2 }
"1.1.10  P: Agent returns correct error-status when writing read-only variables\n        T: For each read-only variable returned in test 1.1.2: issue a set\n          with identical value\n        X: Agent returns noSuchName\n        Y: Agent returns notWritable"
    }
    { 1.1.11.1  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.1 T: identical value\n                    E: Agent returns noError\n                       (Note: some objects with peculiar semantics, e.g.,\n                       tcpConnState, will return badValue/wrongValue instead\n                       of noError.)"
    }
    { 1.1.11.2  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.2 T: wrong tag\n                    X: Agent returns badValue\n                    Y: Agent returns wrongType"
    }
    { 1.1.11.3  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.3 T: length field not minimally-encoded\n                    E: Agent returns noError"
    }
    { 1.1.11.4  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.4 T: for integer-values, too small an integer\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.5  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.5 T: for integer-values, too large an integer\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.6  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.6 T: for enumerated integer-values, a non-enumeration\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.7  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.7 T: for NVT string-values, non-ASCII value\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.8  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.8 T: for NVT string-values, \\r not followed by \\n or \\0\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.9  { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n           1.1.11.9 T: for any string-values, constructed (not primitive) value\n                    E: Agent drops packet\n                    X: Agent returns noError or badValue\n                    Y: Agent returns noError or wrongEncoding"
    }
    { 1.1.11.10 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.10 T: for any string-values with SIZE, too few octets\n                    X: Agent returns badValue\n                    Y: Agent returns wrongValue"
    }
    { 1.1.11.11 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.11 T: for any string-values with SIZE, too many octets\n                    X: Agent returns badValue\n                    Y: Agent returns wrongLength"
    }
    { 1.1.11.12 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.12 T: an earlier variable-binding, being read-only\n                    X: Agent returns noSuchName\n                    Y: Agent returns notWritable"
    }
    { 1.1.11.13 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.13 T: a later variable-binding, being read-only\n                    X: Agent returns noSuchName\n                    Y: Agent returns notWritable"
    }
    { 1.1.11.14 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.14 T: an earlier variable-binding, being read-write, but with\n                       a bad value\n                    X: Agent returns badValue\n                    Y: Agent returns wrong*"
    }
    { 1.1.11.15 { 1.1.2 }
"1.1.11  P: Agent returns correct error-status when mis-writing variables\n        T: For each read-write variable returned in test 1.1.2, issue a\n           set with/using:\n          1.1.11.15 T: a later variable-binding, being read-write but, with a\n                       bad value\n                    X: Agent returns badValue\n                    Y: Agent returns wrong*"
    }
    { 1.1.12.1 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.1 T: the indefinite length form\n                    X: Agent drops packet, increments snmpInASNParseErrs, or\n                       returns noError"
    }
    { 1.1.12.2 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.2 T: a non-unitary version number\n                    X: Agent drops packet, increments snmpInBadVersions"
    }
    { 1.1.12.3 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.3 T: a zero-length community string\n                    X: Agent drops packet, increments snmpInBadNames"
    }
    { 1.1.12.4 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.4 T: the largest possible request-id\n                    X: Agent responds"
    }
    { 1.1.12.5 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.5 T: a non-zero error-status\n                    X: Agent responds"
    }
    { 1.1.12.6 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.6 T: a non-zero error-index\n                    X: Agent responds"
    }
    { 1.1.12.7 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.7 T: zero variable-bindings\n                    X: Agent responds"
    }
    { 1.1.12.8 { 1.1.2 }
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.8 T: (get, get-next only) variable-bindings with values other\n                       than ASN.1 NULL\n                    X: Agent responds"
    }
    { 1.1.12.9 { 1.1.2 }    
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n           1.1.12.9 T: (get, get-next only) a large number of\n                       variable-bindings so as to force a tooBig response\n                    Y: Agent returns tooBig"
    }
    { 1.1.12.10 { 1.1.2 }    
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n          1.1.12.10 T: the BER length is smaller than UDP data\n                    X: Agent drops packet, increments snmpInASNParseErrs, or\n                       returns noError"
    }
    { 1.1.12.11 { 1.1.2 }    
"1.1.12  P: V1 agent handles message boundary conditions correctly\n        T: For { get get-next set }, issue a request with/using:\n          1.1.12.11 T: the BER length is larger than UDP data\n                    X: Agent drops packet, increments snmpInASNParseErrs"
    }
    { 1.1.13.1 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.1 T: the indefinite length form\n                    Y: Agent drops packet, increments snmpStatsEncodingErrors,\n                       or returns noError"
    }
    { 1.1.13.2 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.2 T: an unknown dstParty\n                    Y: Agent drops packet, increments\n                      snmpStatsUnknownDstParties"
    }
    { 1.1.13.3 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.3 T: dstParty mismatch in privacy/authentication wrappers\n                    Y: Agent drops packet, increments\n                       snmpStatsDstPartyMismatches"
    }
    { 1.1.13.4 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.4 T: an unknown srcParty\n                    Y: Agent drops packet, increments\n                       snmpStatsUnknownSrcParties"
    }
    { 1.1.13.5 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.5 T: authInfo inappropriate srcParty\n                    Y: Agent drops packet, increments snmpStatsBadAuths"
    }
    { 1.1.13.6 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.6 T: an unknown context\n                    Y: Agent drops packet, increments snmpStatsUnknownContexts"
    }
    { 1.1.13.7 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.7 T: the largest possible request-id\n                    Y: Agent responds "
    }
    { 1.1.13.8 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.8 T: (get, get-next, set only) a non-zero error-status\n                    Y: Agent responds "
    }
    { 1.1.13.9 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n           1.1.13.9 T: (get, get-next, set only) a non-zero error-index\n                    Y: Agent responds "
    }
    { 1.1.13.10 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n          1.1.13.10 T: zero variable-bindings\n                    Y: Agent responds"
    }
    { 1.1.13.11 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n          1.1.13.11 T: (get, get-next only) variable-bindings with values other\n                       than ASN.1 NULL\n                    Y: Agent responds"
    }
    { 1.1.13.12 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n          1.1.13.12 T: (get, get-next only) a large number of\n                       variable-bindings so as to force a tooBig response\n                    Y: Agent returns tooBig"
    }
    { 1.1.13.13 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n          1.1.13.13 T: the BER length is smaller than UDP data\n                    Y: Agent drops packet, increments snmpStatsEncodingErrors,\n                       or returns noError"
    }
    { 1.1.13.14 { 1.1.2 }
"1.1.13  P: V2 agent handles message boundary conditions correctly\n        T: For { get get-next bulk set }, issue a request with/using:\n          1.1.13.14 T: the BER length is larger than UDP data\n                    Y: Agent drops packet, increments snmpStatsEncodingErrors"
    }
    { 1.1.14 { }
"1.1.14  P: Agent handles protocol boundary conditions correctly\n        T: For { response trap xxx }: send each to the agent\n        X: Agent drops packet, on xxx, increments snmpInASNParseErrs\n        Y: Agent drops packet, on response, trap, increments\n           snmpStatsBadOperations, on xxx, increments snmpStatsEncodingErrors"
    }
    { 1.1.15 { 1.1.2 }
"1.1.15  P: Agent correlates requests to responses\n        T: For { 2 3 ... 16 }: issue that many get-nexts for different\n           variables returned in test 1.1.2\n        E: Any responses should be returned with the right request-ids"
    }
    { 1.1.16.1 { 1.1.2 }
"1.1.16  P: Agent performs direct lookup with unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n       1.1.16.1 T: issue a get of variables from unrelated tables\n                E: Agent returns same variables (static table), or (for dynamic\n                   (tables)\n                X: Agent returns noSuchName\n                Y: Agent returns noSuchInstance"
    }
    { 1.1.16.2 { 1.1.2 }
"1.1.16  P: Agent performs direct lookup with unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n       1.1.16.2 T: issue a get of variables containing both scalar and\n                   columnar objects\n                E: Agent returns same variables (static table), or (for dynamic\n                   (tables)\n                X: Agent returns noSuchName\n                Y: Agent returns noSuchInstance"
    }
    { 1.1.16.3 { 1.1.2 }
"1.1.16  P: Agent performs direct lookup with unrelated objects\n        T: For a subset of the variables returned in test 1.1.2:\n       1.1.16.3 T: issue a get of multiple columnar objects from the same table\n                E: Agent returns same variables (static table), or (for dynamic\n                   (tables)\n                X: Agent returns noSuchName\n                Y: Agent returns noSuchInstance"
    }
    { 1.2.1 {}
"1.2.1   P: Agent implements objects that it claims to, according to its\n           AGENT-CAPABILITIES statement (RFC 1444)\n        T: For each object in each group claimed by agent, retrieve any\n           instance\n        E: At least one instance should be returned\n           (Note: tables that are implemented, but empty, will need to be\n                  primed prior to running this test.)"
    }
    { 1.2.2 { 1.1.2 }
"1.2.2   P: Agent doesn't implements any objects not listed in its\n           AGENT-CAPABILITIES statement\n        T: For each variable returned in test 1.1.2, verify that the agent's\n           AGENT-CAPABILITIES statement claims support\n        E: All variables returned in test 1.1.2 should be present\n"
    }
}

set	tests	{}
foreach	set	$testset {
    lappend	tests		[set test [lindex $set 0]]
    .test.list	insert		end \
				[format "%-10s %-12s" $test "never run"]
}
set	testsInProgress		 0
set	updateInterval		25


#
# redefine select context hook
#

set	sysDescr		[snmpinfo mibprop sysDescr     value]
set	sysUpTime		[snmpinfo mibprop sysUpTime    value]
set	snmpTrapOID		[snmpinfo mibprop snmpTrapOID  value]
set	coldStart		[snmpinfo mibprop coldStart    value]
set	partyMIB		[snmpinfo mibprop partyMIB     value]
if {[catch { set tcpConnState [snmpinfo mibprop tcpConnState value] }]} {
    set	tcpConnState		""
}

rename	snmptcl_callback_select_contexts_aux	snmptcl_callback_select_contexts_aux_orig
proc	snmptcl_callback_select_contexts_aux	{{agent {}}} {
    global	tests snmptcl_context testsInProgress testsFailed
    global	sysDescr

    .test.system	configure	-text ""
    catch { .test.system configure	-background bisque1 }

    .test.list		delete		0 end
    foreach	test	$tests {
        .test.list	insert	end	[format "%-10s %-12s" \
					    $test "never run"]

	catch { destroy .failed[join [split $test "."] "_"] }
    }
    set		testsInProgress		0
    catch { unset testsFailed }

    snmptcl_callback_select_contexts_aux_orig	[set agent agent]
    if {[catch { $agent config active }]} {
	return
    }
    .test.system	configure	-text [format "%-74.74s" \
						  "Context: $snmptcl_context"]

    $agent bulk		SelectContextCallback [list [list $sysDescr]]
}

proc	SelectContextCallback	{pHandle iHandle eStatus eIndex vBinds} {
    global	sysDescr snmptcl_agent snmptcl_context

    if {$eStatus != "noError"} {
	return
    }

    set		descr			""
    foreach	vb	$vBinds {
	if {([set syntax [lindex $vb 1]] == "exception") \
		|| ($syntax == "error")} {
	    continue
	}

	if {"$sysDescr.0" == [lindex $vb 0]} {
	    set		descr		[lindex $vb 2]
	    break
	}
    }
    set		descr			[lindex [split $descr "\n"] 0]
    set		objectid		[lindex [$snmptcl_agent config \
						     resources] 0]

    if {($descr != "") && ($objectid != "")} {
	set		banner		[format "Agent: %s (%s)" \
					    $descr $objectid]
    } elseif {$descr != ""} {
	set		banner		"sysDescr: $descr"
    } elseif {$objectid != ""} {
	set		banner		"sysObjectID: $objectid"
    } else {
	return
    }
    if {[set len [string length [set tail " \[$snmptcl_context\]"]]] > 70} {
	set	len	[string length [set tail ""]]
    }
    set		len	[expr 73-$len]
    set		fmt	[format "%%-%d.%ds%%s" $len $len]
    if {[string length $banner] > $len} {
	set	blen	[expr $len-3]
	set	banner	[format [format "%%-%d.%ds..." $blen $blen] $banner]
    }

    .test.system	configure	-text [format $fmt $banner $tail]
    catch { .test.system configure	-background bisque2 }
}


#
# run a test, etc.
#

proc	runTest		{} {
    global	tests testset snmptcl_agent

    if {[catch { $snmptcl_agent config active }]} {
	.test.list	select clear
	blt_bell	100
	snmptcl_proc_mkmsg \
			"first, select a management context"
	return
    }

    foreach	i	[.test.list curselection] {
	set	set	[lindex $testset $i]
	if {[statusTest [set test [lindex $set 0]]] == "in progress"} {
	    snmptcl_proc_mkmsg	"test $test is already in progress..."
	} elseif {[set waits [checkTest $test [lindex $set 1]]] != ""} {
	    updateTest	$test	"waiting"	$waits
	} else {
	    runTestAux	$test
	}
    }
}

proc	runTestAux	{test} {
    global	testsInProgress testsFailed

    snmptcl_proc_mkmsg	""

    catch { destroy .failed[join [split $test "."] "_"] }

    catch { unset testsFailed($test) }
    if {[catch { test$test $test } result]} {
	snmptcl_proc_mkmsg	$result
	updateTest	$test	"bye"	$result
	checkTests
    } else {
	incr		testsInProgress
    }
}

proc	checkTests	{} {
    global	testset	testsInProgress

    foreach	set	$testset {
	if {[statusTest [set test [lindex $set 0]]] == "waiting"} {
	    if {[set waits [checkTest $test [lindex $set 1]]] != ""} {
		updateTest	$test	"waiting"	$waits
	    } elseif {$testsInProgress > 0} {
		updateTest	$test	"waiting"
	    } else {
		runTestAux	$test
	    }
	}
    }
}

proc	checkTest	{test preqs} {
    set		waits	""
    foreach	preq	$preqs {
	if {[statusTest $preq] != "passed"} {
	    lappend	waits	$preq
	}
    }
    if {[llength $waits] > 0} {
	if {[llength $waits] > 1} {
	    set		waits	"for tests $waits"
	} else {
	    set		waits	"for test $waits"
	}
    }

    return	$waits
}

proc	statusTest	{test} {
    global	tests

    return	[string trim \
		     [string range [.test.list get \
				        [lsearch -exact $tests $test]] 11 22]]
}

proc	updateTest	{test status {reason {}}} {
    global	tests

    if {[set i [lsearch -exact $tests $test]] < 0} {
	snmptcl_dialog		Error "Internal Error: lost test $test" \
				questhead 0 Abort
	return
    }

    .test.list	delete		$i
    .test.list	insert		$i \
				[format "%-10s %-12s %s" $test $status $reason]

    updateMsg
}

proc	updateMsg	{} {
    global	 testsInProgress

    if {$testsInProgress == 0} {
	update
   }
}

proc	passedTest	{test {diag {}}} {
    global	testsInProgress testsFailed testset

    snmptcl_proc_mkmsg	""

    if {[winfo exists [set w .failed[join [split $test "."] "_"]]]} {
	set	status	failed
	set	diag	"see window \"Failure for Test $test\""
    } elseif {[info exists testsFailed($test)]} {
	set	status	failed
    } else {
	set	status	passed
    }

    incr	testsInProgress	-1
    updateTest	$test	$status	$diag
    checkTests
}

proc	failedTest	{test diag text} {
    global	testsInProgress

    if {$text != ""} {
	losingTest	$test $diag $text
    }
    if {($diag == "") \
	    && [winfo exists [set w .failed[join [split $test "."] "_"]]]} {
	set	diag	"see window \"Failure for Test $test\""
    }
    if {$diag != ""} {
	set	msg	"test $test: $diag"
    } else {
	set	msg	"test $test: failed"
    }
    snmptcl_proc_mkmsg	$msg

    incr	testsInProgress	-1
    updateTest	$test	failed	$diag
    checkTests
}

proc	losingTest	{test diag text} {
    global	testsFailed

    set		testsFailed($test)	""
    set		w		.failed[join [split $test "."] "_"]
    if {$diag != ""} {
	set	text		"Remarks:      $diag\n$text"
    }
    talkWindow	$w		$test "Failure for Test $test" $text
}

proc	resultsTest	{} {
    global	tests

    snmptcl_proc_mkmsg	""
    foreach	i	[.test.list curselection] {
	set	test	[lindex $tests $i]
	if {[winfo exists [set w .failed[join [split $test "."] "_"]]]} {
	    wm	iconify		$w
	    wm	deiconify	$w
	}
    }
}

proc	talkTest	{} {
    global	tests testset

    snmptcl_proc_mkmsg	""
    foreach	i	[.test.list curselection] {
	set	test	[lindex $tests $i]

	if {[winfo exists [set w .info[join [split $test "."] "_"]]]} {
	    wm	iconify		$w
	    wm	deiconify	$w
	    continue
	}
    
	set	text		[lindex [lindex $testset $i] 2]
	if {[llength [set preqs [lindex [lindex $testset $i] 1]]] > 0} {
	    append	text	"\n\nPrequisites:\n  $preqs"
	}

	talkWindow	$w	$test "Information about Test $test" $text
    }
}

proc	talkWindow	{w test title text} {
    global	tests testset snmptcl_testing_icon

    if {[winfo exists $w]} {
	$w.frame.list	insert	end	"-------"

	foreach	i		[split $text "\n"] {
	    $w.frame.list	insert	end	$i
	}

	return
    } elseif {($test != "") \
		 && ([string first .failed $w] == 0) \
		 && ([set i [lsearch -exact $tests $test]] >= 0)} {
	append		info	[lindex [lindex $testset $i] 2] "\n\n\n" $text
	set		text	$info
    }

    toplevel	$w
    wm title	$w		$title
    if {$test != ""} {
	wm iconname	$w	$test
    }
    wm iconbitmap \
		$w		@$snmptcl_testing_icon

    frame $w.frame		-borderwidth 10

    pack append \
	$w.frame	[scrollbar $w.frame.yscroll \
				-command "$w.frame.list yview" \
				-orient vertical \
				-relief sunken] \
			    {right filly} \
			[scrollbar $w.frame.xscroll \
				-command "$w.frame.list xview" \
				-orient horizontal \
				-relief sunken] \
			    {bottom fillx} \
			[listbox $w.frame.list \
				-font -*-courier-medium-r-normal-*-120-* \
				-geometry 80x20 \
				-relief sunken \
				-xscroll "$w.frame.xscroll set" \
				-yscroll "$w.frame.yscroll set"] \
			    {top expand}

    bind	$w.frame.list	<1>		""
    bind	$w.frame.list	<B1-Motion>	""
    bind	$w.frame.list	<Shift-1>	""
    bind	$w.frame.list	<Shift-B1-Motion> ""
    bind	$w.frame.list	<Double-1>	""

    foreach	i		[split $text "\n"] {
	$w.frame.list	insert	end	$i
    }

    catch { $w.frame.list configure -background lightgoldenrod }

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[button	$w.bot.save \
					-command "FSBox {} {} \"saveWindow $w $test\"" \
					-text Save...] \
				    {left expand padx 5 pady 5} \
				[button	$w.bot.dismiss \
					-command "destroy $w" \
					-text Dismiss] \
				    {left expand padx 5 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}

proc	saveWindow	{w {test {}}} {
    global	fsBox

    if {[catch { set fd [open $fsBox(path)/$fsBox(name) "a"] } result]} {
	snmptcl_proc_mkmsg	$result
	return
    }

    if {$test != ""} {
	set	msg	"\n\n\nInformation about Test $test"
    } else {
	set	msg	"\n\n\nInformation"
    }
    set		wid	[expr "(80-[string length $msg])/2"]
    puts	$fd	[format [format "%*.*s%%s\n\n%%s\n\n" $wid $wid ""] \
			    $msg [lindex [.test.system configure -text] 4]]

    for {set i 0; set j [$w.frame.list size]} {$i < $j} {incr i} {
	puts	$fd	[$w.frame.list get $i]
    }

    if {[catch { close $fd } result]} {
	snmptcl_proc_mkmsg	$result
    }
}

proc	rsp2text	{eStatus eIndex vBinds} {
    set		text	"Error-Status: $eStatus"
    switch -- $eStatus {
	destroyed
	    -
	failed
	    -
	timeout {
	    return	$text
	}
	    
	default {
	    append	text	"\nError-Index:  $eIndex\nBindings:"
	}
    }

    foreach	vb	$vBinds {
	append	text	"\n    [oid2ode [lindex $vb 0]] [lindex $vb 2]"
	append	text	"\n        [lindex $vb 0] [lindex $vb 1]"
    }

    return	$text
}


#
# 
#

proc	test1.1.1.1	{test} { test1.1.1 $test 0.0 }
proc	test1.1.1.2	{test} { test1.1.1 $test 1.0 }

proc	test1.1.1	{test start} {
    global	info snmptcl_agent

    updateTest	$test		"in progress"
    set		iHandle		[$snmptcl_agent next callback1.1.1 \
				     [list [list $start]]]
    set		info($iHandle)	[list $test]
}

proc	callback1.1.1	{pHandle iHandle eStatus eIndex vBinds} {
    global	info

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb	[lindex $vBinds 0]
	    if {[lindex $vb 1] == "exception"} {
		failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	}

	default {
	    failedTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    passedTest		$test "agent returned [lindex [lindex $vBinds 0] 0]"
}


#
#
#

proc	test1.1.2	{test} {
    global	variables scalars info snmptcl_agent

    snmptcl_proc_mkmsg	"this may take a while..."
    updateMsg

    set		variables	""
    catch { unset scalars }
    catch { unset columns }
    catch { unset readonly }
    catch { unset readwrite }

    foreach	resource	[$snmptcl_agent config resources] {
	catch { snmpinfo agentprop $resource sysUpTime support }
    }

    updateTest	$test		"in progress"
    set		iHandle		[$snmptcl_agent next callback1.1.2 {{0.0}}]
    set		info($iHandle)	[list $test 0.0]
}

proc	callback1.1.2	{pHandle iHandle eStatus eIndex vBinds} {
    global	info variables snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[lindex $vb 1] == "exception"} {
		switch [lindex $vb 2] {
		    noSuchObject
		        -
		    noSuchInstance
			-
		    default {
			failedTest	$test "agent returned [lindex $vb 2]" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }

		    endOfMibView {
			if {[llength $variables] <= 0} {
			    failedTest	$test "no variables!"
			} else {
			    passedTest	$test \
					"[llength $variables] variables walked"
			}
		    }
		}

		return
	    }
	    if {[oidcmp $last [set var [lindex $vb 0]]] >= 0} {
		failedTest	$test "lexicographic error, $last >= $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    vBind1.1.2	$test	$pHandle $eStatus $eIndex $vBinds
	}

	noSuchName {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		failedTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    } elseif {$eIndex != 1} {
		failedTest	$test "expecting error-index=1, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
	    } elseif {[llength $variables] <= 0} {
		failedTest	$test "no variables!"
	    } else {
		passedTest	$test "[llength $variables] variables walked"
	    }
	    return
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    lappend	variables	$var

    if {$snmptcl_verbose} {
	if {([set l [llength $variables]] % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $l variables walked, continuing..."
	    update
	}
    }

    set	    iHandle		[$pHandle next callback1.1.2 \
					     [list [list $var]]]
    set	    info($iHandle)	[list $test $var]
}

proc	vBind1.1.2	{test pHandle eStatus eIndex vBinds} {
    set		vb		[lindex $vBinds 0]
    set		var		[lindex $vb 0]

    if {([catch { set SYNTAX [snmpinfo mibprop $var SYNTAX] }]) \
	    || ($SYNTAX == "")} {
	talkWindow \
		.note		"" "Note Well" "unknown object: $var"
	return	-1
    }

    set		full		[snmpinfo mibprop $var full]
    if {([lindex $full 4] == "") && ([lindex $full 3] != "0")} {
	losingTest	$test	"agent returned malformed instance-ID" \
				[rsp2text $eStatus $eIndex $vBinds]
    }

    set		syntax		[lindex $vb 1]
    set		value		[lindex $vb 2]

    set		asyntax		[agentSyntax $pHandle $var [lindex $full 2]]

    switch $SYNTAX {
	Integer32 {
	    if {$syntax != $SYNTAX} {
		losingTest	$test	"agent returned wrong syntax (expecting $SYNTAX)" \
					[rsp2text $eStatus $eIndex $vBinds]
		return		0
	    }
	    if {[set enums [snmpinfo mibprop $var enums]] != ""} {
		if {[lindex $asyntax 0] == "enums"} {
		    if {[lsearch -exact [lrange $asyntax 1 end] $value] >= 0} {
			return	1
		    }
		} else {
		    foreach enum	$enums {
		        if {[lindex $enum 0] == $value} {
			    return	1
		        }
		    }
		}
		losingTest	$test	"agent returned non-enumeration" \
					[rsp2text $eStatus $eIndex $vBinds]
		return	0
	    }
	    if {[set range [snmpinfo mibprop $var range]] != ""} {
		if {[lindex $asyntax 0] == "range"} {
		    set	    range	[lrange $asyntax 1 end]
		}
		if {($value >= [lindex $range 0]) \
		        && ($value <= [lindex $range 1])} {
		    return	1
		}
		losingTest	$test	"agent returned out of range" \
					[rsp2text $eStatus $eIndex $vBinds]
		return	0
	    }
	    return	1
	}

	OctetString
	    -
	DisplayString {
	    if {($syntax != "OctetString") && ($syntax != "DisplayString")} {
		losingTest	$test	"agent returned wrong syntax (expecting $SYNTAX)" \
					[rsp2text $eStatus $eIndex $vBinds]
		return		0
	    }
	    if {[set size [snmpinfo mibprop $var size]] == ""} {
		return		1
	    }
	    if {[lindex $asyntax 0] == "size"} {
		set    size	[lrange $asyntax 1 end]
	    }
	    set		len	[string length $value]
	    if {$syntax == "OctetString"} {
		set	len	[expr ($len-1)/3]
	    }
	    if {($len >= [lindex $size 0]) \
		    && ($len <= [lindex $size 1])} {
		return	1
	    }
	    losingTest	$test	"agent returned wrong size" \
				[rsp2text $eStatus $eIndex $vBinds]
	}

	NetworkAddress {
	    if {$syntax == "IpAddress"} {
		return	1
	    }
	}

	default {
	    if {$syntax == $SYNTAX} {
		return	1
	    }
	}
    }
    losingTest	$test	"agent returned wrong syntax (expecting $SYNTAX)" \
			[rsp2text $eStatus $eIndex $vBinds]
    return	0
}

proc	agentSyntax	{pHandle var {value {}}} {
    if {$value == ""} {
	set	    SYNTAX	WRITE-SYNTAX
	set	    value	[snmpinfo mibprop $var value]
    } else {
	set	    SYNTAX	SYNTAX
    }

    foreach	resource	[$pHandle config resources] {
	if {[catch { set asyntax [snmpinfo agentprop $resource $value \
		    		      $SYNTAX] }]} {
	    continue
	}
	if {($asyntax == "") \
		&& ($SYNTAX == "WRITE-SYNTAX")} {
	    set		asyntax	    [snmpinfo agentprop $resource $value \
				         SYNTAX]
	    continue
	}
	if {$asyntax != ""} {
	    return	$asyntax
	}
    }

    return	""
}


#
#
#

proc	test1.1.3	{test} {
    global	variables info snmptcl_agent

    if {[lsearch $variables "2.*"] < 0} {
	error		"no variables lexi-greater than 2.0"
    }

    updateTest	$test		"in progress"
    set		iHandle		[$snmptcl_agent next callback1.1.4 {{2.0}}]
    set		info($iHandle)	[list $test 2.0 [llength $variables] 1]
}

proc	test1.1.4	{test} {
    global	variables info snmptcl_agent

    set		l	0
    foreach	var	$variables {
	incr	l	[expr [llength [split $var "."]]-2]
    }

    if {[set i [string last . [set var [lindex $variables 0]]]] < 0} {
	snmptcl_dialog		Error "Internal Error: lost dot in $var" \
				questhead 0 Abort
	return
    }
    set		last		[string range $var 0 [expr $i-1]]

    updateTest	$test		"in progress" "estimating $l iterations"
    set		iHandle		[$snmptcl_agent next callback1.1.4 \
				     [list [list $last]]]
    set		info($iHandle)	[list $test $last 0 1]
}

proc	callback1.1.4	{pHandle iHandle eStatus eIndex vBinds} {
    global	info variables snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		index		[lindex $data 2]
    set		iters		[lindex $data 3]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb		[lindex $vBinds 0]
	    if {[lindex $vb 1] == "exception"} {
		failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    if {[oidcmp $last [set var [lindex $vb 0]]] >= 0} {
		losingTest	$test \
				"lexicographic error, $last >= $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    set		var		$last
    if {[set i [string last . $var]] < 2} {
	if {[incr index] >= [llength $variables]} {
	    if {$test == "1.1.3"} {
		passedTest	$test "agent returned [lindex $vb 0]"
	    } else {
		passedTest	$test "$iters iterations"
	    }
	    return
	}
	set	var		[lindex $variables $index]
	if {[set i [string last . $var]] < 0} {
	    snmptcl_dialog	Error "Internal Error: lost dot in $var" \
				questhead 0 Abort
	    return
	}
    }
    set		last		[string range $var 0 [expr $i-1]]

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set		iHandle		[$pHandle next callback1.1.4 \
				     [list [list $last]]]
    set		info($iHandle)	[list $test $last $index $iters]
}


#
#
#

set	oidsuffixes	{	     0
				 127 128
				 255 256
			       65535 65536
			    16777215 16777216
			  4294967295          }

proc	test1.1.5	{test} {
    global	variables oidsuffixes info snmptcl_agent

    if {[catch { set value [snmpinfo mibprop [set var [lindex $variables 0]] \
			        value] }]} {
	set	index2		1
    } else {
	set	index2		[expr [llength [split $var "."]]-[llength [split $value "."]]]
    }
    set		offset		[expr [llength [set oid [split $var "."]]]-$index2]
    set		var		[join [lreplace $oid $offset $offset \
					    4294967295] "."]
    incr	index2		-1

    updateTest	$test		"in progress" \
				"at least [llength $variables] iterations, probably a lot more"
    set		iHandle		[$snmptcl_agent next callback1.1.5 \
				     [list [list $var]]]
    set		info($iHandle)	[list $test $var 0 $index2 1]
}

proc	test1.1.6	{test} {
    global	variables oidsuffixes info snmptcl_agent

    set		var		[lindex $variables 0]
    set		index2		0
    for {set nsubID [llength [split $var "."]]} \
	    {$nsubID < 128} \
	    {incr nsubID} {
	if {[incr index2] >= [llength $oidsuffixes]} {
	    set	index2		0
	}
	append	var		"." [lindex $oidsuffixes $index2]
    }

    updateTest	$test		"in progress" \
				"estimating [llength $variables] iterations"
    set		iHandle		[$snmptcl_agent next callback1.1.5 \
				     [list [list $var]]]
    set		info($iHandle)	[list $test $var 0 $index2 1]
}

proc	callback1.1.5	{pHandle iHandle eStatus eIndex vBinds} {
    global	info variables oidsuffixes snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		index1		[lindex $data 2]
    set		index2		[lindex $data 3]
    set		iters		[lindex $data 4]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb		[lindex $vBinds 0]
	    set		var		[lindex $vb 0]
	    if {[lindex $vb 1] == "exception"} {
		if {[lindex $vb 2] != "endOfMibView"} {
		    failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
	    } elseif {[oidcmp $last $var] >= 0} {
		losingTest	$test \
				"lexicographic error, $last >= $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    }
	}

	noSuchName {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb		[lindex $vBinds 0]
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		failedTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    } elseif {$eIndex != 1} {
		failedTest	$test "expecting error-index=1, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    switch $test {
	1.1.5 {
	    if {$index2 <= 0} {
		if {[incr index1] >= [llength $variables]} {
		    passedTest		$test "$iters iterations"
		    return
		}
		set		var	[lindex $variables $index1]
		if {[catch { set value [snmpinfo mibprop $var value] }]} {
		    set		index2	1
		} else {
		    set		index2	[expr [llength [split $var "."]]-[llength [split $value "."]]]
		}
	    } else {
		set		var	[lindex $variables $index1]
	    }
	    set		offset		[expr [llength [set oid [split $var "."]]]-$index2]
	    set		var		[join [lreplace $oid $offset $offset \
						    4294967295] "."]
	    incr	index2		-1
	}

	1.1.6 {
	    if {[incr index1] >= [llength $variables]} {
		passedTest	$test "$iters iterations"
		return
	    }
	    set		var		[lindex $variables $index1]
	    for {set nsubID [llength [split $var "."]]} \
		    {$nsubID < 128} \
		    {incr nsubID} {
		if {[incr index2] >= [llength $oidsuffixes]} {
		    set	index2		0
		}
		append		var	"." [lindex $oidsuffixes $index2]
	    }
	}
    }

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set		iHandle		[$pHandle next callback1.1.5 \
				     [list [list $var]]]
    set		info($iHandle)	[list $test $var $index1 $index2 $iters]
}


#
#
#

proc	test1.1.7.1	{test} { test1.1.7 $test }
proc	test1.1.7.2	{test} { test1.1.7 $test }
proc	test1.1.16.1	{test} { test1.1.7 $test }
proc	test1.1.16.2	{test} { test1.1.7 $test }

proc	test1.1.7	{test} {
    global	scalars columns info snmptcl_agent

    if {![info exists scalars]} {
	readyScalarsAndColumnars
    }
    if {(![info exists columns]) \
	    || ([llength [set tables [array names columns]]] <= 0)} {
	error		"no columnar variables"
    }

    switch $test {
	"1.1.7.1"
	    -
	"1.1.7.2" {
	    set		op	next
	}

	"1.1.16.1"
	    -
	"1.1.16.2" {
	    set		op	get
	}
    }

    set		i	0
    set		j	0
    if {($test == "1.1.7.2") || ($test == "1.1.16.2")} {
	incr	i
	incr	j
	set	k	0
	if {[set kmax [llength $scalars]] <= 0} {
	    error	"no scalar variables"
	}
    }

    set		pairs	{}
    foreach	table	[lrange $tables 0 [expr [llength $tables]-2]] {
	set	imax	[llength $columns($table)]

	set	others	[lrange $tables \
			    [expr [lsearch -exact $tables $table]+1] end]
	foreach	other	$others {
	    set		jmax	[llength $columns($other)]

	    if {[incr i] >= $imax} {
		set	i	0
	    }
	    set		pair	[lindex $columns($table) $i]

	    if {($test == "1.1.7.2") || ($test == "1.1.16.2")} {
		if {[incr k] >= $kmax} {
		    set	k	0
		}
		lappend	pair	[lindex $scalars $k]
	    }

	    if {[incr j] >= $jmax} {
		set	j	0
	    }
	    lappend	pair	[lindex $columns($other) $j]

	    if {($test == "1.1.7.2") || ($test == "1.1.16.2")} {
		if {[incr k] >= $kmax} {
		    set	k	0
		}
		lappend	pair	[lindex $scalars $k]
	    }

	    lappend	pairs	$pair
	}
    }

    set		pair		[lindex $pairs 0]
    set		pairs		[lrange $pairs 1 end]

    updateTest	$test		"in progress" \
				"estimating [expr [llength $pairs]+1] iterations"
    set		iHandle		[$snmptcl_agent $op callback1.1.7 $pair]
    set		info($iHandle)	[list $test $op $pair $pairs 1]
}

proc	test1.1.16.3	{test} { test1.1.7.3 $test }

proc	test1.1.7.3	{test} {
    global	columns info snmptcl_agent

    if {![info exists columns]} {
	readyScalarsAndColumnars
    }
    if {(![info exists columns]) \
	    || ([llength [set tables [array names columns]]] <= 0)} {
	error		"no columnar variables"
    }

    switch $test {
	"1.1.7.3" {
	    set		op	next
	}

	"1.1.16.3" {
	    set		op	get
	}
    }

    set		i	0
    foreach	table	$tables {
	lappend	tables	$table
    }
    foreach	table	$tables {
	lappend	tables	$table
    }
    foreach	table	$tables {
	set	pair	""
	set	imax	[llength $columns($table)]

	for {set j 0} {$j < 3} {incr j} {
	    if {[incr i] >= $imax} {
		set	i	0
	    }
	    lappend	pair	[lindex $columns($table) $i]
	}

	lappend		pairs	$pair
    }

    set		pair		[lindex $pairs 0]
    set		pairs		[lrange $pairs 1 end]

    updateTest	$test		"in progress" \
				"estimating [expr [llength $pairs]+1] iterations"
    set		iHandle		[$snmptcl_agent $op callback1.1.7 $pair]
    set		info($iHandle)	[list $test $op $pair $pairs 1]
}

proc	callback1.1.7	{pHandle iHandle eStatus eIndex vBinds} {
    global	info snmptcl_verbose snmptcl_agent

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		op		[lindex $data 1]
    set		pair		[lindex $data 2]
    set		pairs		[lindex $data 3]
    set		iters		[lindex $data 4]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != [set k [llength $pair]]} {
		failedTest	$test "expecting $k VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    while {[incr l -1] >= 0} {
		set	vb		[lindex $vBinds $l]
		if {([lindex $vb 1] == "exception") \
			&& (($op != "get") \
				|| ([lindex $vb 2] != "noSuchInstance"))} {
		    failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
		if {$op == "get"} {
		    if {[oidcmp [set last [lindex $pair $l]] \
			        [set var [lindex $vb 0]]] != 0} {
		        losingTest	$test \
				    "lexicographic error, $last != $var" \
				    [rsp2text $eStatus $eIndex $vBinds]
		    }
		} else {
		    if {[oidcmp [set last [lindex $pair $l]] \
			        [set var [lindex $vb 0]]] >= 0} {
		        losingTest	$test \
				    "lexicographic error, $last >= $var" \
				    [rsp2text $eStatus $eIndex $vBinds]
		    }
		}
	    }
	}

	noSuchName {
	    if {$op != "get"} {
		failedTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		return
	    }

	    if {[set l [llength $vBinds]] != [set k [llength $pair]]} {
		failedTest	$test "expecting $k VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    if {($eIndex < 1) || ($eIndex > $l)} {
		    failedTest	$test \
				"expecting error-index=1..$l, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
	    }
	    while {[incr l -1] >= 0} {
		set	vb		[lindex $vBinds $l]
		if {[oidcmp [set last [lindex $pair $l]] \
			    [set var [lindex $vb 0]]] != 0} {
		    losingTest	$test \
				"lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {[llength $pairs] == 0} {
	passedTest	$test "$iters iterations"
	return
    }

    set		pair	[lindex $pairs 0]
    set		pairs	[lrange $pairs 1 end]

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set		iHandle		[$snmptcl_agent $op callback1.1.7 $pair]
    set		info($iHandle)	[list $test $op $pair $pairs $iters]
}

proc	readyScalarsAndColumnars {} {
    global	scalars variables columns

    snmptcl_proc_mkmsg	"computing scalar and columnar variables..."
    updateMsg

    set		scalars	""
    catch { unset columns }

    foreach	var 	$variables {
	if {([catch { set full [snmpinfo mibprop $var full] } result])
		|| ([lindex $full 3] == "")} {
	    continue
	}

	set	oid	[split [lindex $full 2] "."]
	set	parent	[join [lrange $oid 0 [expr [llength $oid]-2]] "."]
	if {[catch { set parent [snmpinfo mibprop $parent descriptor] } \
		result]} {
	    talkWindow	.note	"" "Note Well" "unknown object: $parent"
	    continue
	}

	set	column	[lindex $full 2]
	if {[string match "*Entry" $parent]} {
	    if {![info exists columns($parent)]} {
		set	columns($parent)	$column
	    } elseif {[lsearch -exact $columns($parent) $column] < 0} {
		lappend	columns($parent)	$column
	    }

	} else {
	    lappend	scalars			$column
	}
    }

    if {[info exists columns]} {
        foreach	table	[array names columns] {
	    set	columns($table)	[lsort -command sortable $columns($table)]
	}
    }
    set		scalars		[lsort -command sortable $scalars]

    snmptcl_proc_mkmsg	""
    updateMsg
}

proc	sortable	{a b} {
    return	[string compare [snmpinfo mibprop $a descriptor] \
				[snmpinfo mibprop $b descriptor]]
}


#
#
#

proc	test1.1.8	{test} {
    global	variables info snmptcl_agent

    set		var		[lindex $variables 0]

    updateTest	$test		"in progress" \
				"estimating [llength $variables] iterations"
    set		iHandle		[$snmptcl_agent get callback1.1.8 \
				     [list [list $var]]]
    set		info($iHandle)	[list $test $var 0 0 0 1]
}

proc	test1.1.9	{test} {
    global	variables oidsuffixes info snmptcl_agent

    set		var		[lindex $variables 0]
    set		index2		0
    for {set nsubID [llength [split $var "."]]} \
	    {$nsubID < 128} \
	    {incr nsubID} {
	if {[incr index2] >= [llength $oidsuffixes]} {
	    set	index2		0
	}
	append	var		"." [lindex $oidsuffixes $index2]
    }

    updateTest	$test		"in progress" \
				"estimating [llength $variables] iterations"
    set		iHandle		[$snmptcl_agent get callback1.1.8 \
				     [list [list $var]]]
    set		info($iHandle)	[list $test $var 0 $index2 0 1]
}

proc	callback1.1.8	{pHandle iHandle eStatus eIndex vBinds} {
    global	info variables oidsuffixes snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		index1		[lindex $data 2]
    set		index2		[lindex $data 3]
    set		fails		[lindex $data 4]
    set		iters		[lindex $data 5]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[set syntax [lindex $vb 1]] == "exception"} {
		if {[lindex $vb 2] != "noSuchInstance"} {
		    failedTest		$test "agent returned [lindex $vb 2]" \
					[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
		incr	fails
	    }
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		failedTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    if {($test == "1.1.9") && ($syntax != "exception")} {
		losingTest	$test "shouldn't have been able to retrieve $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    }
	}

	noSuchName {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb		[lindex $vBinds 0]
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		losingTest	$test \
				"lexicographic error, $last >= $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    } elseif {$eIndex != 1} {
		failedTest	$test "expecting error-index=1, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    incr	fails
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {[incr index1] >= [llength $variables]} {
	switch $test {
	    "1.1.8" {
		if {$fails} {
		    passedTest	$test \
				"$fails retrieval failures (dynamic tables?)"
		} else {
		    passedTest	$test "$iters iterations"
		}
	    }

	    "1.1.9" {
		if {$iters != $fails} {
		    passedTest	$test \
				"only $fails retrieval failures (hmmm...)"
		} else {
		    passedTest	$test "$iters iterations"
		}
	    }
	}

	return
    }
    set		var	[lindex $variables $index1]
    if {$test == "1.1.9"} {
	for {set nsubID [llength [split $var "."]]} \
		{$nsubID < 128} \
		{incr nsubID} {
	    if {[incr index2] >= [llength $oidsuffixes]} {
		set	index2		0
	    }
	    append	var		"." [lindex $oidsuffixes $index2]
	}
    }

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set	    iHandle		[$pHandle get callback1.1.8 [list [list $var]]]
    set	    info($iHandle)	[list $test $var $index1 $index2 $fails $iters]
}


#
#
#

proc	test1.1.10	{test} {
    global	readonly readwrite tcpConnState info snmptcl_agent

    set		varlist		""
    switch $test {
	"1.1.10" {
	    if {![info exists readonly]} {
		readyReadOnlyVars
	    }
	    if {[set l [llength [set varlist $readonly]]] <= 0} {
		error		"no read-only variables"
	    }
	}

	"1.1.11.1" {
	    if {![info exists readwrite]} {
		readyReadWriteVars
	    }
	    set		varlist		$readwrite
	    if {$tcpConnState != ""} {
		while {[set l [lsearch -glob $varlist "$tcpConnState.*"]] \
			    >= 0} {
		    set	varlist		[lreplace $varlist $l $l]
		}
	    }
	    if {[set l [llength $varlist]] <= 0} {
		error		"no read-write variables"
	    }
	}

	"1.1.11.9" {
	    if {![info exists readwrite]} {
		readyReadWriteVars
	    }
	    foreach	var	$readwrite {
		if {[catch { set syntax [snmpinfo mibprop $var SYNTAX] }]} {
		    continue
		}
		switch $syntax {
		    "OctetString"
			-
		    "DisplayString" {
		    }

		    default {
			continue
		    }
		}
		lappend	varlist	$var
	    }
	    if {[set l [llength $varlist]] <= 0} {
		error		"no read-write string-valued variables"
	    }
	}

	"1.1.11.12"
	    -
	"1.1.11.13" {
	    if {![info exists readonly]} {
		readyReadOnlyVars
	    }
	    if {[llength $readonly] <= 0} {
		error		"no read-only variables"
	    }
	    if {![info exists readwrite]} {
		readyReadWriteVars
	    }
	    if {[set l [llength [set varlist $readwrite]]] <= 0} {
		error		"no read-write variables"
	    }
	}

	default {
	    if {![info exists readwrite]} {
		readyReadWriteVars
	    }
	    if {[set l [llength [set varlist $readwrite]]] <= 0} {
		error		"no read-write variables"
	    }
	}
    }

    set		var		[lindex $varlist 0]
    set		varlist		[lrange $varlist 1 end]
    switch $test {
	"1.1.11.9" {
	    set		vb		[list [list $var]]
	    snmptcl_proc_mkmsg	"this involves timeouts, so may take a while..."
	    updateMsg
	}

	"1.1.11.12" {
	    set		vb		[list [list [lindex $readonly 0]] \
					      [list $var]]
	}

	"1.1.11.13" {
	    set		vb		[list [list $var] \
					      [list [lindex $readonly 0]]]
	}

	"1.1.11.14" {
	    set		vb		[list [list [lindex $readwrite 0]] \
					      [list $var]]
	}

	"1.1.11.15" {
	    set		vb		[list [list $var] \
					      [list [lindex $readwrite 0]]]
	}

	default {
	    set		vb		[list [list $var]]
	}
    }

    updateTest	$test		"in progress" \
				"estimating [expr $l*2] iterations"
    set		iHandle		[$snmptcl_agent get callback1.1.10a $vb]
    set		info($iHandle)	[list $test $var $varlist 0 1]
}

proc	callback1.1.10a	{pHandle iHandle eStatus eIndex vBinds} {
    global	info readonly readwrite snmptcl_verbose syntaxes nsyntax snmptcl_debug

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		varlist		[lindex $data 2]
    set		fails		[lindex $data 3]
    set		iters		[lindex $data 4]

    set		failed		0
    switch $test {
	"1.1.11.12" {
	    set		nvb		2
	    set		last2		$last
	    set		last		[lindex $readonly 0]
	}

	"1.1.11.13" {
	    set		nvb		2
	    set		last2		[lindex $readonly 0]
	}

	"1.1.11.14" {
	    set		nvb		2
	    set		last2		$last
	    set		last		[lindex $readwrite 0]
	}

	"1.1.11.15" {
	    set		nvb		2
	    set		last2		[lindex $readwrite 0]
	}

	default {
	    set		nvb		1
	}
    }
    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != $nvb} {
		failedTest	$test "expecting $nvb VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[set syntax [lindex $vb 1]] == "unknown"} {
		failedTest	$test "agent returned unknown syntax" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    } elseif {$syntax == "exception"} {
		if {[lindex $vb 2] != "noSuchInstance"} {
		    failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
		incr	fails
		set	failed		1
	    }
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		failedTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    if {$l > 1} {
		set	    vb2		[lindex $vBinds 1]
		if {[set syntax [lindex $vb2 1]] == "unknown"} {
		    failedTest	$test "agent returned unknown syntax" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
		} elseif {$syntax == "exception"} {
		    if {[lindex $vb2 2] != "noSuchInstance"} {
		        failedTest	$test \
					"agent returned [lindex $vb2 2]" \
					[rsp2text $eStatus $eIndex $vBinds]
			return
		    }
		    incr	fails
		    set		failed	1
		}
		if {[oidcmp $last2 [set var2 [lindex $vb2 0]]] != 0} {
		    failedTest	$test "lexicographic error, $last2 != $var2" \
				[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
		if {($test == "1.1.11.12") || ($test == "1.1.11.14")} {
		    set		var	$last2
		}
	    }
	}

	noSuchName {
	    if {[set l [llength $vBinds]] != $nvb} {
		failedTest	$test "expecting 1 VBs, got $nvb" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set		vb		[lindex $vBinds 0]
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		failedTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    } elseif {($eIndex < 1) || ($eIndex > $nvb)} {
		failedTest	$test \
				"expecting error-index <= $nvb, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    incr	fails
	    set		failed		1
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {$failed} {
	if {[llength $varlist] == 0} {
	    if {$fails} {
		passedTest	$test \
				"$fails retrieval failures (dynamic tables?)"
	    } else {
		passedTest	$test "$iters iterations"
	    }

	    return
	}

	set	var		[lindex $varlist 0]
	set	varlist		[lrange $varlist 1 end]
	switch $test {
	    "1.1.11.12" {
	        set	vb		[list [list [lindex $readonly 0]] \
					      [list $var]]
	    }

	    "1.1.11.13" {
		set	vb		[list [list $var] \
					      [list [lindex $readonly 0]]]
	    }

	    "1.1.11.14" {
	        set	vb		[list [list [lindex $readwrite 0]] \
					      [list $var]]
	    }

	    "1.1.11.15" {
		set	vb		[list [list $var] \
					      [list [lindex $readwrite 0]]]
	    }

	    default {
		set	vb		[list [list $var]]
	    }
	}

	incr	iters
	if {$snmptcl_verbose} {
	    if {($iters % 25) == 0} {
		snmptcl_proc_mkmsg \
				"$test: $iters iterations, continuing..."
		update
	    }
	}

	set	iHandle		[$pHandle get callback1.1.10a $vb]
	set	info($iHandle)	[list $test $var $varlist $fails $iters]

	return
    }

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set		timeout		""
    switch $test {
	"1.1.11.3" {
	    set		magic	"notminlen"
	}

	"1.1.11.9" {
	    set		magic	"constring"
	    set		timeout	"-timeout 5"
	}

	"1.1.11.14"
	    -
	"1.1.11.15" {
	    if {$test == "1.1.11.14"} {
		set		var2	[lindex [lindex $vBinds 0] 0]
		set		syn2	[lindex [lindex $vBinds 0] 1]
	    } else {
		set		var2	[lindex [lindex $vBinds 1] 0]
		set		syn2	[lindex [lindex $vBinds 1] 1]
	    }
	    switch $syn2 {
		"Integer32" {
		    set		vb2	[list $var2 DisplayString surprise!]
		}

		default {
		    set		vb2	[list $var2 Integer32 42]
		}
	    }
	    if {$test == "1.1.11.14"} {
		set	vBinds		[list $vb2 [lindex $vBinds 1]]
	    } else {
		set	vBinds		[list [lindex $vBinds 0] $vb2]
	    }
	    set	magic		"none"
	}

	default {
	    set	magic		"none"
	}
    }

    if {$snmptcl_debug} {
	set	vb		[lindex $vBinds 0]
	puts	stdout		"-> set [oid2ode [lindex $vb 0]] [lrange $vb 1 end] -magic $magic"
	if {[llength $vBinds] > 1} {
	    set		vb	[lindex $vBinds 1]
	    puts	stdout	"   set [oid2ode [lindex $vb 0]] [lrange $vb 1 end] -magic $magic"
	}
    }
    set		iHandle		[eval $pHandle set callback1.1.10b \
				    [list $vBinds] -magic $magic $timeout]
    set		info($iHandle)	[list $test $var $varlist $fails $iters]
}

proc	callback1.1.10b	{pHandle iHandle eStatus eIndex vBinds} {
    global	info readonly readwrite snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		varlist		[lindex $data 2]
    set		fails		[lindex $data 3]
    set		iters		[lindex $data 4]

    switch -- $eStatus {
	destroyed {
	    return
	}

	timeout {
	    switch $test {
		"1.1.11.9" {
		}

		default {
		    failedTest		$test "" \
					[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
	    }
	}

	noError {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.1"
		    -
		"1.1.11.3"
		    -
		"1.1.11.9" {
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	noSuchName
	    -
	notWritable {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.10" {
		}

		"1.1.11.12" {
		    if {$eIndex != 1} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		"1.1.11.13" {
		    if {$eIndex != 2} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	badValue {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.9" {
		}

		"1.1.11.14" {
		    if {$eIndex != 1} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		"1.1.11.15" {
		    if {$eIndex != 2} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	wrongType
	    -
	wrongLength {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.14" {
		    if {$eIndex != 1} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		"1.1.11.15" {
		    if {$eIndex != 2} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	wrongEncoding {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.9" {
		}

		"1.1.11.14" {
		    if {$eIndex != 1} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		"1.1.11.15" {
		    if {$eIndex != 2} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	wrongValue {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.9" {
		}

		"1.1.11.14" {
		    if {$eIndex != 1} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		"1.1.11.15" {
		    if {$eIndex != 2} {
			losingTest	$test \
					"agent returned wrong error-index" \
					[rsp2text $eStatus $eIndex $vBinds]
		    }
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {[llength $varlist] == 0} {
	passedTest	$test "$iters iterations"
	return
    }

    set		var		[lindex $varlist 0]
    set		varlist		[lrange $varlist 1 end]
    switch $test {
	"1.1.11.12" {
	    set		vb		[list [list [lindex $readonly 0]] \
					      [list $var]]
	}

	"1.1.11.13" {
	    set		vb		[list [list $var] \
					      [list [lindex $readonly 0]]]
	}

	"1.1.11.14" {
	    set		vb		[list [list [lindex $readwrite 0]] \
					      [list $var]]
	    }

	"1.1.11.15" {
	    set		vb		[list [list $var] \
					      [list [lindex $readwrite 0]]]
	}

	default {
	    set		vb		[list [list $var]]
	}
    }

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    set		iHandle		[$pHandle get callback1.1.10a $vb]
    set		info($iHandle)	[list $test $var $varlist $fails $iters]
}

proc	readyReadOnlyVars {}	{
    global	variables readonly snmptcl_agent

    snmptcl_proc_mkmsg	"computing read-only variables..."
    updateMsg

    set		resources	[$snmptcl_agent config resources]

    set		readonly	""
    foreach	var		$variables {
	if {[catch { set max [snmpinfo mibprop $var MAX-ACCESS] } result]} {
	    continue
	}

	if {![catch { set value [snmpinfo mibprop $var value] }]} {
	    foreach	resource	$resources {
		if {[catch { set access [snmpinfo agentprop $resource $value \
					     ACCESS] }]} {
		    continue
		}
		if {$access != ""} {
		    set		max	$access
		    break
		}
	    }
	}

	if {$max == "read-only"} {
	    lappend	readonly	$var
	}
    }

    snmptcl_proc_mkmsg	""
    updateMsg
}

proc	readyReadWriteVars {}	{
    global	variables readwrite partyMIB snmptcl_agent

    snmptcl_proc_mkmsg	"computing read-write variables..."
    updateMsg

    set		resources	[$snmptcl_agent config resources]

    set		readwrite	""
    foreach	var		$variables {
	if {[string first $partyMIB $var] == 0} {
	    continue
	}

	if {[catch { set max [snmpinfo mibprop $var MAX-ACCESS] } result]} {
	    continue
	}

	if {![catch { set value [snmpinfo mibprop $var value] }]} {
	    foreach	resource	$resources {
		if {[catch { set access [snmpinfo agentprop $resource $value \
					     ACCESS] }]} {
		    continue
		}
		if {$access != ""} {
		    set		max	$access
		    break
		}
	    }
	}

	if {($max == "read-write") || ($max == "read-create")} {
	    lappend	readwrite	$var
	}
    }

    snmptcl_proc_mkmsg	""
    updateMsg
}

proc	vBinds1.1.10b	{test last eStatus eIndex vBinds} {
    global	readonly readwrite

    switch $test {
	"1.1.11.12" {
	    set		nvb		2
	    set		last2		$last
	    set		last		[lindex $readonly 0]
	}

	"1.1.11.13" {
	    set		nvb		2
	    set		last2		[lindex $readonly 0]
	}

	"1.1.11.14" {
	    set		nvb		2
	    set		last2		$last
	    set		last		[lindex $readwrite 0]
	}

	"1.1.11.15" {
	    set		nvb		2
	    set		last2		[lindex $readwrite 0]
	}

	default {
	    set		nvb		1
	}
    }
    if {[set l [llength $vBinds]] != $nvb} {
	failedTest	$test "expecting $nvb VBs, got $l" \
			[rsp2text $eStatus $eIndex $vBinds]
	return	0
    }
    set	    vb			[lindex $vBinds 0]
    if {($eStatus == "noError") && ([lindex $vb 1] == "exception")} {
	failedTest	$test "agent returned $vb" \
			[rsp2text $eStatus $eIndex $vBinds]
	return	0
    }
    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
	failedTest	$test "lexicographic error, $last != $var" \
			[rsp2text $eStatus $eIndex $vBinds]
	return	0
    }
    if {$l > 1} {
	set	    vb2		[lindex $vBinds 1]
	if {($eStatus == "noError") && ([lindex $vb2 1] == "exception")} {
	    failedTest		$test "agent returned $vb2" \
				[rsp2text $eStatus $eIndex $vBinds]
	    return	0
	}
	if {[oidcmp $last2 [set var2 [lindex $vb2 0]]] != 0} {
	    failedTest		$test "lexicographic error, $last2 != $var2" \
				[rsp2text $eStatus $eIndex $vBinds]
	    return	0
	}
	if {($eStatus != "noError") \
		&& (($eIndex < 1) || ($eIndex > $nvb))} {
	    failedTest		$test \
				"expecting error-index <= $nvb, got $eIndex" \
				[rsp2text $eStatus $eIndex $vBinds]
	    return	0
	}
    }

    return	1
}


#
#
#

proc	test1.1.11.1	{test} { test1.1.10 $test }
proc	test1.1.11.2	{test} { test1.1.11 $test }
proc	test1.1.11.3	{test} { test1.1.10 $test }
proc	test1.1.11.4	{test} { test1.1.11 $test }
proc	test1.1.11.5	{test} { test1.1.11 $test }
proc	test1.1.11.6	{test} { test1.1.11 $test }
proc	test1.1.11.7	{test} { test1.1.11 $test }
proc	test1.1.11.8	{test} { test1.1.11 $test }
proc	test1.1.11.9	{test} { test1.1.10 $test }
proc	test1.1.11.10	{test} { test1.1.11 $test }
proc	test1.1.11.11	{test} { test1.1.11 $test }
proc	test1.1.11.12	{test} { test1.1.10 $test }
proc	test1.1.11.13	{test} { test1.1.10 $test }
proc	test1.1.11.14	{test} { test1.1.10 $test }
proc	test1.1.11.15	{test} { test1.1.10 $test }

proc	test1.1.11	{test} {
    global	readwrite info snmptcl_debug snmptcl_agent

    if {![info exists readwrite]} {
	readyReadWriteVars
    }
    set		varlist		""
    switch $test {
	"1.1.11.2" {
	    set	varlist		$readwrite
	}

	"1.1.11.4"
	    -
	"1.1.11.5" {
	    foreach	var	$readwrite {
		if {[catch { set syntax [snmpinfo mibprop $var SYNTAX] }]} {
		    continue
		}
		switch $syntax {
		    "Integer32" {
		    }

		    default {
			continue
		    }
		}
		if {([catch { set range [snmpinfo mibprop $var range] }]) \
			|| ($range == "")} {
		    continue
		}
		if {[lindex [set asyntax [agentSyntax $snmptcl_agent $var]] \
		         0] == "range"} {
		    set	    range	[lrange $asyntax 1 end]
		}
		if {$test == "1.1.11.4"} {
		    if {[lindex $range 0] < -2147483647} {
			continue
		    }
		} elseif {[lindex $range 1] > 2147483646} {
		    continue
		}
		lappend	varlist	$var
	    }
	}

	"1.1.11.6" {
	    foreach	var	$readwrite {
		if {([catch { set syntax [snmpinfo mibprop $var SYNTAX] }]) \
			|| ($syntax != "Integer32") \
			|| ([catch { set enums [snmpinfo mibprop $var enums] }]) \
			|| ($enums == "")} {
		    continue
		}
		lappend	varlist	$var
	    }
	}

	"1.1.11.7"
	    -
	"1.1.11.8" {
	    foreach	var	$readwrite {
		if {([catch { set syntax [snmpinfo mibprop $var SYNTAX] }]) \
			|| ($syntax != "DisplayString")} {
		    continue
		}
		lappend	varlist	$var
	    }
	}

	"1.1.11.10"
	    -
	"1.1.11.11" {
	    foreach	var	$readwrite {
		if {[catch { set syntax [snmpinfo mibprop $var SYNTAX] }]} {
		    continue
		}
		switch $syntax {
		    "OctetString"
			-
		    "DisplayString" {
		    }

		    default {
			continue
		    }
		}
		if {([catch { set size [snmpinfo mibprop $var size] }]) \
			|| ($size == "")} {
		    continue
		}
		if {[lindex [set asyntax [agentSyntax $snmptcl_agent $var]] \
		         0] == "size"} {
		    set	    size	[lrange $asyntax 1 end]
		}
		if {($test == "1.1.11.10") && ([lindex $size 0] < 1)} {
		    continue
		}
		lappend	varlist	$var
	    }
	}
    }
    if {[llength $varlist] <= 0} {
	error		"no applicable read-write variables"
    }

    set		var		[lindex $varlist 0]
    set		varlist		[lrange $varlist 1 end]
    if {[set vb [makeVBind1.1.11 $test var varlist]] == ""} {
	error		"expected additional read-write variables"
    }

    if {$snmptcl_debug} {
	puts	stdout		"-> set [oid2ode [lindex $vb 0]] [lrange $vb 1 end]"
    }
    updateTest	$test		"in progress" \
				"estimating [expr [llength $varlist]+1] iterations"
    set		iHandle		[$snmptcl_agent set callback1.1.11 [list $vb]]
    set		info($iHandle)	[list $test $var $varlist 0 1]
}

proc	callback1.1.11	{pHandle iHandle eStatus eIndex vBinds} {
    global	info snmptcl_verbose snmptcl_debug

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		varlist		[lindex $data 2]
    set		fails		[lindex $data 3]
    set		iters		[lindex $data 4]

    switch -- $eStatus {
	destroyed {
	    return
	}

# noCreation really shouldn't be here...
	noCreation
	    -
	badValue {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	}

	wrongType {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.2" {
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	wrongLength {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.10"
		    -
		"1.1.11.11" {
		}

		default {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}
	    }
	}

	wrongValue {
	    if {![vBinds1.1.10b $test $last $eStatus $eIndex $vBinds]} {
		return
	    }
	    switch $test {
		"1.1.11.2"
		    -
		"1.1.11.10"
		    -
		"1.1.11.11" {
		    losingTest	$test "" [rsp2text $eStatus $eIndex $vBinds]
		}

		default {
		}
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {[llength $varlist] == 0} {
	passedTest	$test "$iters iterations"
	return
    }

    set		var		[lindex $varlist 0]
    set		varlist		[lrange $varlist 1 end]
    if {[set vb [makeVBind1.1.11 $test var varlist]] == ""} {
	passedTest	$test "expected additional read-write variables"
	return
    }

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: $iters iterations, continuing..."
	    update
	}
    }

    if {$snmptcl_debug} {
	puts	stdout		"-> set [oid2ode [lindex $vb 0]] [lrange $vb 1 end]"
    }
    set		iHandle		[$pHandle set callback1.1.11 [list $vb]]
    set		info($iHandle)	[list $test $var $varlist $fails $iters]
}

proc	makeVBind1.1.11	{test VAR VARLIST} {
    global	snmptcl_agent

    upvar 1	$VAR		var
    upvar 1	$VARLIST	varlist

    switch $test {
	"1.1.11.2" {
	    while {[catch { set syntax [snmpinfo mibprop $var SYNTAX] }]} {
		if {[set var [lindex $varlist 0]] == ""} {
		    return	""
		}
		set	varlist	[lrange $varlist 1 end]
	    }
	    switch $syntax {
		"Integer32" {
		    return	[list $var DisplayString surprise!]
		}

		default {
		    return	[list $var Integer32 42]
		}
	    }
	}

	"1.1.11.4"
	    -
	"1.1.11.5" {
	    while {([catch { set range [snmpinfo mibprop $var range] }]) \
		        || ($range == "")} {
		if {[set var [lindex $varlist 0]] == ""} {
		    return	""
		}
		set	varlist	[lrange $varlist 1 end]
	    }
	    if {[lindex [set asyntax [agentSyntax $snmptcl_agent $var]] 0] \
		    == "range"} {
		set    range	[lrange $asyntax 1 end]
	    }
	    if {$test == "1.1.11.4"} {
		return		[list $var Integer32 \
				    [expr [lindex $range 0]-1]]
	    } else {
		return		[list $var Integer32 \
				    [expr [lindex $range 1]+1]]
	    }
	}

	"1.1.11.6" {
	    while {([catch { set enums [snmpinfo mibprop $var enums] }]) \
		        || ($enums == "")} {
		if {[set var [lindex $varlist 0]] == ""} {
		    return	""
		}
		set	varlist	[lrange $varlist 1 end]
	    }
	    if {[lindex [set asyntax [agentSyntax $snmptcl_agent $var]] 0] \
		    == "enums"} {
	        set	min	[lindex $asyntax 1]
		foreach	enum	[lrange $asyntax 2 end] {
		    if {$enum < $min} {
			set	min	$enum
		    }
		}
	    } else {
		set	min	[lindex [lindex $enums 0] 0]
	        foreach	enum	[lrange $enums 1 end] {
		    if {[set nin [lindex $enum 0]] < $min} {
		        set	min	$nin
		    }
		}
	    }

	    return		[list $var Integer32 [expr $min-1]]
	}

	"1.1.11.7" {
	    return	[list $var OctetString "0x77:6f:77:ff"]
	}

	"1.1.11.8" {
	    return	[list $var DisplayString "\rhello"]
	}

	"1.1.11.10"
	    -
	"1.1.11.11" {
	    while {([catch { set size [snmpinfo mibprop $var size] }]) \
		        || ($size == "")} {
		if {[set var [lindex $varlist 0]] == ""} {
		    return	""
		}
		set	varlist	[lrange $varlist 1 end]
	    }
	    if {[lindex [set asyntax [agentSyntax $snmptcl_agent $var]] 0] \
		    == "size"} {
		set    size	[lrange $asyntax 1 end]
	    }
	    if {$test == "1.1.11.10"} {
		set		len	[expr [lindex $size 0]-1]
	    } else {
		set		len	[expr [lindex $size 1]+1]
	    }
	    set		value	""
	    for {} {$len > 0} {incr len -1} {
		append	value	"A"
	    }
	    return	[list $var DisplayString $value]
	}
    }
}


#
#
#

proc	test1.1.12.1	{test} { test1.1.12 $test }
proc	test1.1.12.2	{test} { test1.1.12 $test }
proc	test1.1.12.3	{test} { test1.1.12 $test }
proc	test1.1.12.4	{test} { test1.1.12 $test }
proc	test1.1.12.5	{test} { test1.1.12 $test }
proc	test1.1.12.6	{test} { test1.1.12 $test }
proc	test1.1.12.7	{test} { test1.1.12 $test }
proc	test1.1.12.8	{test} { test1.1.12 $test }
proc	test1.1.12.9	{test} { test1.1.12 $test }
proc	test1.1.12.10	{test} { test1.1.12 $test }
proc	test1.1.12.11	{test} { test1.1.12 $test }

proc	test1.1.13.1	{test} { test1.1.12 $test }
proc	test1.1.13.2	{test} { test1.1.12 $test }
proc	test1.1.13.3	{test} { test1.1.12 $test }
proc	test1.1.13.4	{test} { test1.1.12 $test }
proc	test1.1.13.5	{test} { test1.1.12 $test }
proc	test1.1.13.6	{test} { test1.1.12 $test }
proc	test1.1.13.7	{test} { test1.1.12 $test }
proc	test1.1.13.8	{test} { test1.1.12 $test }
proc	test1.1.13.9	{test} { test1.1.12 $test }
proc	test1.1.13.10	{test} { test1.1.12 $test }
proc	test1.1.13.11	{test} { test1.1.12 $test }
proc	test1.1.13.12	{test} { test1.1.12 $test }
proc	test1.1.13.13	{test} { test1.1.12 $test }
proc	test1.1.13.14	{test} { test1.1.12 $test }
proc	test1.1.13.15	{test} { test1.1.12 $test }

proc	test1.1.14	{test} { test1.1.12 $test }

proc	test1.1.12	{test} {
    global	info readonly readwrite sysUpTime snmpTrapOID coldStart snmptcl_agent

    switch $test {
	"1.1.12.7"
	    -
	"1.1.13.10"
	    -
	"1.1.14" {
	    set		var	""
	}

	"1.1.12.8"
	    -
	"1.1.12.9"
	    -
	"1.1.13.11"
	    -
	"1.1.13.12" {
	    if {![info exists readonly]} {
		readyReadOnlyVars
	    }
	    if {[llength $readonly] <= 0} {
		error		"no read-only variables"
	    }
	    set		var	[lindex $readonly 0]
	}

	default {
	    if {![info exists readwrite]} {
		readyReadWriteVars
	    }
	    set		var	[lindex $readwrite 0]
	}
    }

    set		ops	[list next get set]
    set		vb	[list [list $var]]
    switch $test {
	"1.1.12.1"
	    -
	"1.1.13.1" {
	    set		magic	"indflen"
	}

	"1.1.12.2" {
	    set		magic	"nonunitvr"
	}

	"1.1.12.3" {
	    set		magic	"nullcomm"
	}

	"1.1.13.2" {
	    set		magic	"bdstparty"
	}

	"1.1.13.3" {
	    set		magic	"ndstmatch"
	}

	"1.1.13.4" {
	    set		magic	"bsrcparty"
	}

	"1.1.13.5" {
	    set		magic	"nsrcmatch"
	}

	"1.1.13.6" {
	    set		magic	"bcontext"
	}

	"1.1.12.4"
	    -
	"1.1.13.7" {
	    set		magic	"bigreqid"
	}

	"1.1.12.5"
	    -
	"1.1.13.8" {
	    set		magic	"nonzeroes"
	}

	"1.1.12.6"
	    -
	"1.1.13.9" {
	    set		magic	"nonzeroei"
	}

	"1.1.12.7"
	    -
	"1.1.13.10" {
	    set		magic	"none"
	    set		vb	""
	}

	"1.1.12.8"
	    -
	"1.1.13.11" {
	    set		ops	[list get next]
	    set		magic	"none"
	    set		vb	[list [list $var Integer32 0]]
	}

	"1.1.12.9"
	    -
	"1.1.13.12" {
	    set		ops	[list get next]
	    set		magic	"none"
	    for {set i 0} {$i < 30} {incr i} {
		lappend	vb	[list $var]
	    }
	}

	"1.1.12.10"
	    -
	"1.1.13.13" {
	    set		magic	"berlesudp"
	}

	"1.1.12.11"
	    -
	"1.1.13.14" {
	    set		magic	"udplesber"
	}

	"1.1.14" {
	    set		ops	[list resp trap xxx]
	    set		magic	"none"
	    set		vb	[list [list $sysUpTime.0 TimeTicks 0] \
				      [list $snmpTrapOID.0 ObjectID \
					    $coldStart]]
	}
    }
    switch $test {
	"1.1.13.1"
	    -
	"1.1.13.2"
	    -
	"1.1.13.3"
	    -
	"1.1.13.4"
	    -
	"1.1.13.5"
	    -
	"1.1.13.6"
	    -
	"1.1.13.7"
	    -
	"1.1.13.10"
	    -
	"1.1.13.13"
	    -
	"1.1.13.14" {
	    lappend	ops	"bulk"
	}
    }
    set		op	[lindex $ops 0]
    set		ops	[lrange $ops 1 end]

    switch $test {
	"1.1.12.1"
	    -
	"1.1.12.2"
	    -
	"1.1.12.3"
	    -
	"1.1.12.10"
	    -
	"1.1.12.11"
	    -
	"1.1.13.1"
	    -
	"1.1.13.2"
	    -
	"1.1.13.3"
	    -
	"1.1.13.4"
	    -
	"1.1.13.5"
	    -
	"1.1.13.6"
	    -
	"1.1.13.13"
	    -
	"1.1.13.14"
	    -
	"1.1.14" {
	    set		retries		1
	    set		timeout		"-timeout 5"
	    snmptcl_proc_mkmsg \
			"this involves timeouts, so may take a while..."
	    updateMsg
	}

	default {
	    set		retries		[$snmptcl_agent config retries]
	    set		timeout		""
	}
    }

    updateTest	$test		"in progress"
    set		iHandle		[eval $snmptcl_agent $op callback1.1.12 \
				     [list $vb] -magic $magic -retries \
				     $retries $timeout]
    set		info($iHandle)	[list $test $var $op $ops $magic $retries \
				     $timeout]
}

proc	callback1.1.12	{pHandle iHandle eStatus eIndex vBinds} {
    global	info sysUpTime snmpTrapOID coldStart

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		var		[lindex $data 1]
    set		op		[lindex $data 2]
    set		ops		[lindex $data 3]
    set		magic		[lindex $data 4]
    set		retries		[lindex $data 5]
    set		timeout		[lindex $data 6]

    switch -- $eStatus {
	destroyed {
	    return
	}

	timeout {
	    switch $test {
		"1.1.12.1"
		    -
		"1.1.12.2"
		    -
		"1.1.12.3"
		    -
		"1.1.12.10"
		    -
		"1.1.12.11"
		    -
		"1.1.13.1"
		    -
		"1.1.13.2"
		    -
		"1.1.13.3"
		    -
		"1.1.13.4"
		    -
		"1.1.13.5"
		    -
		"1.1.13.6"
		    -
		"1.1.13.13"
		    -
		"1.1.13.14"
		    -
		"1.1.14" {
		}

		default {
		    failedTest		$test "" \
					[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
	    }
	}

	noError {
# 1.1.12.9 and 1.1.13.12 really shouldn't be here...
	    switch $test {
		"1.1.12.1"
		    -
		"1.1.12.4"
		    -
		"1.1.12.5"
		    -
		"1.1.12.6"
		    -
		"1.1.12.7"
		    -
		"1.1.12.8"
		    -
		"1.1.12.9"
		    -
		"1.1.12.10"
		    -
		"1.1.13.1"
		    -
		"1.1.13.7"
		    -
		"1.1.13.8"
		    -
		"1.1.13.9"
		    -
		"1.1.13.10"
		    -
		"1.1.13.11"
		    -
		"1.1.13.12"
		    -
		"1.1.13.13" {
		}

		default {
		    failedTest		$test "" \
					[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
	    }
	}

	tooBig {
	    switch $test {
		"1.1.12.9" {
		}

		"1.1.13.12" {
		    if {[llength $vBinds] != 0} {
			failedTest	$test "" \
					[rsp2text $eStatus $eIndex $vBinds]
			return
		    }
		}

		default {
		    failedTest		$test "" \
					[rsp2text $eStatus $eIndex $vBinds]
		    return
		}
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {[llength $ops] == 0} {
	passedTest	$test
	return
    }
    set		op	[lindex $ops 0]
    set		ops	[lrange $ops 1 end]

    switch $test {
	"1.1.12.7"
	    -
	"1.1.13.10" {
	    set		vb	""
	}

	"1.1.12.8"
	    -
	"1.1.13.11" {
	    set		vb	[list [list $var Integer32 0]]
	}

	"1.1.12.9"
	    -
	"1.1.13.12" {
	    for {set i 0} {$i < 24} {incr i} {
		lappend	vb	[list $var]
	    }
	}

	"1.1.14" {
	    set		vb	[list [list $sysUpTime.0 TimeTicks 0] \
				      [list $snmpTrapOID.0 ObjectID \
					    $coldStart]]
	}

	default {
	    if {$op == "set"} {
		set	vb	$vBinds
	    } else {
		set	vb	[list [list $var]]
	    }
	}
    }

    set		iHandle		[eval $pHandle $op callback1.1.12 [list $vb] \
				     -magic $magic -retries $retries $timeout]
    set		info($iHandle)	[list $test $var $op $ops $magic $retries \
				     $timeout]
}


#
#
#

proc	test1.1.15	{test} {
    global	info variables tour snmptcl_agent

    set		var		[lindex $variables 0]
    set		tours		{ 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 }

    updateTest	$test		"in progress"
    set		tour		[lindex $tours 0]
    set		tours		[lrange $tours 1 end]
    for {set i 0} {$i < $tour} {incr i} {
	set	iHandle		[$snmptcl_agent next callback1.1.15 \
				     [list [list $var]]]
	set	info($iHandle)	[list $test $var $tours]
    }
}

proc	callback1.1.15	{pHandle iHandle eStatus eIndex vBinds} {
    global	info tour

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		var		[lindex $data 1]
    set		tours		[lindex $data 2]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		set		tour	0
		return
	    }
	    set		vb	[lindex $vBinds 0]
	    if {[lindex $vb 1] == "exception"} {
		failedTest	$test "agent returned [lindex $vb 2]" \
				[rsp2text $eStatus $eIndex $vBinds]
		set		tour	0
		return
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    set		tour	0
	    return
	}
    }

    if {[incr tour -1] > 0} {
	return
    }
    if {[llength $tours] == 0} {
	passedTest	$test
    }

    set		tour		[lindex $tours 0]
    set		tours		[lrange $tours 1 end]
    for {set i 0} {$i < $tour} {incr i} {
	set	iHandle		[$pHandle next callback1.1.15 \
				     [list [list $var]]]
	set	info($iHandle)	[list $test $var $tours]
    }
}


#
#
#

proc	test1.2.1	{test} {
    global	info snmptcl_agent

    if {[set resources [$snmptcl_agent config resources]] == ""} {
	error	"no object resources for agent?!?"
    }
    foreach	resource	$resources {
	if {[catch { snmpinfo agentprop $resource sysUpTime support }]} {
	    error	"no A-C information on [oid2ode $resource]"
	}
    }
    set		resource	[lindex $resources 0]
    set		resources	[lrange $resources 1 end]

    global	[set array snmptcl_agentcap_[join [split $resource "."] "_"]]

    set		objects		""
    foreach	object		[array names $array] {
	set	elems		[split $object ","]
	if {[lindex $elems 1] != "support"} {
	    continue
	}
	lappend	objects		[set var [lindex $elems 0]]
	if {[catch { set value [snmpinfo mibprop $var value] }]} {
	    talkWindow	.note	"" "Note Well" "unknown object: $var"
	}
    }
    set		objects		[lsort -command oidcmp $objects]
    set		object		[lindex $objects 0]
    set		objects		[lrange $objects 1 end]

    updateTest	$test		"in progress" "checking [oid2ode $resource]"
    set		iHandle		[$snmptcl_agent next callback1.2.1 \
				     [list [list $object]]]
    set		info($iHandle)	[list $test $object $objects $resources 1]
}

proc	callback1.2.1	{pHandle iHandle eStatus eIndex vBinds} {
    global	info snmptcl_verbose

    set		data		$info($iHandle)
    unset	info($iHandle)
    set		test		[lindex $data 0]
    set		last		[lindex $data 1]
    set		objects		[lindex $data 2]
    set		resources	[lindex $data 3]
    set		iters	[lindex $data 4]

    switch -- $eStatus {
	destroyed {
	    return
	}

	noError {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[lindex $vb 1] == "exception"} {
		losingTest	$test	"agent returned [lindex $vb 2] when asked for [oid2ode $last]" \
					[rsp2text $eStatus $eIndex $vBinds]
	    } elseif {![snmptcl_proc_oidcontains $last \
		            [set var [lindex $vb 0]]]} {
		losingTest	$test	"agent returned [oid2ode $var] when asked for [oid2ode $last]" \
				[rsp2text $eStatus $eIndex $vBinds]
	    } else {
		vBind1.1.2	$test $pHandle $eStatus $eIndex $vBinds
	    }
	}

	noSuchName {
	    if {[set l [llength $vBinds]] != 1} {
		failedTest	$test "expecting 1 VBs, got $l" \
				[rsp2text $eStatus $eIndex $vBinds]
		return
	    }
	    set	    vb			[lindex $vBinds 0]
	    if {[oidcmp $last [set var [lindex $vb 0]]] != 0} {
		losingTest	$test "lexicographic error, $last != $var" \
				[rsp2text $eStatus $eIndex $vBinds]
	    } else {
		losingTest	$test	"agent returned [oid2ode $var] when asked for [oid2ode $last]" \
					[rsp2text $eStatus $eIndex $vBinds]
	    }
	}

	default {
	    failedTest		$test "" [rsp2text $eStatus $eIndex $vBinds]
	    return
	}
    }

    if {$objects == ""} {
	if {$resources == ""} {
	    passedTest		$test	"$iters objects examined"
	    return
	}
	set	resource	[lindex $resources 0]
	set	resources	[lrange $resources 1 end]

	global	[set array snmptcl_agentcap_[join [split $resource "."] "_"]]

	set	objects		""
	foreach	object		[array names $array] {
	    set		elems	[split $object ","]
	    if {[lindex $elems 1] != "support"} {
	        continue
	    }
	    lappend	objects	[lindex $elems 0]
	    if {[catch { set value [snmpinfo mibprop $var value] }]} {
		talkWindow	.note	"" "Note Well" "unknown object: $var"
	    }
	}
	set	objects		[lsort -command oidcmp $objects]
	set	object		[lindex $objects 0]
	set	objects		[lrange $objects 1 end]

	updateTest \
		$test		"in progress" "checking [oid2ode $resource]"
    }
    set		object		[lindex $objects 0]
    set		objects		[lrange $objects 1 end]

    incr	iters
    if {$snmptcl_verbose} {
	if {($iters % 25) == 0} {
	    snmptcl_proc_mkmsg	"$test: continuing with [oid2ode $object]..."
	    update
	}
    }

    set		iHandle		[$pHandle next callback1.2.1 \
				     [list [list $object]]]
    set		info($iHandle)	[list $test $object $objects $resources $iters]
}

#
#
#

proc	test1.2.2	{test} {
    global	variables snmptcl_verbose snmptcl_agent

    if {[set resources [$snmptcl_agent config resources]] == ""} {
	error	"no object resources for agent?!?"
    }
    foreach	resource	$resources {
	if {[catch { snmpinfo agentprop $resource sysUpTime support }]} {
	    error	"no A-C information on [oid2ode $resource]"
	}
    }

    updateTest	$test		"in progress"

    set		iters		0
    foreach	var		$variables {
	if {[catch { set value [snmpinfo mibprop $var value] }]} {
	    losingTest	$test	"unknown object: $var" ""
	    continue
	}

	set	support		0
	foreach	resource	$resources {
	    if {[set support [snmpinfo agentprop $resource $value support]]} {
		break
	    }
	}
	if {!$support} {
	    losingTest	$test	"agent returned $var" \
				"[oid2ode $value] isn't in AGENT-CAPABILITIES"
	}

	incr	iters
	if {$snmptcl_verbose} {
	    if {($iters % 25) == 0} {
	        snmptcl_proc_mkmsg	"$test: $iters variables examined, continuing..."
	        update
	    }
	}
    }

    passedTest	$test		"$iters variables examined"
}

#
# finally, see if snmptcl already set the management context
#

destroy		.title.contexts
pack after	.title.options		[button .title.contexts \
					-command snmptcl_callback_select_contexts \
					-text Contexts...] \
				    {right}

if {![catch { $snmptcl_agent config active }]} {
    .test.system	configure	-text [format "%-74.74s" \
						    "Context: [$snmptcl_agent config name]"]

    $snmptcl_agent bulk	SelectContextCallback [list [list $sysDescr]]
}


set	snmptcl_help(help)	\
"To run a test, double-click left on the test you want.  To get a description of a test, double-click right on the test.  When a test fails, a window containing a summary of the failure pops up.  If you've iconified this window, double-click middle on the test and the window will re-appear.  (If you've destroyed this window, you have to re-run the test.)\n\nThe \"Applications\" button is inactive, don't waste your time on it.\n\nTo select a new agent for testing, click on \"Contexts...\" which will create a dialog for you to enter the identity of the agent you want to test.  The \"Help\" button in this dialog will tell you how to identify the agent.\n\nTo watch the PDU traffic, click on \"Options\" and then on \"Watch\".\n\nFinally, the \"Help\" button displays this window."

update
snmptcl_callback_help .help "Help Running the Test Suite" "test help" help


return
