/* snmptcl.c - SNMP-capable Tcl */

/* snmpVars:

	snmptcl_library
		directory were Tcl programs are stored

	snmptcl_myname
		for top-level window

	snmptcl_debug
		for debugging

	snmptcl_watch
		for traffic tracing

	snmptcl_verbose
		application-specific
 */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/*
 * $Header$
 *
 *
 * $Log$
 */

/*			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <varargs.h>
#include <sys/file.h>
#include <tk.h>


#define	NOTOK	(-1)
#define	NULLCP	((char *) 0)

/*    DATA */

/* MAIN */

static	char   *myname = "snmptcl";

	int	debug = 0;
static	char   *objects = NULLCP;
static	char   *pParty = "";
static	int	partysync = 1;
static	int	verbose = 0;
static	int	watch = 0;

static	Tk_ArgvInfo tclArgv[] = {
    { "-debug",		TK_ARGV_CONSTANT, (char *) 1,	(char *) &debug,
	"Use debug mode" },
    { "-nodebug",	TK_ARGV_CONSTANT, (char *) 0,	(char *) &debug,
	"Don't ..." },
    { "-objectsfile",	TK_ARGV_STRING,   NULLCP,	(char *) &objects,
	"Pathname of objects file" },
    { "-party",		TK_ARGV_STRING,   NULLCP,	(char *) &pParty,
	"Prefix of party database" },
    { "-verbose",	TK_ARGV_CONSTANT, (char *) 1,	(char *) &verbose,
	"Use verbose mode" },
    { "-noverbose",	TK_ARGV_CONSTANT, (char *) 0,	(char *) &verbose,
	"Don't ..." },
    { "-update",	TK_ARGV_CONSTANT, (char *) 1,	(char *) &partysync,
	"Synchronize party database when exiting (if necessary)" },
    { "-noupdate",	TK_ARGV_CONSTANT, (char *) 0,	(char *) &partysync,
	"Don't ..." },
    { "-watch",		TK_ARGV_CONSTANT, (char *) 1,	(char *) &watch,
	"Use watch mode" },
    { "-nowatch",	TK_ARGV_CONSTANT, (char *) 0,	(char *) &watch,
	"Don't ..." },
    { NULLCP,		TK_ARGV_END,	  NULLCP,	NULLCP,	 NULLCP }
};


static	Tcl_Interp *tclInterp;


int	Tcl_AppInit (), Tcl_AppInitAux ();
void	intrser ();
int	ExitCommand ();

/*    MAIN */

int	Tcl_AppInit (interp)
Tcl_Interp *interp;
{
    if (Tcl_AppInitAux (interp) == TCL_OK)
	return TCL_OK;

    fprintf (stderr, "%s\n", interp -> result);
    exit (1);
    /* NOTREACHED */
}

/*  */

static int  Tcl_AppInitAux (interp)
Tcl_Interp *interp;
{
    int	    argc;
    char   *cp,
	    buffer[BUFSIZ],
	    file[BUFSIZ],
	    library[BUFSIZ],
	   *args,
	  **argv;
    Tk_Window	main;

    main = Tk_MainWindow (tclInterp = interp);

    if (Blt_Init (interp) != TCL_OK
	    || Tcl_Init (interp) != TCL_OK
	    || Tk_Init (interp) != TCL_OK)
	return TCL_ERROR;

    argc = atoi (Tcl_GetVar (interp, "argc", TCL_GLOBAL_ONLY));
    Tcl_SplitList (interp, Tcl_GetVar (interp, "argv", TCL_GLOBAL_ONLY),
		   &argc, &argv);

    if (Tk_ParseArgv (interp, main, &argc, argv, tclArgv,
		      TK_ARGV_DONT_SKIP_FIRST_ARG) != TCL_OK)
	return TCL_ERROR;

    args = Tcl_Merge (argc, argv);
    Tcl_SetVar (interp, "argv", args, TCL_GLOBAL_ONLY);
    free (args);
    (void) sprintf (buffer, "%d", argc);
    Tcl_SetVar (interp, "argc", buffer, TCL_GLOBAL_ONLY);

    if (SNMP_Init (interp, myname, objects, pParty, partysync, &watch, library)
	    != TCL_OK)
	return TCL_ERROR;
    Tcl_CreateCommand (interp, "exit", ExitCommand, (ClientData) 0,
		       (void (*) ()) 0);

    if (MISC_Init (interp) != TCL_OK)
	return TCL_ERROR;

    (void) Tcl_SetVar2 (interp, "snmptcl_library", NULLCP, library,
			TCL_GLOBAL_ONLY);
    (void) Tcl_SetVar2 (interp, "snmptcl_myname", NULLCP, myname,
			TCL_GLOBAL_ONLY);

    (void) Tcl_LinkVar (interp, "snmptcl_debug", &debug, TCL_LINK_INT);
    (void) Tcl_LinkVar (interp, "snmptcl_verbose", &verbose, TCL_LINK_INT);
    (void) Tcl_LinkVar (interp, "snmptcl_watch", &watch, TCL_LINK_INT);

    (void) sprintf (file, "%s.tcl", myname);
    if (access (cp = file, R_OK) == NOTOK) {
	cp = library;
	(void) sprintf (library + strlen (library), "/%s", file);
    }
    if (Tcl_EvalFile (interp, cp) != TCL_OK)
	return TCL_ERROR;

    (void) signal (SIGINT, intrser);

    return TCL_OK;
}

/*  */

#ifdef	BSDSIGS
/* ARGSUSED */
#endif

static void  intrser (sig)
int	sig;
{
    (void) signal (sig, SIG_DFL);

    (void) Tcl_Eval (tclInterp, "exit");
    exit (1);
    /* NOTREACHED */
}

/*  */

/* ARGSUSED */

static int  ExitCommand (clientData, interp, argc, argv)
ClientData	clientData;
Tcl_Interp     *interp;
int		argc;
char	      **argv;
{
    int	    status;
    char    buffer[BUFSIZ];

    switch (argc) {
	case 1:
	    status = 0;
	    break;

	case 2:
	    if (Tcl_GetInt (interp, argv[1], &status) != TCL_OK)
		return TCL_ERROR;
	    break;

	default:
	    (void) sprintf (buffer, "argc=%d; expecting 1 or 2", argc);
	    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
	    return TCL_ERROR;
    }

    if (SNMP_Exit (interp) != TCL_OK) {
	fprintf (stderr, "%s\n", interp -> result);
	if (!status)
	    status = 1;
    }

    exit (status);
    /* NOTREACHED */
}
